////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

%pythoncode %{

#################################################

import os.path

import modellib

#################################################

def createInputPars():
	"""
	createInputPars()

	Creates a LimeInputPars object with default values		
	"""		

# Create LimeInputPars struct
	# and initialize with default values 
	# as expected by LIME
	
	par = LimeInputPars()
	
	# Set default values:
	par.radius	    = 0.0
	par.minScale	= 0.0
	par.pIntensity	= 0   
	par.sinkPoints	= 0

	par.sampling	= 0
	par.tcmb       	= 2.725
	par.moldatfile	= ""
	par.dust        = None
	par.outputfile	= None
	par.gridfile    = None
	par.pregrid     = None

	par.lte_only    = 0
	par.blend       = 0
	par.antialias   = 1
	par.polarization = 0
	
	return par
	
#################################################

def _adoptInputPar(limePar, nImages):

	# Copy LimeInputPars struct limePar (used in Python)
	# to inputPars struct par (used in LIME).
	# Since a char** as element of a structure cannot be
	# passed to LIME, lime.Par.moldatfile is copied into
	# a list which is returned and can be passed to LIME

	par = inputPars()
	
	par.nImages = nImages
	par.nSpecies = 1

	par.radius = limePar.radius
	par.minScale = limePar.minScale
	par.tcmb = limePar.tcmb
	par.sinkPoints = limePar.sinkPoints
	par.pIntensity = limePar.pIntensity
	par.blend = limePar.blend
	par.outputfile = limePar.outputfile
	par.gridfile = limePar.gridfile
	par.pregrid = limePar.pregrid
	par.dust = limePar.dust
	par.sampling = limePar.sampling
	par.lte_only = limePar.lte_only
	par.antialias = limePar.antialias
	par.polarization = limePar.polarization
	
	par.moldatfile = []
	moldatfiles = [limePar.moldatfile]
			
	return par, moldatfiles
		
#################################################

def createImage():
	"""
	createImage()

	Creates a LimeImage object with default values		
	"""		

	# Create LimeImage struct
	# and initialize with default values 
	# as expected by LIME

	img = LimeImage()
	
	img.pxls       	= 0
	img.imgres     	= 0.0
	img.theta	    = 0.0
	img.distance	= 0.0		
	img.unit       	= 0
	img.filename   	= None

	img.phi	       	= 0.0		
	img.source_vel	= 0.0
	img.nchan      	= 0
	img.velres     	= -1.0
	img.trans      	= -1
	img.freq	    = -1.0
	img.bandwidth	= -1.0
	
	return img

#################################################
			
def _adoptImage(limeImg):

	# Copy LimeImage struct limeImg (used in Python)
	# to image struct img (used in LIME).

	img = image()
	
	img.nchan = limeImg.nchan
	img.trans = limeImg.trans
	img.velres = limeImg.velres
	img.imgres = limeImg.imgres
	img.pxls = limeImg.pxls
	img.unit = limeImg.unit
	img.freq = limeImg.freq
	img.bandwidth = limeImg.bandwidth
	img.filename = limeImg.filename
	img.source_vel = limeImg.source_vel
	img.theta = limeImg.theta
	img.phi = limeImg.phi
	img.distance = limeImg.distance

	img.pixel = None	
	
	return img	
	
#################################################

def runLime(par, images, debug=False):
	"""
    runLime(par, images, debug=False)

    Performs basic tests on the passed input parameters and image definitions and runs LIME.

    Parameters:
    -----------
    LimeInputPars par: Basic settings for the LIME run
    
    list of LimeImage images: Definitions for images for the LIME run
    
    bool debug: If True, print values of arguments passed to LIME before actually calling LIME 
	"""

	# Check the various parameters in
	# LimeInputPars par and list of LimeImage objects images,
	# make list images an imageArray, and run LIME
	#
	# If debug is True, print debug output
	#
	_check(par, images)
	
	iPar, moldatfiles = _adoptInputPar(par, len(images))

	ia = imageArray(len(images))
	iImage = 0
	for image in images:
		img = _adoptImage(image)
		ia[iImage] = img 
		iImage += 1				

	if debug:
		_printDebugModel()
		_printDebugPars(iPar, moldatfiles)
		for iImage in range(len(images)):		
			_printDebugImage(ia[iImage], iImage)
	
	#
	# Check that configuration is complete and finalized:
	#
	if not modellib.isConfigurationComplete():
		raise Exception("Configuration not complete")
	if not modellib.isFinalizedConfiguration():
		raise Exception("Configuration not finalized")
		
	run(iPar, ia, moldatfiles)

#################################################

def _printDebugModel():
	print "DEBUG Model:"
	modelID = modellib.getCurrentModelID()
	if modelID:
		print "DEBUG   modelID:\t", modelID
	
		print "DEBUG   Parameters:"
		for paramID in modellib.getParamIDs(modelID):
			paramType = modellib.getParamType(modelID, paramID) 
			val = "?"
			if paramType == "int":
				val = modellib.getParamInt(paramID)
			if paramType == "double":
				val = modellib.getParamDouble(paramID)
			if paramType == "string":
				val = modellib.getParamString(paramID)
			print "DEBUG     paramID:\t", paramID, "\t", val
			
		print "DEBUG   Results:"
		for resultID in modellib.getResultIDs():
			if modellib.isCurrentResultModel(resultID):
				print "DEBUG     resultID:\t", resultID, "\tMODEL"
			else:
				if modellib.isCurrentResultFunction(resultID):
					print "DEBUG     resultID:\t", resultID, "\tFUNCTION"
					functionID = modellib.getFunctionID(resultID)
					print "DEBUG       functionID:\t", functionID
					print "DEBUG       FunctionParameters:"
					for paramID in modellib.getFunctionParamIDs(functionID):	
						paramType = modellib.getFunctionParamType(functionID, paramID) 
						val = "?"
						if paramType == "int":
							val = modellib.getFunctionParamInt(resultID, paramID)
						if paramType == "double":
							val = modellib.getFunctionParamDouble(resultID, paramID)
						if paramType == "string":
							val = modellib.getFunctionParamString(resultID, paramID)
						print "DEBUG         functionParamID:\t", paramID, "\t", val
				else:
					print "DEBUG     resultID:\t", resultID, "\tNONE"
	else:					
		print "DEBUG   modelID:\t -"

#################################################

def _printDebugPars(iPar, moldatfiles):
	print "DEBUG ******"
	print "DEBUG Input Parameter:"
	print "DEBUG   radius:\t", iPar.radius
	print "DEBUG   minScale:\t", iPar.minScale
	print "DEBUG   tcmb:\t", iPar.tcmb
	print "DEBUG   ncell:\t", iPar.ncell
	print "DEBUG   sinkPoints:\t", iPar.sinkPoints
	print "DEBUG   pIntensity:\t", iPar.pIntensity
	print "DEBUG   nImages:\t", iPar.nImages
	print "DEBUG   nSpecies:\t", iPar.nSpecies
	print "DEBUG   blend:\t", iPar.blend
	print "DEBUG   outputfile:\t", iPar.outputfile
	print "DEBUG   inputfile:\t", iPar.inputfile
	print "DEBUG   gridfile:\t", iPar.gridfile
	print "DEBUG   pregrid:\t", iPar.pregrid
	print "DEBUG   dust:\t", iPar.dust
	print "DEBUG   sampling:\t", iPar.sampling
	print "DEBUG   collPart:\t", iPar.collPart
	print "DEBUG   lte_only:\t", iPar.lte_only
	print "DEBUG   antialias:\t", iPar.antialias
	print "DEBUG   polarization:\t", iPar.polarization
	print "DEBUG   moldatfiles:\t", moldatfiles
	
#################################################

def _printDebugImage(img, iImage):
	print "DEBUG ******"
	print "DEBUG Image ", iImage
	print "DEBUG   nchan:\t", img.nchan
	print "DEBUG   trans:\t", img.trans
	print "DEBUG   velres:\t", img.velres
	print "DEBUG   imgres:\t", img.imgres
	print "DEBUG   pxls:\t", img.pxls
	print "DEBUG   unit:\t", img.unit
	print "DEBUG   freq:\t", img.freq
	print "DEBUG   bandwidth:\t", img.bandwidth
	print "DEBUG   filename:\t", img.filename
	print "DEBUG   souce_vel:\t", img.source_vel
	print "DEBUG   theta:\t", img.theta
	print "DEBUG   phi:\t", img.phi
	print "DEBUG   distance:\t", img.distance

#################################################

def _check(par, images):

	# Perform checks on parameters
	
	_checkPar(par)
	_checkImages(par, images)
	_checkPolarization(par, images)
	_checkInputFiles(par)
  
#################################################

def _checkPar(par):
	
	# Perform check on parameters in 
	# inputPars object par
	
	pGtZero = {
		"radius": par.radius,
		"minScale": par.minScale,
		"pIntensity": par.pIntensity,
		"sinkPoins": par.sinkPoints,
		"tcmb": par.tcmb,
	}
	for key in pGtZero.keys():
		if not pGtZero[key] > 0:
			raise Exception("Invalid input parameter '%s': must be > 0" % key)

	pOneOrZero = {
		"sampling": par.sampling,
		"lte_only": par.lte_only,
		"blend": par.blend,
		"polarization": par.polarization,
	}
	for key in pOneOrZero.keys():
		if not pOneOrZero[key] in [0, 1]:
			raise Exception("Invalid input parameter '%s': must be 0 or 1" % key)

#################################################

def _checkInputFiles(par):

	# Check the filenames specified in par
	
	# Make sure that empty filenames are None, not "":
	for filename in [par.dust, par.outputfile, par.gridfile, par.pregrid, par.moldatfile]:
		if not filename:
			filename = None

	# moldatfile or dust table required:
	if not par.moldatfile and not par.dust:
		raise Exception("Invalid input files: no dust tables or moldata file specified")
		
	# Make sure that all specified input files exist:
	infiles = {
 		"dust": par.dust,
 		"moldatfile": par.moldatfile,
		"pregrid": par.pregrid,
	}
	for key in infiles.keys():
		if infiles[key]:
			if not os.path.isfile(infiles[key]):
				raise Exception("Invalid input parameter '%s': file does not exist" % key)

#################################################

def _checkImages(par, images):

	# Check parameters of image objects in
	# list images

	# At least one image must be specified:
	if not images:
		raise Exception("Invalid images: no images specified")
  
  	# Check each image:
	imgID = 0
	for img in images:  
 		_checkImage(img, imgID, par)
 		imgID += 1

#################################################
 		
def _checkImage(img, imgID, par):

	# Check parameters of image img including
	# consistency with inputPars par

  	iGtZero = {
		"pxls": img.pxls,
		"imgres": img.imgres,
		"distance": img.distance,
	}
	for key in iGtZero.keys():
		if not iGtZero[key] > 0:
			raise Exception("Invalid parameter '%s' for image %d: must be > 0" % (key, imgID))

	if not img.unit in [0, 1, 2, 3]:
		raise Exception("Invalid parameter 'unit' for image %d: must be in [0, 1, 2, 3]" % imgID)
		
	if not img.filename:
		raise Exception("No filename specified for image %d" % imgID)
		
	# Check complex dependencies for parameters:
	isOK = True
	if not img.nchan:
		# Continuum image:
		if not par.dust:
			isOK = False
		if par.moldatfile:
			isOK = False
		
		if (img.velres >= 0) or (img.trans >= 0) or (img.bandwidth >= 0):
			isOK = False
		if img.freq <= 0:
			isOK = False

	else:
		# Line image:
		if not par.moldatfile:
			isOK = False

		if img.freq >= 0:
			# Frequency specified:
			if (img.trans >= 0) or (img.velres >= 0):
				isOK = False
			if img.bandwidth <= 0:
				isOK = False
		else:
			# Transition specified:
			if (img.freq >= 0) or (img.bandwidth >= 0):
				isOK = False
			if img.velres <= 0:
				isOK = False

	if not isOK:
		raise Exception("Invalid combination of parameters for image %d" % imgID)	
		
#################################################

def _checkPolarization(par, images):

	# Check par.polarization for consistency
	# with setup of the Model Library and
	# with image parameters

	if (par.polarization):
		# Nothing to check if 0
		
		# Does Model Library provide magnetic field?
		if not (modellib.isCurrentResultModel('bmag') or modellib.isCurrentResultFunction('bmag')):
			raise Exception("Invalid setup for polarization: Model Library does not provide magnetic field")
			
		# No polarization for line images:
		for img in images:
			if img.nchan:
				# Line image:
				raise Exception("Invalid setup for polarization: Not implemented for line images")
		
%}
