////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <stdexcept>
#include <cmath>
#include <iostream>

#define PI 3.141592653589793


#include "artistGrid.h"

using namespace std;


ArtistRegularGrid::ArtistRegularGrid(const vector<int>& nGridPoints,
        							 const int& iCoordSys,
        							 const int& iGridLocation)
: m_iCoordSys(iCoordSys),
  m_iGridLocation(iGridLocation),
  m_iLastFoundIndices(vector<unsigned int>(3, 0))
{
	// Allocate vectors for grid points:
	m_xGrid = vector<vector<double> >(3);
	m_iNextIndex = vector<unsigned int>(3);
	for (int iDim=0; iDim<3; iDim++) {
		if ( iGridLocation==0 ) {
			// Cell centered: Make room for one more grid point:
			if ( nGridPoints[iDim] ) {
				// But only if this dimension is used
				m_xGrid[iDim] = vector<double>(nGridPoints[iDim]+1);
				m_iNextIndex[iDim] = 1; // Leave first element free
			} else {
				m_xGrid[iDim] = vector<double>(0);
				m_iNextIndex[iDim] = 0;
			}
		} else {
			// Cell Interfaces:
			m_xGrid[iDim] = vector<double>(nGridPoints[iDim]);
			m_iNextIndex[iDim] = 0;
		}
	}

	// Upper and lower bounds:
	m_hasLowerBound = vector<bool>(3);
	m_lowerBound = vector<double>(3);
	m_hasUpperBound = vector<bool>(3);
	m_upperBound = vector<double>(3);

	if ( iCoordSys == 0) {
		// Cartesian: x, y z
		for ( int iDim=0; iDim<3; iDim++ ) {
			m_hasLowerBound[iDim] = false;
			m_lowerBound[iDim] = 0.;
			m_hasUpperBound[iDim] = false;
			m_upperBound[iDim] = 0.;
		}
	} else if ( iCoordSys == 1 ) {
		// Cylidrical: r, phi, z
		m_hasLowerBound[0] = true;
		m_lowerBound[0] = 0.;
		m_hasUpperBound[0] = false;
		m_upperBound[0] = 0.;

		m_hasLowerBound[1] = true;
		m_lowerBound[1] = 0.;
		m_hasUpperBound[1] = true;
		m_upperBound[1] = 2. * PI;

		m_hasLowerBound[2] = false;
		m_lowerBound[2] = 0.;
		m_hasUpperBound[2] = false;
		m_upperBound[2] = 0.;
	} else if ( iCoordSys==2 ) {
		// Spherical: r, theta, phi
		m_hasLowerBound[0] = true;
		m_lowerBound[0] = 0.;
		m_hasUpperBound[0] = false;
		m_upperBound[0] = 0.;

		m_hasLowerBound[1] = true;
		m_lowerBound[1] = 0.;
		m_hasUpperBound[1] = true;
		m_upperBound[1] = PI;

		m_hasLowerBound[2] = true;
		m_lowerBound[2] = 0.;
		m_hasUpperBound[2] = true;
		m_upperBound[2] = 2. * PI;
	} else {
        throw domain_error("Invalid coordinate system");
	}
}

void
ArtistRegularGrid::addGridPoint(const double& x, const int& iDim)
{
	if ( m_iGridLocation == 0 )
		addPointCenter(x, iDim);
	else
		addPointInterface(x, iDim);
}

void
ArtistRegularGrid::addPointCenter(const double& x, const int& iDim)
{
	unsigned int& index = m_iNextIndex[iDim];

	// Check that grid point is within bounds:
	if ( m_hasLowerBound[iDim] ) {
		if ( x <= m_lowerBound[iDim] )
			throw  domain_error("Grid point out of bounds");
	}
	if ( m_hasUpperBound[iDim] ) {
		if ( x >= m_upperBound[iDim] )
			throw  domain_error("Grid point out of bounds");
	}

	// Check that grid points in strictly monotonically increasing order:
	bool isFirstGridPoint = false;
	if ( index == 1 )
		isFirstGridPoint = true;
	if ( ! isFirstGridPoint ) {
		if ( m_xGrid[iDim][index-1] >= x )
			throw  domain_error("Grid points not in strictly monotonically increasing order");
	}

	m_xGrid[iDim].at(index) = x;
	index++;
}

void
ArtistRegularGrid::addPointInterface(const double& x, const int& iDim)
{
	unsigned int& index = m_iNextIndex[iDim];

	bool isFirstGridPoint = false;
	if ( index == 0 )
		isFirstGridPoint = true;
	bool isLastGridPoint = false;
	if ( index == m_xGrid[iDim].size()-1 )
		isLastGridPoint = true;

	if ( isFirstGridPoint ) {
		m_xGrid[iDim].at(index) = x;

		// If lower bound exists and x below it, set first grid point to lower bound:
		if ( m_hasLowerBound[iDim] )
			if ( x < m_lowerBound[iDim])
				m_xGrid[iDim].at(index) = m_lowerBound[iDim];
	}

	if ( isLastGridPoint ) {
		m_xGrid[iDim].at(index) = x;

		// If upper bound exists and x above it, set last grid point to upper bound:
		if ( m_hasUpperBound[iDim] )
			if ( x > m_upperBound[iDim])
				m_xGrid[iDim].at(index) = m_upperBound[iDim];

		// Check that grid points in strictly monotonically increasing order:
		if ( m_xGrid[iDim][index-1] >= m_xGrid[iDim][index] )
			throw  domain_error("Grid points not in strictly monotonically increasing order");
	}

	if ( ! isFirstGridPoint && ! isLastGridPoint ) {
		// Check that grid point is within bounds:
		if ( m_hasLowerBound[iDim] ) {
			if ( x < m_lowerBound[iDim] )
				throw  domain_error("Grid point out of bounds");
		}
		if ( m_hasUpperBound[iDim] ) {
			if ( x > m_upperBound[iDim] )
				throw  domain_error("Grid point out of bounds");
		}

		// Check that grid points in strictly monotonically increasing order:
		if ( m_xGrid[iDim][index-1] >= x )
			throw  domain_error("Grid points not in strictly monotonically increasing order");

		m_xGrid[iDim].at(index) = x;
	}
	index++;
}


void
ArtistRegularGrid::finalizeGrid()
{
	// Check that all points have been defined:
	for ( int iDim=0; iDim<3; iDim++ ) {
		unsigned int npoints = m_xGrid[iDim].size();
		if ( npoints ) {
			if ( m_iNextIndex[iDim]<npoints )
				throw domain_error("Not all grid points defined");
		}
	}

	// Transfer cell centered grid to cell interface grid:
	if ( m_iGridLocation==0 ) {
		for ( int iDim=0; iDim<3; iDim++ ) {
			unsigned int npoints = m_xGrid[iDim].size();
			if ( npoints ) {
				if ( npoints == 1 ) {
					// Somewhat pathological...
					if ( ! m_hasLowerBound[iDim] )
						throw domain_error("Invalid grid");
					if ( ! m_hasUpperBound[iDim] )
						throw domain_error("Invalid grid");
					m_xGrid[iDim][0] = m_lowerBound[iDim];
					m_xGrid[iDim][1] = m_upperBound[iDim];
				} else {
					double x1 = m_xGrid[iDim][1]; // Needed later
					unsigned int n = npoints -1;      // Last index

					// Middle points:
					// Set interface location in middle between two cell centers:
					for( unsigned int i=1; i<n; i++ )
						m_xGrid[iDim][i] = ( m_xGrid[iDim][i] + m_xGrid[iDim][i+1] ) / 2.;

					// First and last point:
					// Set either to lower or upper bound, or make interfaces symmetric
					// wrt original cell center:
					if ( m_hasLowerBound[iDim] )
						m_xGrid[iDim][0] = m_lowerBound[iDim];
					else
						m_xGrid[iDim][0] = x1 - ( m_xGrid[iDim][1] - x1 );

					if ( m_hasUpperBound[iDim] )
						m_xGrid[iDim][n] = m_upperBound[iDim];
					else
						m_xGrid[iDim][n] = m_xGrid[iDim][n] + ( m_xGrid[iDim][n] - m_xGrid[iDim][n-1] );
				}
			}
		}
	}

	// <RS_DEBUG>
	/*
	for ( int iDim=0; iDim<3; iDim++) {
		cout << iDim << ": " << m_xGrid[iDim].size() << endl;
		for ( unsigned int i=0; i<m_xGrid[iDim].size(); i++)
			cout << m_xGrid[iDim][i] << " ";
		cout << endl;
	}
	*/
	// </RS_DEBUG>
}

const int
ArtistRegularGrid::getMaxIndex() const
{
	int maxIndex = 1;
	bool pointsFound = false;

	for ( int iDim=0; iDim<3; iDim++ ) {
		unsigned int npoints = m_xGrid[iDim].size();
		if ( npoints ) {
			maxIndex *= ( npoints - 1 );
			pointsFound = true;
		}
	}

	if ( ! pointsFound )
		return -1;
	return maxIndex - 1;
}


const int
ArtistRegularGrid::getIndex(const vector<double>& coords) const
{
	int index = 0;

	for ( int iDim=0; iDim<3; iDim++ ) {
		const vector<double>& xGrid = m_xGrid[iDim];
		const unsigned int& nGrid = xGrid.size();

		if ( nGrid ) {
			const double& x = coords[iDim];
			unsigned int& lastFoundIndex = m_iLastFoundIndices[iDim];

			int dimIndex;

			if ( xGrid[lastFoundIndex] <= x && x <= xGrid[lastFoundIndex + 1] ) {
				// Try last found position:
				dimIndex = lastFoundIndex;
			} else {
				dimIndex = binarySearch(xGrid, x);
			}

			// <RS_DEBUG>
			/*
			cout << dimIndex << " ";
			*/
			// </RS_DEBUG>

			if ( dimIndex < 0 ) {
				// <RS_DEBUG>
				/*
				cout << endl;
				*/
				// </RS_DEBUG>
				return -1;
			}

			lastFoundIndex = dimIndex;
			index = index * ( nGrid - 1 ) + dimIndex;
		}
	}


	// <RS_DEBUG>
	/*
	cout << "\t" << index << endl;
	*/
	// </RS_DEBUG>

	return index;
}

const int
ArtistRegularGrid::binarySearch(const vector<double>& xGrid, const double& x) const
{
	int low=0;
	int high = xGrid.size() - 2;

	while (low <= high) {
        int middle = low + (high - low)/2;
        if ( xGrid[middle] <= x  && x <= xGrid[middle+1] )
            return middle;
        else if ( x  < xGrid[middle] )
        	high = middle - 1;
        else
            low = middle + 1;
    }
    return -1;
}

