////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <vector>
#include <stdexcept>
#include <math.h>

#define PI 3.141592653589793


#include "artistTabdata.h"

using namespace std;

ArtistTabdata::ArtistTabdata()
: m_iNextIndex(0)
{}

ArtistTabdata::~ArtistTabdata()
{}

ArtistScalarTabdata::ArtistScalarTabdata(const unsigned int& nDataPoints)
: ArtistTabdata()
{
	m_data = vector<double>(nDataPoints);
}

ArtistScalarTabdata::~ArtistScalarTabdata()
{}

void
ArtistScalarTabdata::addData(const double& data)
{
	m_data.at(m_iNextIndex) = data;
	m_iNextIndex++;
}

const double&
ArtistScalarTabdata::getData(const unsigned int& index) const
{
	return m_data.at(index);
}


ArtistVectorTabdata::ArtistVectorTabdata(const unsigned int& nDataPoints)
: ArtistTabdata()
{
	m_data = vector<vector<double> >(nDataPoints);
}

ArtistVectorTabdata::~ArtistVectorTabdata()
{}

void
ArtistVectorTabdata::addData(const vector<double>& data)
{
	m_data.at(m_iNextIndex) = data;
	m_iNextIndex++;
}

const vector<double>&
ArtistVectorTabdata::getData(const unsigned int& index) const
{
	return m_data.at(index);
}


const vector<double>
CoordSystemCart::coordCartToSys(const double& x, const double& y, const double& z) const
{
	vector<double> coordSys = vector<double>(3);

	coordSys[0] = x;
	coordSys[1] = y;
	coordSys[2] = z;

	return coordSys;
}

const vector<double>
CoordSystemCart::vecSysToCart(const vector<double>& vecSys, const vector<double>& coordSys) const
{
	vector<double> vecCart = vecSys;
	return vecCart;
}

const vector<double>
CoordSystemCyl::coordCartToSys(const double& x, const double& y, const double& z) const
{
	vector<double> coordSys = vector<double>(3);

	double& rho = coordSys[0];
	double& phi = coordSys[1];

	rho = sqrt(x*x + y*y);

	if ( rho == 0. ) {
		phi = 0.;
	} else {
		if ( y >= 0. )
			phi = acos(x / rho);
		else
			phi = 2. * PI - acos(x / rho);
	}

	coordSys[2] = z;

	return coordSys;
}

const vector<double>
CoordSystemCyl::vecSysToCart(const vector<double>& vecSys, const vector<double>& coordSys) const
{
	vector<double> vecCart = vector<double>(3);

	double& vX = vecCart[0];
	double& vY = vecCart[1];

	const double& vRho = vecSys[0];
	const double& vPhi = vecSys[1];

	const double& phi = coordSys[1];

	vX = vRho * cos(phi) - vPhi * sin(phi);
	vY = vRho * sin(phi) + vPhi * cos(phi);

	vecCart[2] = vecSys[2];

	return vecCart;
}

const vector<double>
CoordSystemSphe::coordCartToSys(const double& x, const double& y, const double& z) const
{
	vector<double> coordSys = vector<double>(3);

	double& r     = coordSys[0];
	double& theta = coordSys[1];
	double& phi   = coordSys[2];

	r = sqrt(x*x + y*y + z*z);

	if ( r == 0. )
		theta = 0.;
	else
		theta = acos(z / r);

	double rho = sqrt(x*x + y*y);

	if ( rho == 0. ) {
		phi = 0.;
	} else {
		if ( y >= 0. )
			phi = acos(x / rho);
		else
			phi = 2. * PI - acos(x / rho);
	}

	return coordSys;
}

const vector<double>
CoordSystemSphe::vecSysToCart(const vector<double>& vecSys, const vector<double>& coordSys) const
{
	vector<double> vecCart = vector<double>(3);

	double& vX = vecCart[0];
	double& vY = vecCart[1];
	double& vZ = vecCart[2];

	const double& vR     = vecSys[0];
	const double& vTheta = vecSys[1];
	const double& vPhi   = vecSys[2];

	const double& theta = coordSys[1];
	const double& phi   = coordSys[2];

	vX = ( vR * sin(theta) + vTheta * cos(theta) ) * cos(phi) - vPhi * sin(phi);
	vY = ( vR * sin(theta) + vTheta * cos(theta) ) * sin(phi) + vPhi * cos(phi);
	vZ = vR * cos(theta) - vTheta * sin(theta);

	return vecCart;
}



