////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <fstream>
#include <stdexcept>

#include <iostream>

#include "artistTabdataReader.h"
#include "artistGrid.h"
#include "artistTabdata.h"

using namespace std;


TabdataIndex::TabdataIndex(const vector<double>& coords, const unsigned int& index, const bool& isValid)
: coords(coords),
  index(index),
  isValid(isValid)
{}




TabdataReader::TabdataReader(const string& filename)
: m_filename(filename),
  m_headerParsed(false),
  m_gridParsed(false),
  m_dataParsed(false),
  m_pGrid(NULL),
  m_pCoordSys(NULL)
{
	m_infile.exceptions( ifstream::failbit | ifstream::badbit | ifstream::eofbit );
}

TabdataReader::~TabdataReader()
{
	delete m_pGrid;
	delete m_pCoordSys;
}

void
TabdataReader::openFile()
{
  m_infile.open(m_filename.c_str(), ios_base::in);
  if ( m_infile.fail() )
    throw domain_error("Cannot open data file " + m_filename);
}

void
TabdataReader::closeFile()
{
  if ( m_infile.is_open() )
    m_infile.close();
}

void
TabdataReader::parseFile()
{
	if ( !m_infile.is_open() )
		throw domain_error("Data file " + m_filename + " not open");

	try {
		parseHeader();
		parseGrid();
		parseData();
	} catch ( ifstream::failure& e ) {
		if ( m_infile.eof() )
			throw domain_error("Data file " + m_filename + " does not contain enough data");
		throw domain_error("Error reading data file " + m_filename);
	} catch ( exception& e ) {
		throw domain_error("Error in data file " + m_filename +": " + e.what() );
	}

}

void
TabdataReader::parseHeader()
{
	if ( !m_headerParsed ) {
		m_infile >> m_dataName;            // density, temperature, etc.

		m_infile >> m_iRegularGrid;        // 0: Regular; 1: Irregular
		if ( m_iRegularGrid < 0 || m_iRegularGrid > 1 )
			throw domain_error("Invalid value for regular/irregular grid");

		m_infile >> m_iGridType;           // Regular Grids: 0 Cartesian; 1 Cylindrical; 2 Spherical
						                   // Irregular Grids: 0 ARM; 1 Delaunay; 2 Voronoi
		if ( m_iRegularGrid == 0 ) {
			// Regular Grid
			if ( m_iGridType == 0 )
				m_pCoordSys = dynamic_cast<const CoordSystem*>(new CoordSystemCart());
			else if ( m_iGridType == 1 )
				m_pCoordSys = dynamic_cast<const CoordSystem*>(new CoordSystemCyl());
			else if ( m_iGridType == 2 )
				m_pCoordSys = dynamic_cast<const CoordSystem*>(new CoordSystemSphe());
			else
				throw domain_error("Invalid value for grid type");
		}

		m_infile >> m_iGridLocation;       // 0 Cell centered; 1 Cell interfaces
		if ( m_iGridLocation < 0 || m_iGridLocation > 1 )
			throw domain_error("Invalid value for cell/interface grid");

		if (m_iRegularGrid==0) {
			// Regular grid:
			m_nGridPoints = vector<int>(3);
			m_infile >> m_nGridPoints[0] >> m_nGridPoints[1] >> m_nGridPoints[2];
			if ( m_nGridPoints[0] < 0 || m_nGridPoints[1] < 0 || m_nGridPoints[2] < 0 )
				throw domain_error("Invalid number of grid points");

		} else {
			// Irregular grid:
			m_nGridPoints = vector<int>(1);
			m_infile >> m_nGridPoints[0];
			if ( m_nGridPoints[0] < 0 )
				throw domain_error("Invalid number of grid points");
		}
		m_headerParsed = true;

		// <RS_DEBUG>
		/*
		cout << m_dataName << endl;            // density, temperature, etc.
		cout << m_iRegularGrid << endl;        // 0: Regular; 1: Irregular
		cout << m_iGridType << endl;           // Regular Grids: 0 Cartesian; 1 Cylindrical; 2 Spherical
						                       // Irregular Grids: 0 ARM; 1 Delaunay; 2 Voronoi
		cout << m_iGridLocation << endl;       // 0 Cell centered; 1 Cell interfaces
		if (m_iRegularGrid==0) {
			// Regular grid:
			cout << m_nGridPoints[0] << " " <<
					m_nGridPoints[1] << " " <<
					m_nGridPoints[2] << endl;
		} else {
			// Regular grid:
			cout << m_nGridPoints[0] << endl;
		}
		*/
		// </RS_DEBUG>
	}
}

void
TabdataReader::parseGrid()
{
	if ( !m_gridParsed ) {
		if ( m_iRegularGrid==0 ) {
			// Regular grid:
			ArtistRegularGrid* pGrid = new ArtistRegularGrid(m_nGridPoints, m_iGridType, m_iGridLocation);

			double x;
			for ( int iDim=0; iDim<3; iDim++ ) {
				for ( int iPoint=0; iPoint<m_nGridPoints[iDim]; iPoint++) {
					m_infile >> x;
					pGrid->addGridPoint(x, iDim);
				}
			}

			pGrid->finalizeGrid();
			m_pGrid = dynamic_cast<const ArtistGrid*>(pGrid);

		}  else {
			// Irregular grid:
			throw domain_error("Irregular grids not yet implemented");
		}
		m_gridParsed = true;
	}
}

const string
TabdataReader::getDataname() const
{
	return m_dataName;
}

const TabdataIndex
TabdataReader::getIndex(const double& x, const double& y, const double& z) const
{
	// Convert to system coordinates
	vector<double> coordSys = m_pCoordSys->coordCartToSys(x, y, z);

	int gridIndex = m_pGrid->getIndex(coordSys);
	if ( gridIndex < 0 )
		return TabdataIndex(coordSys, 0, false);
	return TabdataIndex(coordSys, gridIndex, true);
}


ScalarTabdataReader::ScalarTabdataReader(const string& filename)
: TabdataReader(filename),
  m_pData(NULL)
{}

ScalarTabdataReader::~ScalarTabdataReader()
{
	delete m_pData;
}


void
ScalarTabdataReader::parseData()
{
	if ( !m_dataParsed ) {
		int nDataPoints = m_pGrid->getMaxIndex() + 1;
		if ( nDataPoints > 0 ) {
			m_pData = new ArtistScalarTabdata(nDataPoints);

			double data;
			for ( int iPoint=0; iPoint<nDataPoints; iPoint++) {
				m_infile >> data;
				m_pData->addData(data);
			}

			// <RS_DEBUG>
			/*
			cout << "Data: " << nDataPoints << endl;
			for ( int iPoint=0; iPoint<nDataPoints; iPoint++)
				cout << m_pData->getData(iPoint) << endl;
			*/
			// </RS_DEBUG>
		}
		m_dataParsed = true;
	}
}





const double&
ScalarTabdataReader::getData(const TabdataIndex& tdIndex) const
{
	return m_pData->getData(tdIndex.index);
}



VectorTabdataReader::VectorTabdataReader(const string& filename)
: TabdataReader(filename),
  m_pData(NULL)
{}

VectorTabdataReader::~VectorTabdataReader()
{
	delete m_pData;
}


void
VectorTabdataReader::parseData()
{
	if ( !m_dataParsed ) {
		int nDataPoints = m_pGrid->getMaxIndex() + 1;
		if ( nDataPoints > 0 ) {
			m_pData = new ArtistVectorTabdata(nDataPoints);

			vector<double> data(3);
			for ( int iPoint=0; iPoint<nDataPoints; iPoint++) {
				for( int iDim=0; iDim<3; iDim++ )
					m_infile >> data[iDim];
				m_pData->addData(data);
			}

			// <RS_DEBUG>
			/*
			cout << "Data: " << nDataPoints << endl;
			for ( int iPoint=0; iPoint<nDataPoints; iPoint++) {
				const vector<double>& data = m_pData->getData(iPoint);
				for ( int iDim=0; iDim<3; iDim++ )
					cout << data[iDim] << " ";
				cout << endl;
			}
			*/
			// </RS_DEBUG>
		}
		m_dataParsed = true;
	}
}

const vector<double>
VectorTabdataReader::getData(const TabdataIndex& tdIndex) const
{
	if ( !m_dataParsed )
		throw domain_error("Data was not parsed");

	const vector <double>& sysData = m_pData->getData(tdIndex.index);

	// Transformation of vector component
	const vector<double>& cartData = m_pCoordSys->vecSysToCart(sysData, tdIndex.coords);


	// <RS_DEBUG>
	/*
	cout << "sysData: " << endl;
	for ( int iDim=0; iDim<3; iDim++ )
		cout << sysData[iDim] << " ";
	cout << endl;

	cout << "sysCoords: " << endl;
	for ( int iDim=0; iDim<3; iDim++ )
		cout << tdIndex.coords[iDim] << " ";
	cout << endl;

	cout << "cartData: " << endl;
	for ( int iDim=0; iDim<3; iDim++ )
		cout << cartData[iDim] << " ";
	cout << endl;
	*/
	// </RS_DEBUG>

	return cartData;
}
