////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <stdexcept>

#include "functionScalar.h"
#include "iProvider.h"

#include <cmath>

using namespace std;


const string
FunctionScalar::getType()
{
  return providerTypeScalar;
}

const double
FunctionScalar::density(const double x, const double y, const double z)
{
  return value(x, y, z);
}

const double
FunctionScalar::temperature(const double x, const double y, const double z)
{
  return value(x, y, z);
}

const double
FunctionScalar::abundance(const double x, const double y, const double z)
{
  return value(x, y, z);
}

const double
FunctionScalar::t_dust(const double x, const double y, const double z)
{
  return value(x, y, z);
}

const double
FunctionScalar::doppler(const double x, const double y, const double z)
{
  return value(x, y, z);
}

//////////////////////////////////////////////////////////////////////////////
//
// FS_Const
//
// Constant scalar value
//
// Parameters:
// "val" (double): Constant scalar value

string
FS_Const::functionID = "scalarConst";

const FunctionReg
FS_Const::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Const Scalar Function";
  functionReg.description = "Constant scalar value (double)";
  return functionReg;
}

const set<string>
FS_Const::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("val");
  return paramIDs;
}

const ParamReg*
FS_Const::getParamReg(string paramID)
{
  if (paramID == "val")
      return new ParamRegDouble("val",
                                "Constant scalar value",
                                0.,
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FS_Const::onFinalizeConfiguration()
{
  m_value = m_paramDouble["val"];
}

const double
FS_Const::value(const double x, const double y, const double z)
{
  return m_value;
}


//////////////////////////////////////////////////////////////////////////////
//
// FS_pR
//
// Scalar with power-law dependency on spherical radius r
// value = factor * r**exponent + offset (if r >= lowerR)
//       = factor * lowerR**exponent + offset (if r < lowerR)
//
// Parameters:
// "factor" (double): Factor of power-law
// "exponent" (double): Exponent of power-law
// "offset" (double): Offset of power-law
// "lowerR" (double): Lowest r for which power-law applies

string
FS_pR::functionID = "scalarPowerR";

const FunctionReg
FS_pR::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Powerlaw R Scalar Function";
  functionReg.description = "Scalar with power-law in spherical radius r (double)\n";
  functionReg.description += "value = factor * r**exponent + offset (if r >= lowerR)\n";
  functionReg.description += "      = factor * lowerR**exponent + offset (if r < lowerR)";
  return functionReg;
}

const set<string>
FS_pR::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("factor");
  paramIDs.insert("exponent");
  paramIDs.insert("offset");
  paramIDs.insert("lowerR");
  return paramIDs;
}

const ParamReg*
FS_pR::getParamReg(string paramID)
{
  if (paramID == "factor")
      return new ParamRegDouble("factor",
                                "Factor of power-law",
                                0.,
                                "");
  if (paramID == "exponent")
      return new ParamRegDouble("exponent",
                                "Exponent of power-law",
                                1.,
                                "");
  if (paramID == "offset")
      return new ParamRegDouble("offset",
                                "Offset of power-law",
                                0.,
                                "");
  if (paramID == "lowerR")
      return new ParamRegDouble("lowerR",
                                "Lower r value of power-law",
                                0.,
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FS_pR::onFinalizeConfiguration()
{
  m_factor = m_paramDouble["factor"];
  m_exponent = m_paramDouble["exponent"];
  m_offset = m_paramDouble["offset"];
  m_lowerR = m_paramDouble["lowerR"];
}

const double
FS_pR::value(const double x, const double y, const double z)
{
  double r;
  r = sqrt(x*x + y*y + z*z);
  if ( r >= m_lowerR )
	  return m_factor * pow(r, m_exponent) + m_offset;
  else
	  return m_factor * pow(m_lowerR, m_exponent) + m_offset;
}


//////////////////////////////////////////////////////////////////////////////
//
// FS_pRZ
//
// Scalar with power-law dependency on spherical radius r and z
// value = valR * valZ with
// valR = factR * r**expR + offsetR (if r >= lowerR)
//      = factR * lowerR**expR + offsetR (if r < lowerR)
// valZ = factZ * abs(z)**expZ + offsetZ (if abs(z) >= lowerZ)
//      = factZ * abs(lowerZ)**expZ + offsetZ (if abs(z) < lowerZ)
//
// Parameters:
// "factR" (double): Factor of power-law in r
// "expR" (double): Exponent of power-law in r
// "offsetR" (double): Offset of power-law in r
// "lowerR" (double): Lowest r for which power-law applies
// "factZ" (double): Factor of power-law in z
// "expZ" (double): Exponent of power-law in z
// "offsetZ" (double): Offset of power-law in z
// "lowerZ" (double): Lowest z for which power-law applies

string
FS_pRZ::functionID = "scalarPowerRZ";

const FunctionReg
FS_pRZ::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Powerlaw r z Scalar Function";
  functionReg.description = "Scalar with power-law in spherical radius r and z (double)\n";
  functionReg.description += "value = valR * valZ with\n";
  functionReg.description += "valR = factR * r**expR + offsetR (if r >= lowerR)\n";
  functionReg.description += "     = factR * lowerR**expR + offsetR (if r < lowerR)\n";
  functionReg.description += "valZ = factZ * abs(z)**expZ + offsetZ (if abs(z) >= lowerZ)\n";
  functionReg.description += "     = factZ * abs(lowerZ)**expZ + offsetZ (if abs(z) < lowerZ)";
  return functionReg;
}

const set<string>
FS_pRZ::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("factR");
  paramIDs.insert("expR");
  paramIDs.insert("offsetR");
  paramIDs.insert("lowerR");
  paramIDs.insert("factZ");
  paramIDs.insert("expZ");
  paramIDs.insert("offsetZ");
  paramIDs.insert("lowerZ");
  return paramIDs;
}

const ParamReg*
FS_pRZ::getParamReg(string paramID)
{
  if (paramID == "factR")
      return new ParamRegDouble("factR",
                                "Factor of power-law in r",
                                0.,
                                "");
  if (paramID == "expR")
      return new ParamRegDouble("expR",
                                "Exponent of power-law in r",
                                1.,
                                "");
  if (paramID == "offsetR")
      return new ParamRegDouble("offsetR",
                                "Offset of power-law in r",
                                0.,
                                "");
  if (paramID == "lowerR")
      return new ParamRegDouble("lowerR",
                                "Lower r value of power-law",
                                0.,
                                "");
  if (paramID == "factZ")
      return new ParamRegDouble("factZ",
                                "Factor of power-law in z",
                                0.,
                                "");
  if (paramID == "expZ")
      return new ParamRegDouble("expZ",
                                "Exponent of power-law in z",
                                -1.,
                                "");
  if (paramID == "offsetZ")
      return new ParamRegDouble("offsetZ",
                                "Offset of power-law in z",
                                0.,
                                "");
  if (paramID == "lowerZ")
      return new ParamRegDouble("lowerZ",
                                "Lower z value of power-law",
                                0.,
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FS_pRZ::onFinalizeConfiguration()
{
	m_factR = m_paramDouble["factR"];
	m_expR = m_paramDouble["expR"];
	m_offsetR = m_paramDouble["offsetR"];
	m_lowerR = m_paramDouble["lowerR"];
	m_factZ = m_paramDouble["factZ"];
	m_expZ = m_paramDouble["expZ"];
	m_offsetZ = m_paramDouble["offsetZ"];
	m_lowerZ = m_paramDouble["lowerZ"];
}

const double
FS_pRZ::value(const double x, const double y, const double z)
{
  double r;
  r = sqrt(x*x + y*y + z*z);

  double valR;
  if ( r >= m_lowerR )
	  valR = m_factR * pow(r, m_expR) + m_offsetR;
  else
	  valR = m_factR * pow(m_lowerR, m_expR) + m_offsetR;

  double valZ;
  if ( abs(z) >= m_lowerZ )
	  valZ = m_factZ * pow(abs(z), m_expZ) + m_offsetZ;
  else
	  valZ = m_factZ * pow(abs(m_lowerZ), m_expZ) + m_offsetZ;

  return valR * valZ;
}


//////////////////////////////////////////////////////////////////////////////
//
// FS_pReZ
//
// Scalar with power-law dependency on spherical radius r and exponential in z
// value = valR * valZ with
// valR = factR * r**expR + offsetR (if r >= lowerR)
//      = factR * lowerR**expR + offsetR (if r < lowerR)
// valZ = factZ * e**(-abs(z) / scaleZ) + offsetZ
//
// Parameters:
// "factR" (double): Factor of power-law in r
// "expR" (double): Exponent of power-law in r
// "offsetR" (double): Offset of power-law in r
// "lowerR" (double): Lowest r for which power-law applies
// "factZ" (double): Factor of exponential in z
// "scaleZ" (double): Scale of exponential in z
// "offsetZ" (double): Offset of exponential in z

string
FS_pReZ::functionID = "scalarPowerRExpZ";

const FunctionReg
FS_pReZ::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Powerlaw r exponential z Scalar Function";
  functionReg.description = "Scalar with power-law in spherical radius r and exponential in z (double)\n";

  functionReg.description += "value = valR * valZ with\n";
  functionReg.description += "valR = factR * r**expR + offsetR (if r >= lowerR)\n";
  functionReg.description += "     = factR * lowerR**expR + offsetR (if r < lowerR)\n";
  functionReg.description += "valZ = factZ * e**(-abs(z) / scaleZ) + offsetZ";
  return functionReg;
}

const set<string>
FS_pReZ::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("factR");
  paramIDs.insert("expR");
  paramIDs.insert("offsetR");
  paramIDs.insert("lowerR");
  paramIDs.insert("factZ");
  paramIDs.insert("scaleZ");
  paramIDs.insert("offsetZ");
  return paramIDs;
}

const ParamReg*
FS_pReZ::getParamReg(string paramID)
{
  if (paramID == "factR")
      return new ParamRegDouble("factR",
                                "Factor of power-law in r",
                                0.,
                                "");
  if (paramID == "expR")
      return new ParamRegDouble("expR",
                                "Exponent of power-law in r",
                                1.,
                                "");
  if (paramID == "offsetR")
      return new ParamRegDouble("offsetR",
                                "Offset of power-law in r",
                                0.,
                                "");
  if (paramID == "lowerR")
      return new ParamRegDouble("lowerR",
                                "Lower r value of power-law",
                                0.,
                                "");
  if (paramID == "factZ")
      return new ParamRegDouble("factZ",
                                "Factor of exponential in z",
                                0.,
                                "");
  if (paramID == "scaleZ")
      return new ParamRegDouble("expZ",
                                "Scale length of exponential in z",
                                1.,
                                "");
  if (paramID == "offsetZ")
      return new ParamRegDouble("offsetZ",
                                "Offset of exponential in z",
                                0.,
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FS_pReZ::onFinalizeConfiguration()
{
	m_factR = m_paramDouble["factR"];
	m_expR = m_paramDouble["expR"];
	m_offsetR = m_paramDouble["offsetR"];
	m_lowerR = m_paramDouble["lowerR"];
	m_factZ = m_paramDouble["factZ"];
	m_scaleZ = m_paramDouble["scaleZ"];
	m_offsetZ = m_paramDouble["offsetZ"];
}

const double
FS_pReZ::value(const double x, const double y, const double z)
{
  double r;
  r = sqrt(x*x + y*y + z*z);

  double valR;
  if ( r >= m_lowerR )
	  valR = m_factR * pow(r, m_expR) + m_offsetR;
  else
	  valR = m_factR * pow(m_lowerR, m_expR) + m_offsetR;

  double valZ;
  valZ = m_factZ * exp(-abs(z) / m_scaleZ) + m_offsetZ;

  return valR * valZ;
}

//////////////////////////////////////////////////////////////////////////////
//
// FS_pRT
//
// Scalar with power-law dependency on spherical radius r and theta
// value = valR * valTheta with
// valR = factR * r**expR + offsetR (if r >= lowerR)
//      = factR * lowerR**expR + offsetR (if r < lowerR)
// valTheta = factTheta * theta**expTheta + offsetTheta (if theta >= lowerTheta)
//          = factTheta * lowerTheta**expTheta + offsetTheta (if theta < lowerTheta)
// and with theta = acos( abs(z) / r
//
// Parameters:
// "factR" (double): Factor of power-law in r
// "expR" (double): Exponent of power-law in r
// "offsetR" (double): Offset of power-law in r
// "lowerR" (double): Lowest r for which power-law applies
// "factTheta" (double): Factor of power-law in theta
// "expTheta" (double): Exponent of power-law in theta
// "offsetTheta" (double): Offset of power-law in theta
// "lowerTheta" (double): Lowest theta for which power-law applies

string
FS_pRT::functionID = "scalarPowerRTheta";

const FunctionReg
FS_pRT::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Powerlaw r theta Scalar Function";
  functionReg.description = "Scalar with power-law in spherical radius r and theta (double)\n";
  functionReg.description += "value = valR * valTheta with\n";
  functionReg.description += "valR = factR * r**expR + offsetR (if r >= lowerR)\n";
  functionReg.description += "     = factR * lowerR**expR + offsetR (if r < lowerR)\n";
  functionReg.description += "valTheta = factTheta * theta**expTheta + offsetTheta (if theta >= lowerTheta)\n";
  functionReg.description += "         =  factTheta * lowerTheta**expTheta + offsetTheta (if theta < lowerTheta)\n";
  functionReg.description += "and with theta = acos( abs(z) / r )";
  return functionReg;
}

const set<string>
FS_pRT::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("factR");
  paramIDs.insert("expR");
  paramIDs.insert("offsetR");
  paramIDs.insert("lowerR");
  paramIDs.insert("factTheta");
  paramIDs.insert("expTheta");
  paramIDs.insert("offsetTheta");
  paramIDs.insert("lowerTheta");
  return paramIDs;
}

const ParamReg*
FS_pRT::getParamReg(string paramID)
{
  if (paramID == "factR")
      return new ParamRegDouble("factR",
                                "Factor of power-law in r",
                                0.,
                                "");
  if (paramID == "expR")
      return new ParamRegDouble("expR",
                                "Exponent of power-law in r",
                                1.,
                                "");
  if (paramID == "offsetR")
      return new ParamRegDouble("offsetR",
                                "Offset of power-law in r",
                                0.,
                                "");
  if (paramID == "lowerR")
      return new ParamRegDouble("lowerR",
                                "Lower r value of power-law",
                                0.,
                                "");
  if (paramID == "factTheta")
      return new ParamRegDouble("factTheta",
                                "Factor of power-law in theta",
                                0.,
                                "");
  if (paramID == "expTheta")
      return new ParamRegDouble("expTheta",
                                "Exponent of power-law in theta",
                                -1.,
                                "");
  if (paramID == "offsetTheta")
      return new ParamRegDouble("offsetTheta",
                                "Offset of power-law in theta",
                                0.,
                                "");
  if (paramID == "lowerTheta")
      return new ParamRegDouble("lowerTheta",
                                "Lower theta value of power-law",
                                0.,
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FS_pRT::onFinalizeConfiguration()
{
	m_factR = m_paramDouble["factR"];
	m_expR = m_paramDouble["expR"];
	m_offsetR = m_paramDouble["offsetR"];
	m_lowerR = m_paramDouble["lowerR"];
	m_factTheta = m_paramDouble["factTheta"];
	m_expTheta = m_paramDouble["expTheta"];
	m_offsetTheta = m_paramDouble["offsetTheta"];
	m_lowerTheta = m_paramDouble["lowerTheta"];
}

const double
FS_pRT::value(const double x, const double y, const double z)
{
	double r, theta;
	r = sqrt(x*x + y*y + z*z);
	if ( r == 0. )
		theta = 0.;
	else
		theta = acos( abs(z) / r);

	double valR;
	if ( r >= m_lowerR )
		valR = m_factR * pow(r, m_expR) + m_offsetR;
	else
		valR = m_factR * pow(m_lowerR, m_expR) + m_offsetR;

	double valTheta;
	if ( theta >= m_lowerTheta )
		valTheta = m_factTheta * pow(abs(theta), m_expTheta) + m_offsetTheta;
	else
		valTheta = m_factTheta * pow(abs(m_lowerTheta), m_expTheta) + m_offsetTheta;

	return valR * valTheta;
}

