////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <string>
#include <stdexcept>

#include <math.h>

#include "functionTab.h"
#include "artistTabdataReader.h"
#include "utilities.h"


using namespace std;


//////////////////////////////////////////////////////////////////////////////
//
// FS_Tabdata
//
// Scalar data from ARTIST Tabdata file
//
// Parameters:
// "filename" (string): Full path to the Tabdata file
// "defValue" (double): Default data value outside grid defined in Tabdata file

string
FS_Tabdata::functionID = "scalarTabdata";

FS_Tabdata::FS_Tabdata()
: m_pReader(NULL)
{}

FS_Tabdata::~FS_Tabdata()
{
  delete m_pReader;
}

const FunctionReg
FS_Tabdata::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Scalar Tabdata Function";
  functionReg.description = "Scalar data from ARTISTS Tabdata file";
  return functionReg;

}

const set<string>
FS_Tabdata::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("filename");
  paramIDs.insert("defValue");
  return paramIDs;
}

const ParamReg*
FS_Tabdata::getParamReg(string paramID)
{
  if (paramID == "filename")
      return new ParamRegString("filename",
                                "Full path to the Tabdata file",
                                "");
  if (paramID == "defValue")
      return new ParamRegDouble("defValue",
                                "Default data value outside grid defined in Tabdata file",
                                0,
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FS_Tabdata::onFinalizeConfiguration()
{
	string& filename = m_paramString["filename"];
	if ( strip(filename).size() ) {
		m_pReader = new ScalarTabdataReader(filename);
		m_pReader->openFile();
		m_pReader->parseFile();
		m_pReader->closeFile();
	}
	m_defValue = m_paramDouble["defValue"];
}

const double
FS_Tabdata::value(const double x, const double y, const double z)
{
	if ( !m_pReader )
		throw domain_error("No Data read");

	TabdataIndex index = m_pReader->getIndex(x, y, z);
	if ( !index.isValid )
		return m_defValue;

	return m_pReader->getData(index);
}

//////////////////////////////////////////////////////////////////////////////
//
// FV_Tabdata
//
// Vector data from ARTIST Tabdata file
// Default data value outside grid is (0, 0, 0)
//
// Parameters:
// "filename" (string): Full path to the Tabdata file

string
FV_Tabdata::functionID = "vectorTabdata";

FV_Tabdata::FV_Tabdata()
: m_pReader(NULL)
{}

FV_Tabdata::~FV_Tabdata()
{
  delete m_pReader;
}

const FunctionReg
FV_Tabdata::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Vector Tabdata Function";
  functionReg.description = "Vector data from ARTIST Tabdata file\n";
  functionReg.description += "Default data value outside grid is (0, 0, 0)";
  return functionReg;

}

const set<string>
FV_Tabdata::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("filename");
  return paramIDs;
}

const ParamReg*
FV_Tabdata::getParamReg(string paramID)
{
  if (paramID == "filename")
      return new ParamRegString("filename",
                                "Full path to the Tabdata file",
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FV_Tabdata::onFinalizeConfiguration()
{
	string& filename = m_paramString["filename"];
	if ( strip(filename).size() ) {
		m_pReader = new VectorTabdataReader(filename);
		m_pReader->openFile();
		m_pReader->parseFile();
		m_pReader->closeFile();
	}
}

const void
FV_Tabdata::value(const double x, const double y, const double z, double* val)
{
	TabdataIndex index = m_pReader->getIndex(x, y, z);
	if ( !index.isValid ) {
		val[0] = 0.;
		val[1] = 0.;
		val[2] = 0.;
	} else {
		const vector<double>& value = m_pReader->getData(index);
		val[0] = value[0];
		val[1] = value[1];
		val[2] = value[2];
	}
}

