////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <stdexcept>

#include "functionVector.h"
#include "iProvider.h"

#include <cmath>

using namespace std;


const string
FunctionVector::getType()
{
  return providerTypeVector;
}

const void
FunctionVector::velocity(const double x, const double y, const double z, double* v)
{
  value(x, y, z, v);
}

const void
FunctionVector::bmag(const double x, const double y, const double z, double* b)
{
  value(x, y, z, b);
}


//////////////////////////////////////////////////////////////////////////////
//
// FV_rad_const
//
// Constant radial vector
//
// Parameters:
// "val" (double): Constant value of vector

string
FV_rad_const::functionID = "vectorConstR";

const FunctionReg
FV_rad_const::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Const radial vector Function";
  functionReg.description = "Constant radial vector (double)";
  return functionReg;
}

const set<string>
FV_rad_const::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("val");
  return paramIDs;
}

const ParamReg*
FV_rad_const::getParamReg(string paramID)
{
  if (paramID == "val")
      return new ParamRegDouble("val",
                                "Constant value of vector",
                                0.,
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FV_rad_const::onFinalizeConfiguration()
{
  m_val = m_paramDouble["val"];
}

const void
FV_rad_const::value(const double x, const double y, const double z, double* val)
{
	double r, phi, theta;

	r = sqrt(x*x + y*y + z*z);
	if ( r == 0. ) {
		val[0] = 0.;
		val[1] = 0.;
		val[2] = 0.;
	} else {
		double rho = sqrt(x*x + y*y);
		theta = atan2(rho, z);
		if ( rho == 0. )
			phi = 0.;
		else
			phi = atan2(y, x);

		val[0] = m_val * sin(theta) * cos(phi);
		val[1] = m_val * sin(theta) * sin(phi);
		val[2] = m_val * cos(theta);
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// FV_xyz_const
//
// Constant vector with x, y, and z component
//
// Parameters:
// "valX" (double): Constant value of x component
// "valY" (double): Constant value of y component
// "valZ" (double): Constant value of z component

string
FV_xyz_const::functionID = "vectorConstXYZ";

const FunctionReg
FV_xyz_const::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Const cartesian vector Function";
  functionReg.description = "Constant vector with x, y, and z component (double)";
  return functionReg;
}

const set<string>
FV_xyz_const::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("valX");
  paramIDs.insert("valY");
  paramIDs.insert("valZ");
  return paramIDs;
}

const ParamReg*
FV_xyz_const::getParamReg(string paramID)
{
  if (paramID == "valX")
      return new ParamRegDouble("valX",
                                "Constant x component of vector",
                                0.,
                                "");
  if (paramID == "valY")
      return new ParamRegDouble("valY",
                                "Constant y component of vector",
                                0.,
                                "");
  if (paramID == "valZ")
      return new ParamRegDouble("valZ",
                                "Constant z component of vector",
                                0.,
                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FV_xyz_const::onFinalizeConfiguration()
{
	m_valX = m_paramDouble["valX"];
	m_valY = m_paramDouble["valY"];
	m_valZ = m_paramDouble["valZ"];
}

const void
FV_xyz_const::value(const double x, const double y, const double z, double* val)
{
  val[0] = m_valX;
  val[1] = m_valY;
  val[2] = m_valZ;
}

//////////////////////////////////////////////////////////////////////////////
//
// FV_rad_pR
//
// Radial vector with power-law dependency on spherical radius r
// valR = factor * r**exponent + offset (if r >= lowerR)
//      = factor * lowerR**exponent + offset (if r < lowerR)
//
// Parameters:
// "factor" (double): Factor of power-law
// "exponent" (double): Exponent of power-law
// "offset" (double): Offset of power-law
// "lowerR" (double): Lowest r for which power-law applies

string
FV_rad_pR::functionID = "vectorRadialPowerR";

const FunctionReg
FV_rad_pR::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Powerlaw R radial vector Function";
  functionReg.description = "Radial vector with power-law in spherical radius r (double)\n";
  functionReg.description += "valR = factor * r**exponent + offset (if r >= lowerR)\n";
  functionReg.description += "     = factor * lowerR**exponent + offset (if r < lowerR)";
  return functionReg;
}

const set<string>
FV_rad_pR::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("factor");
  paramIDs.insert("exponent");
  paramIDs.insert("offset");
  paramIDs.insert("lowerR");
  return paramIDs;
}

const ParamReg*
FV_rad_pR::getParamReg(string paramID)
{
	  if (paramID == "factor")
	      return new ParamRegDouble("factor",
	                                "Factor of power-law",
	                                0.,
	                                "");
	  if (paramID == "exponent")
	      return new ParamRegDouble("exponent",
	                                "Exponent of power-law",
	                                1.,
	                                "");
	  if (paramID == "offset")
	      return new ParamRegDouble("offset",
	                                "Offset of power-law",
	                                0.,
	                                "");
	  if (paramID == "lowerR")
	      return new ParamRegDouble("lowerR",
	                                "Lower r value of power-law",
	                                0.,
	                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FV_rad_pR::onFinalizeConfiguration()
{
	  m_factor = m_paramDouble["factor"];
	  m_exponent = m_paramDouble["exponent"];
	  m_offset = m_paramDouble["offset"];
	  m_lowerR = m_paramDouble["lowerR"];
}

const void
FV_rad_pR::value(const double x, const double y, const double z, double* val)
{
	double r;

	r = sqrt(x*x + y*y + z*z);
	if ( r == 0. ) {
		val[0] = 0.;
		val[1] = 0.;
		val[2] = 0.;
	} else {
		double rho, phi, theta;
		double vr;

		rho = sqrt(x*x + y*y);
		theta = atan2(rho, z);
		if ( rho == 0. )
			phi = 0.;
		else
			phi = atan2(y, x);

		if ( r >= m_lowerR )
			vr = m_factor * pow(r, m_exponent) + m_offset;
		else
			vr = m_factor * pow(m_lowerR, m_exponent) + m_offset;
		val[0] = vr * sin(theta) * cos(phi);
		val[1] = vr * sin(theta) * sin(phi);
		val[2] = vr * cos(theta);
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// FV_tor_pRho
//
// Toroidal vector with power-law dependency on cylindrical radius rho
// valPhi = factor * r**exponent + offset (if r >= lowerR)
//        = factor * lowerR**exponent + offset (if r < lowerR)
//
// Parameters:
// "factor" (double): Factor of power-law
// "exponent" (double): Exponent of power-law
// "offset" (double): Offset of power-law
// "lowerRho" (double): Lowest rho for which power-law applies

string
FV_tor_pRho::functionID = "vectorToroidalPowerR";

const FunctionReg
FV_tor_pRho::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Powerlaw R toroidal vector Function";
  functionReg.description = "Toroidal vector with power-law in cylindrical radius rho (double)\n";
  functionReg.description += "valPhi = factor * r**exponent + offset (if rho >= lowerRho)\n";
  functionReg.description += "       = factor * lowerRho**exponent + offset (if rho < lowerRho)";
  return functionReg;
}

const set<string>
FV_tor_pRho::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("factor");
  paramIDs.insert("exponent");
  paramIDs.insert("offset");
  paramIDs.insert("lowerRho");
  return paramIDs;
}

const ParamReg*
FV_tor_pRho::getParamReg(string paramID)
{
	  if (paramID == "factor")
	      return new ParamRegDouble("factor",
	                                "Factor of power-law",
	                                0.,
	                                "");
	  if (paramID == "exponent")
	      return new ParamRegDouble("exponent",
	                                "Exponent of power-law",
	                                1.,
	                                "");
	  if (paramID == "offset")
	      return new ParamRegDouble("offset",
	                                "Offset of power-law",
	                                0.,
	                                "");
	  if (paramID == "lowerRho")
	      return new ParamRegDouble("lowerRho",
	                                "Lower rho value of power-law",
	                                0.,
	                                "");
  throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FV_tor_pRho::onFinalizeConfiguration()
{
	  m_factor = m_paramDouble["factor"];
	  m_exponent = m_paramDouble["exponent"];
	  m_offset = m_paramDouble["offset"];
	  m_lowerRho = m_paramDouble["lowerR"];
}

const void
FV_tor_pRho::value(const double x, const double y, const double z, double* val)
{
	double rho;

	rho = sqrt(x*x + y*y);
	if ( rho == 0. ) {
		val[0] = 0.;
		val[1] = 0.;
		val[2] = 0.;

	} else {
		double phi;
		double vphi;

		phi = atan2(y, x);
		if ( rho >= m_lowerRho )
			vphi = m_factor * pow(rho, m_exponent) + m_offset;
		else
			vphi = m_factor * pow(m_lowerRho, m_exponent) + m_offset;

		val[0] = - vphi * sin(phi);
		val[1] =   vphi * cos(phi);
		val[2] = 0.;
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// FV_rad_pRT
//
// Radial vector with power-law dependency on spherical radius r and theta
// value = valR * valTheta with
// valR = factR * r**expR + offsetR (if r >= lowerR)
//      = factR * lowerR**expR + offsetR (if r < lowerR)
// valTheta = factTheta * theta**expTheta + offsetTheta (if theta >= lowerTheta)
//          = factTheta * lowerTheta**expTheta + offsetTheta (if theta < lowerTheta)
// and with theta = acos( abs(z) / r
//
// Parameters:
// "factR" (double): Factor of power-law in r
// "expR" (double): Exponent of power-law in r
// "offsetR" (double): Offset of power-law in r
// "lowerR" (double): Lowest r for which power-law applies
// "factTheta" (double): Factor of power-law in theta
// "expTheta" (double): Exponent of power-law in theta
// "offsetTheta" (double): Offset of power-law in theta
// "lowerTheta" (double): Lowest theta for which power-law applies

string
FV_rad_pRT::functionID = "vectorRadialPowerRTheta";

const FunctionReg
FV_rad_pRT::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Powerlaw r theta radial vector Function";
  functionReg.description = "Radial vector with power-law in spherical radius r and theta (double)\n";
  functionReg.description += "value = valR * valTheta with\n";
  functionReg.description += "valR = factR * r**expR + offsetR (if r >= lowerR)\n";
  functionReg.description += "     = factR * lowerR**expR + offsetR (if r < lowerR)\n";
  functionReg.description += "valTheta = factTheta * theta**expTheta + offsetTheta (if theta >= lowerTheta)\n";
  functionReg.description += "         =  factTheta * lowerTheta**expTheta + offsetTheta (if theta < lowerTheta)\n";
  functionReg.description += "and with theta = acos( abs(z) / r )";
  return functionReg;
}

const set<string>
FV_rad_pRT::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("factR");
  paramIDs.insert("expR");
  paramIDs.insert("offsetR");
  paramIDs.insert("lowerR");
  paramIDs.insert("factTheta");
  paramIDs.insert("expTheta");
  paramIDs.insert("offsetTheta");
  paramIDs.insert("lowerTheta");
  return paramIDs;
}

const ParamReg*
FV_rad_pRT::getParamReg(string paramID)
{
	if (paramID == "factR")
		return new ParamRegDouble("factR",
				"Factor of power-law in r",
				0.,
				"");
	if (paramID == "expR")
		return new ParamRegDouble("expR",
				"Exponent of power-law in r",
				1.,
				"");
	if (paramID == "offsetR")
		return new ParamRegDouble("offsetR",
				"Offset of power-law in r",
				0.,
				"");
	if (paramID == "lowerR")
		return new ParamRegDouble("lowerR",
				"Lower r value of power-law",
				0.,
				"");
	if (paramID == "factTheta")
		return new ParamRegDouble("factTheta",
				"Factor of power-law in theta",
				0.,
				"");
	if (paramID == "expTheta")
		return new ParamRegDouble("expTheta",
				"Exponent of power-law in theta",
				-1.,
				"");
	if (paramID == "offsetTheta")
		return new ParamRegDouble("offsetTheta",
				"Offset of power-law in theta",
				0.,
				"");
	if (paramID == "lowerTheta")
		return new ParamRegDouble("lowerTheta",
				"Lower theta value of power-law",
				0.,
				"");
	throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FV_rad_pRT::onFinalizeConfiguration()
{
	m_factR = m_paramDouble["factR"];
	m_expR = m_paramDouble["expR"];
	m_offsetR = m_paramDouble["offsetR"];
	m_lowerR = m_paramDouble["lowerR"];
	m_factTheta = m_paramDouble["factTheta"];
	m_expTheta = m_paramDouble["expTheta"];
	m_offsetTheta = m_paramDouble["offsetTheta"];
	m_lowerTheta = m_paramDouble["lowerTheta"];
}

const void
FV_rad_pRT::value(const double x, const double y, const double z, double* val)
{
	double r;

	r = sqrt(x*x + y*y + z*z);
	if ( r == 0. ) {
		val[0] = 0.;
		val[1] = 0.;
		val[2] = 0.;
	} else {
		double rho, phi, theta;
		double vr;

		rho = sqrt(x*x + y*y);
		theta = atan2(rho, z);
		if ( rho == 0. )
			phi = 0.;
		else
			phi = atan2(y, x);

		double valR;
		if ( r >= m_lowerR )
			valR = m_factR * pow(r, m_expR) + m_offsetR;
		else
			valR = m_factR * pow(m_lowerR, m_expR) + m_offsetR;

		double valTheta;
		double thetaVal = acos( abs(z) / r);
		if ( thetaVal >= m_lowerTheta )
			valTheta = m_factTheta * pow(abs(thetaVal), m_expTheta) + m_offsetTheta;
		else
			valTheta = m_factTheta * pow(abs(m_lowerTheta), m_expTheta) + m_offsetTheta;

		vr = valR * valTheta;

		val[0] = vr * sin(theta) * cos(phi);
		val[1] = vr * sin(theta) * sin(phi);
		val[2] = vr * cos(theta);
	}
}

//////////////////////////////////////////////////////////////////////////////
//
// FV_dipole
//
// Dipole field with dipole moment
//
// valX = 3 * mDipole * sin(theta) * cos(theta) * cos(phi) / r**3
// valY = 3 * mDipole * sin(theta) * cos(theta) * sin(phi) / r**3
// valZ = 3 * mDipole * (cos(theta)**2 - 1/3) / r**3
// if r >= lowerR
//
// valX = 3 * mDipole * sin(theta) * cos(theta) * cos(phi) / lowerR**3
// valY = 3 * mDipole * sin(theta) * cos(theta) * sin(phi) / lowerR**3
// valZ = 3 * mDipole * (cos(theta)**2 - 1/3) / lowerR**3
// if r < lowerR
//
// Parameters:
// "mDipole" (double): Dipole moment
// "lowerR" (double): Lowest r for which dipole applies

string
FV_dipole::functionID = "vectorDipole";

const FunctionReg
FV_dipole::getFunctionReg()
{
  FunctionReg functionReg;
  functionReg.name = "Dipole vector Function";
  functionReg.description = "Dipole field with dipole moment\n\n";
  functionReg.description += "valX = 3 * mDipole * sin(theta) * cos(theta) * cos(phi) / r**3\n";
  functionReg.description += "valY = 3 * mDipole * sin(theta) * cos(theta) * sin(phi) / r**3\n";
  functionReg.description += "valZ = 3 * mDipole * (cos(theta)**2 - 1/3) / r**3\n";
  functionReg.description += "if r >= lowerR\n\n";
  functionReg.description += "valX = 3 * mDipole * sin(theta) * cos(theta) * cos(phi) / lowerR**3\n";
  functionReg.description += "valY = 3 * mDipole * sin(theta) * cos(theta) * sin(phi) / lowerR**3\n";
  functionReg.description += "valZ = 3 * mDipole * (cos(theta)**2 - 1/3) / lowerR**3\n";
  functionReg.description += "if r < lowerR";
  return functionReg;
}

const set<string>
FV_dipole::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("mDipole");
  paramIDs.insert("lowerR");
  return paramIDs;
}

const ParamReg*
FV_dipole::getParamReg(string paramID)
{
	if (paramID == "mDipole")
		return new ParamRegDouble("mDipole",
				"Dipole moment",
				0.,
				"");
	if (paramID == "lowerR")
		return new ParamRegDouble("lowerR",
				"Lower r value of dipole",
				0.,
				"");
	throw invalid_argument("Invalid paramID " + paramID + " for functionID " + functionID);
}

void
FV_dipole::onFinalizeConfiguration()
{
	m_mDipole = m_paramDouble["mDipole"];
	m_lowerR = m_paramDouble["lowerR"];
}

const void
FV_dipole::value(const double x, const double y, const double z, double* val)
{
	double r, rho, phi, theta;

	r = sqrt(x*x + y*y + z*z);
	rho = sqrt(x*x + y*y);

	if ( r == 0. )
		theta = 0.;
	else
		theta = atan2(rho, z);

	if ( rho == 0. )
		phi = 0.;
	else
		phi = atan2(y, x);

	double rCube;
	if ( r >= m_lowerR )
		rCube = r * r * r;
	else
		rCube = m_lowerR * m_lowerR * m_lowerR;

	double sinTheta = sin(theta);
	double cosTheta = cos(theta);

	val[0] = 3. * m_mDipole * sinTheta * cosTheta * cos(phi) / rCube;
	val[1] = 3. * m_mDipole * sinTheta * cosTheta * sin(phi) / rCube;
	val[2] = 3. * m_mDipole * (cosTheta * cosTheta - 1./3.) / rCube;
}


