////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef FUNCTIONVECTOR_H
#define FUNCTIONVECTOR_H

#include "function.h"
#include "functionregistry.h"
#include "iProvider.h"

#include <string>
#include <set>

using namespace std;

class FunctionVector: public Function,
                      public IVelocityProvider,
                      public IBmagProvider
{
public:
  virtual ~FunctionVector() {};

  // Function interface:
  virtual const string getType();

  // IProvider interfaces:
  virtual const void
  velocity(const double x, const double y, const double z, double* v);

  virtual const void
  bmag(const double x, const double y, const double z, double* b);

protected:
  virtual void onFinalizeConfiguration() = 0;
  virtual const void value(const double x, const double y, const double z, double* val) = 0;
};



class FV_rad_const: public FunctionVector
{
public:
  static string functionID;

  virtual ~FV_rad_const() {};

  virtual const FunctionReg getFunctionReg();
  virtual const set<string> getParamIDs();
  virtual const ParamReg* getParamReg(string paramID);

  virtual const void value(const double x, const double y, const double z, double* val);

protected:
  virtual void onFinalizeConfiguration();

private:
  double m_val;
};


class FV_xyz_const: public FunctionVector
{
public:
  static string functionID;

  virtual ~FV_xyz_const() {};

  virtual const FunctionReg getFunctionReg();
  virtual const set<string> getParamIDs();
  virtual const ParamReg* getParamReg(string paramID);

  virtual const void value(const double x, const double y, const double z, double* val);

protected:
  virtual void onFinalizeConfiguration();

private:
  double m_valX;
  double m_valY;
  double m_valZ;
};


class FV_rad_pR: public FunctionVector
{
public:
  static string functionID;

  virtual ~FV_rad_pR() {};

  virtual const FunctionReg getFunctionReg();
  virtual const set<string> getParamIDs();
  virtual const ParamReg* getParamReg(string paramID);

  virtual const void value(const double x, const double y, const double z, double* val);

protected:
  virtual void onFinalizeConfiguration();

private:
  double m_factor;
  double m_exponent;
  double m_offset;
  double m_lowerR;
};

class FV_tor_pRho: public FunctionVector
{
public:
  static string functionID;

  virtual ~FV_tor_pRho() {};

  virtual const FunctionReg getFunctionReg();
  virtual const set<string> getParamIDs();
  virtual const ParamReg* getParamReg(string paramID);

  virtual const void value(const double x, const double y, const double z, double* val);

protected:
  virtual void onFinalizeConfiguration();

private:
  double m_factor;
  double m_exponent;
  double m_offset;
  double m_lowerRho;
};

class FV_rad_pRT: public FunctionVector
{
public:
  static string functionID;

  virtual ~FV_rad_pRT() {};

  virtual const FunctionReg getFunctionReg();
  virtual const set<string> getParamIDs();
  virtual const ParamReg* getParamReg(string paramID);

  virtual const void value(const double x, const double y, const double z, double* val);

protected:
  virtual void onFinalizeConfiguration();

private:
  double m_factTheta;
  double m_expTheta;
  double m_offsetTheta;
  double m_lowerTheta;
  double m_factR;
  double m_expR;
  double m_offsetR;
  double m_lowerR;
};


class FV_dipole: public FunctionVector
{
public:
  static string functionID;

  virtual ~FV_dipole() {};

  virtual const FunctionReg getFunctionReg();
  virtual const set<string> getParamIDs();
  virtual const ParamReg* getParamReg(string paramID);

  virtual const void value(const double x, const double y, const double z, double* val);

protected:
  virtual void onFinalizeConfiguration();

private:
  double m_mDipole;
  double m_lowerR;
};



#endif // FUNCTIONVECTOR_H
