////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include "functionregistry.h"
#include "parameterregistry.h"
#include "providerregistry.h"
#include "function.h"

#include <stdexcept>

using namespace std;

////////////////////////////////////////////////////////////////////
// Edit here when adding a new model:
//
#include "functionScalar.h"
#include "functionVector.h"
#include "functionTab.h"

void
FunctionRegistry::registerFunctions()
{
  registerFunction(FS_Const::functionID);
  registerFunction(FS_pR::functionID);
  registerFunction(FS_pRZ::functionID);
  registerFunction(FS_pReZ::functionID);
  registerFunction(FS_pRT::functionID);

  registerFunction(FV_rad_const::functionID);
  registerFunction(FV_xyz_const::functionID);
  registerFunction(FV_rad_pR::functionID);
  registerFunction(FV_tor_pRho::functionID);
  registerFunction(FV_rad_pRT::functionID);
  registerFunction(FV_dipole::functionID);

  registerFunction(FS_Tabdata::functionID);
  registerFunction(FV_Tabdata::functionID);
}

Function*
FunctionRegistry::createFunction(const string functionID)
const
{
  if (functionID == FS_Const::functionID)
    return new FS_Const();
  if (functionID == FS_pR::functionID)
    return new FS_pR();
  if (functionID == FS_pRZ::functionID)
    return new FS_pRZ();
  if (functionID == FS_pReZ::functionID)
    return new FS_pReZ();
  if (functionID == FS_pRT::functionID)
    return new FS_pRT();

  if (functionID == FV_rad_const::functionID)
	  return new FV_rad_const();
  if (functionID == FV_xyz_const::functionID)
	  return new FV_xyz_const();
  if (functionID == FV_rad_pR::functionID)
	  return new FV_rad_pR();
  if (functionID == FV_tor_pRho::functionID)
	  return new FV_tor_pRho();
  if (functionID == FV_rad_pRT::functionID)
	  return new FV_rad_pRT();
  if (functionID == FV_dipole::functionID)
	  return new FV_dipole();

  if (functionID == FS_Tabdata::functionID)
    return new FS_Tabdata();
  if (functionID == FV_Tabdata::functionID)
    return new FV_Tabdata();

  throw invalid_argument("Cannot create Function with functionID " + functionID);
}
//
// No further editing necessary below!
////////////////////////////////////////////////////////////////////



////////////////////////////////////////////////////////////////////
// Singleton pattern:
//
FunctionRegistry* FunctionRegistry::m_theFunctionRegistry = 0;

FunctionRegistry::FunctionRegistry(){}

FunctionRegistry::~FunctionRegistry(){}

FunctionRegistry* FunctionRegistry::getInstance()
{
  if (m_theFunctionRegistry == 0)
    m_theFunctionRegistry = new FunctionRegistry();
  return m_theFunctionRegistry;
}
//
////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////
// Other:
//
void
FunctionRegistry::addFunctionReg(const string& functionID,
                                 const FunctionReg functionReg)
{
  if (isRegisteredFunction(functionID))
    throw invalid_argument("Multiple registration of functionID " + functionID);

  m_functionRegs[functionID] = functionReg;
  m_functionIDs.insert(functionID);
}

const FunctionReg&
FunctionRegistry::getFunctionReg(const string& functionID)
const
{
  map<string, FunctionReg>::const_iterator pos;
  pos = m_functionRegs.find(functionID);
  if (pos == m_functionRegs.end())
    throw invalid_argument("Invalid functionID " + functionID);
  return pos->second;
}

void
FunctionRegistry::registerFunction(const string& functionID)
{
  if (isRegisteredFunction(functionID))
    throw invalid_argument("Multiple registration of functionID " + functionID);

  Function* function = createFunction(functionID);

  FunctionReg functionReg = function->getFunctionReg();
  addFunctionReg(functionID, functionReg);

  m_parameterRegistries[functionID] = ParameterRegistry();
  ParameterRegistry& parameterRegistry = m_parameterRegistries[functionID];

  set<string> paramIDs = function->getParamIDs();
  set<string>::iterator paramID;
  for (paramID = paramIDs.begin(); paramID != paramIDs.end(); paramID++) {
    if (parameterRegistry.isRegisteredParam(*paramID))
      throw invalid_argument("Multiple registration of functionID, paramID "
                             + functionID + "," + *paramID);
      const ParamReg* paramReg = function->getParamReg(*paramID);
      parameterRegistry.addParamReg(*paramID, paramReg);
      delete paramReg;
  }

  delete function;
}

bool
FunctionRegistry::isRegisteredFunction(const string& functionID)
const
{
  if (m_functionRegs.count(functionID))
    return true;
  return false;
}

set<string>
FunctionRegistry::getFunctionIDs()
const
{
  return m_functionIDs;
}

string
FunctionRegistry::getFunctionName(const string& functionID)
const
{
  return getFunctionReg(functionID).name;
}


string
FunctionRegistry::getFunctionDesc(const string& functionID)
const
{
  return getFunctionReg(functionID).description;
}

string
FunctionRegistry::getFunctionType(const string& functionID)
const
{
  Function* function = createFunction(functionID);
  string functionType = function->getType();
  delete function;
  return functionType;
}

const ParameterRegistry&
FunctionRegistry::getParameterRegistry(const string& functionID)
const
{
  map<string, ParameterRegistry>::const_iterator pos;
  pos = m_parameterRegistries.find(functionID);
  if (pos == m_parameterRegistries.end())
    throw invalid_argument("Invalid functionID " + functionID);
  return pos->second;
}

bool
FunctionRegistry::isRegisteredParam(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).isRegisteredParam(paramID);
}

set<string>
FunctionRegistry::getParamIDs(const string& functionID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamIDs();
}

string
FunctionRegistry::getParamType(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamType(paramID);
}

string
FunctionRegistry::getParamName(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamName(paramID);
}

string
FunctionRegistry::getParamDesc(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamDesc(paramID);
}

int
FunctionRegistry::getParamDefValInt(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamDefValInt(paramID);
}

double
FunctionRegistry::getParamDefValDouble(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamDefValDouble(paramID);
}

string
FunctionRegistry::getParamDefValString(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamDefValString(paramID);
}

vector<double>
FunctionRegistry::getParamEnumValues(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamEnumValues(paramID);
}

unsigned
FunctionRegistry::getParamEnumDefIndex(const string& functionID, const string& paramID)
const
{
  if (!isRegisteredFunction(functionID))
    throw invalid_argument("Invalid functionID " + functionID);
  return getParameterRegistry(functionID).getParamEnumDefIndex(paramID);
}

