////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef IPROVIDER_H
#define IPROVIDER_H

#include <string>

using namespace std;

static const string providerTypeScalar = "scalar";
static const string providerTypeVector = "vector";

class IProvider
{
public:
  virtual ~IProvider() {};
  virtual void dummy() {};
};

bool isDensityProvider(IProvider* ip);
bool isTemperatureProvider(IProvider* ip);
bool isAbundanceProvider(IProvider* ip);
bool isTdustProvider(IProvider* ip);
bool isDopplerProvider(IProvider* ip);
bool isVelocityProvider(IProvider* ip);
bool isBmagProvider(IProvider* ip);

// Density
class IDensityProvider: public IProvider
{
public:
  const static string providerID;
  const static string providerName;
  const static string providerDesc;
  const static string providerUnit;
  const static string providerType;

  virtual ~IDensityProvider() {};

  virtual const double
  density(const double x, const double y, const double z) = 0;
};

// Gas temperature
class ITemperatureProvider: public IProvider
{
public:
  const static string providerID;
  const static string providerName;
  const static string providerDesc;
  const static string providerUnit;
  const static string providerType;

  virtual ~ITemperatureProvider() {};

  virtual const double
  temperature(const double x, const double y, const double z) = 0;
};

// Abundance
class IAbundanceProvider: public IProvider
{
public:
  const static string providerID;
  const static string providerName;
  const static string providerDesc;
  const static string providerUnit;
  const static string providerType;

  virtual ~IAbundanceProvider() {};

  virtual const double
  abundance(const double x, const double y, const double z) = 0;
};

// Dust temperature
class ITdustProvider: public IProvider
{
public:
  const static string providerID;
  const static string providerName;
  const static string providerDesc;
  const static string providerUnit;
  const static string providerType;

  virtual ~ITdustProvider() {};

  virtual const double
  t_dust(const double x, const double y, const double z) = 0;
};

// Velocity dispersion
class IDopplerProvider: public IProvider
{
public:
  const static string providerID;
  const static string providerName;
  const static string providerDesc;
  const static string providerUnit;
  const static string providerType;

  virtual ~IDopplerProvider() {};

  virtual const double
  doppler(const double x, const double y, const double z) = 0;
};

// Velocity
class IVelocityProvider: public IProvider
{
public:
  const static string providerID;
  const static string providerName;
  const static string providerDesc;
  const static string providerUnit;
  const static string providerType;

  virtual ~IVelocityProvider() {};

  virtual const void
  velocity(const double x, const double y, const double z, double* v) = 0;
};

// Magnetic field
class IBmagProvider: public IProvider
{
public:
  const static string providerID;
  const static string providerName;
  const static string providerDesc;
  const static string providerUnit;
  const static string providerType;

  virtual ~IBmagProvider() {};

  virtual const void
  bmag(const double x, const double y, const double z, double* b) = 0;
};


// Adaptor Temperature - Tdust
// Gives an ITemperatureProvider an ITdustProvider interface
class TempTdustAdaptor: public ITdustProvider
{
public:
	TempTdustAdaptor(ITemperatureProvider* pTemperatureProvider);
	virtual ~TempTdustAdaptor();

	virtual const double
	t_dust(const double x, const double y, const double z);

private:
	ITemperatureProvider* m_pTemperatureProvider;
};

// Adaptor Tdust - Temperature
// Gives an ITdustProvider an ITemperatureProvider interface
class TdustTempAdaptor: public ITemperatureProvider
{
public:
	TdustTempAdaptor(ITdustProvider* pTdustProvider);
	virtual ~TdustTempAdaptor();

	virtual const double
	temperature(const double x, const double y, const double z);

private:
	ITdustProvider* m_pTdustProvider;
};




#endif // IPROVIDER_H
