////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

/*!
 @file icppModellib.cc
 @author rschaaf

 Interface to be used from C++
 Implementation file
 */

#include <string>
#include <set>
#include <vector>

#include "icppModellib.h"
#include "modellib.h"


////////////////////////////////////////////////////////////////////
//! Functions that allow the exploration of the model library.
/**
 * @name Exploration
 */
//@{

// Exploration:

// Models:

////////////////////////////////////////////////////////////////////
//! The IDs of all models registered with the model library.
/*!
 * Returns the IDs of all registered models.
 */
set<string>
getModelIDs()
{
  return ModelLib::getInstance()->getModelIDs();
}

////////////////////////////////////////////////////////////////////
//! Checks if a model is registered with the model library.
/*!
 * Returns true if model with modelID is registered, else false.
 *
 * @param[in]   modelID
 *              ID of the model.
 */
bool
isRegisteredModel(const string modelID)
{
  return ModelLib::getInstance()->isRegisteredModel(modelID);
}

////////////////////////////////////////////////////////////////////
//! Name of a model.
/*!
 * Returns the name of the model.
 *
 * @param       modelID
 *              ID of the model.
 * @exception   invalid_argument
 *              if no model with modelID registered.
 */
string
getModelName(const string modelID)
{
  return ModelLib::getInstance()->getModelName(modelID);
}

////////////////////////////////////////////////////////////////////
//! Description of a model.
/*!
 * Returns the description of the model.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @exception   invalid_argument
 *              if no model with modelID registered.
 */
string
getModelDesc(const string modelID)
{
  return ModelLib::getInstance()->getModelDesc(modelID);
}

////////////////////////////////////////////////////////////////////
//! Bibliographic reference of a model.
/*!
 * Returns the bibliographic reference of the model.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @exception   invalid_argument
 *              if no model with modelID registered.
 */
string
getModelBibref(const string modelID)
{
  return ModelLib::getInstance()->getModelBibref(modelID);
}

////////////////////////////////////////////////////////////////////
//! The IDs of all parameters registered for a model.
/*!
 * Returns the IDs of all registered parameters for model.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @exception   invalid_argument
 *              if no model with modelID registered.
 */
set<string>
getParamIDs(const string modelID)
{
  return ModelLib::getInstance()->getParamIDs(modelID);
}

////////////////////////////////////////////////////////////////////
//! Checks if a model parameter is registered with the model library.
/*!
 * Returns true if parameter with paramID is registered
 *              for model with modelID, else false.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[out]  paramID
 *              ID of the parameter.
 */
bool
isRegisteredParam(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->isRegisteredParam(string(modelID), string(
      paramID));
}

////////////////////////////////////////////////////////////////////
//! Name of a model parameter.
/*!
 * Returns the name of the model parameter.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 */
string
getParamName(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamName(modelID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Description of a model parameter.
/*!
 * Returns the description of the model parameter.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 */
string
getParamDesc(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamDesc(modelID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Datatype of a model parameter.
/*!
 * Returns the datatype of the model parameter.
 * Valid datatypes are "double", "int", and "string".
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 */
string
getParamType(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamType(modelID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Unit of a model parameter.
/*!
 * Returns the unit of the model parameter.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is of type
 *              without unit.
 */
string
getParamUnit(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamUnit(modelID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Default value of an integer model parameter.
/*!
 * Returns the integer default value of the model parameter.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type int.
 */
int
getParamDefValInt(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamDefValInt(modelID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Default value of a double model parameter.
/*!
 * Returns the double default value of the model parameter.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type double.
 */
double
getParamDefValDouble(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamDefValDouble(modelID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Default value of a string model parameter.
/*!
 * Returns the string default value of the model parameter.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type string.
 */
string
getParamDefValString(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamDefValString(modelID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Values of an enum model parameter.
/*!
 * Returns the enum values of the model parameter.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type enum.
 */
vector<double>
getParamEnumValues(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamEnumValues(modelID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Default index of an enum model parameter.
/*!
 * Returns the default index of the model parameter.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type enum.
 */
unsigned getParamEnumDefIndex(const string modelID, const string paramID)
{
  return ModelLib::getInstance()->getParamEnumDefIndex(modelID, paramID);
}



// Results:

////////////////////////////////////////////////////////////////////
//! The IDs of all results registered with the model library.
/*!
 * Returns the IDs of all registered results.
 */
set<string>
getResultIDs()
{
  return ModelLib::getInstance()->getProviderIDs();
}

////////////////////////////////////////////////////////////////////
//! Checks if a result is registered with the model library.
/*!
 * Returns true if result with resultID is registered, else false.
 *
 * @param[in]   resultID
 *              ID of the result.
 */
bool
isRegisteredResult(const string resultID)
{
  return ModelLib::getInstance()->isRegisteredProvider(resultID);
}

////////////////////////////////////////////////////////////////////
//! Name of a result.
/*!
 * Returns the name of the result.
 *
 * @param[in]   resultID
 *              ID of the result.
 * @exception   invalid_argument
 *              if no result with resultID registered.
 */
string
getResultName(const string resultID)
{
  return ModelLib::getInstance()->getProviderName(resultID);
}

////////////////////////////////////////////////////////////////////
//! Description of a result.
/*!
 * Returns the description of the result.
 *
 * @param[in]   resultID
 *              ID of the result.
 * @exception   invalid_argument
 *              if no result with resultID registered.
 */
string
getResultDesc(const string resultID)
{
  return ModelLib::getInstance()->getProviderDesc(resultID);
}

////////////////////////////////////////////////////////////////////
//! Unit of a result.
/*!
 * Returns the unit of the result.
 *
 * @param[in]   resultID
 *              ID of the result.
 * @exception   invalid_argument
 *              if no result with resultID registered.
 */
string
getResultUnit(const string resultID)
{
  return ModelLib::getInstance()->getProviderUnit(resultID);
}

////////////////////////////////////////////////////////////////////
//! Type of a result.
/*!
 * Returns the type of the result.
 * Valid types are "scalar" and "vector"
 *
 * @param[in]   resultID
 *              ID of the result.
 * @exception   invalid_argument
 *              if no result with resultID registered.
 */
string
getResultType(const string resultID)
{
  return ModelLib::getInstance()->getProviderType(resultID);
}

////////////////////////////////////////////////////////////////////
//! The IDs of all results registered for a model.
/*!
 * Returns the resultIDs registered for the model.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @exception   invalid_argument
 *              if no model with modelID registered.
 */
set<string>
getModelResultIDs(const string modelID)
{
  return ModelLib::getInstance()->getModelProviderIDs(modelID);
}

////////////////////////////////////////////////////////////////////
//! Checks if a Result is registered for a Model
/*!
 * Return True if Result with resultID is registered for the Model
 * with modelID, else False
 *
 * @param[in]   resultID
 *              ID of the result.
 * @param[in]   modelID
 *              ID of the model.
 * @exception   invalid_argument
 *              if no result with resultID registered.
 * @exception   invalid_argument
 *              if no model with modelID registered.
 */
bool
isResultModel(const string resultID, const string modelID)
{
  return ModelLib::getInstance()->isProviderModel(resultID, modelID);
}

// Functions:

////////////////////////////////////////////////////////////////////
//! The IDs of all functions registered with the model library.
/*!
 * Returns the IDs of all registered functions.
 */
set<string>
getFunctionIDs()
{
  return ModelLib::getInstance()->getFunctionIDs();
}

////////////////////////////////////////////////////////////////////
//! Checks if a function is registered with the model library.
/*!
 * Returns true if function with functionID is registered, else false.
 *
 * @param[in]   functionID
 *              ID of the model.
 */
bool
isRegisteredFunction(const string functionID)
{
  return ModelLib::getInstance()->isRegisteredFunction(functionID);
}

////////////////////////////////////////////////////////////////////
//! Name of a function.
/*!
 * Returns the name of the function.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @exception   invalid_argument
 *              if no function with functionID registered.
 */
string
getFunctionName(const string functionID)
{
  return ModelLib::getInstance()->getFunctionName(functionID);
}

////////////////////////////////////////////////////////////////////
//! Description of a function.
/*!
 * Returns the description of the function.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @exception   invalid_argument
 *              if no function with functionID registered.
 */
string
getFunctionDesc(const string functionID)
{
  return ModelLib::getInstance()->getFunctionDesc(functionID);
}

////////////////////////////////////////////////////////////////////
//! Type of a function.
/*!
 * Returns the type of the function.
 * Valid types are "scalar" and "vector".
 *
 * @param[in]   functionID
 *              ID of the function.
 * @exception   invalid_argument
 *              if no function with functionID registered.
 */
string
getFunctionType(const string functionID)
{
  return ModelLib::getInstance()->getFunctionType(functionID);
}

////////////////////////////////////////////////////////////////////
//! The IDs of all Results to which a Function can be linked.
/*!
 * Returns the resultIDs of all Results to which the Function can be linked
 *
 * @param[in]   functionID
 *              ID of the function.
 * @exception   invalid_argument
 *              if no function with functionID registered.
 */
set<string>
getFunctionResultIDs(const string functionID)
{
  return ModelLib::getInstance()->getFunctionProviderIDs(functionID);
}

////////////////////////////////////////////////////////////////////
//! The IDs of all Functions that can be linked to a Result.
/*!
 * Returns the functionIDs of all Functions that can be linked to the Result.
 *
 * @param[in]   resultID
 *              ID of the result.
 * @exception   invalid_argument
 *              if no result with resultID registered.
 */
set<string>
getResultFunctionIDs(const string resultID)
{
  return ModelLib::getInstance()->getProviderFunctionIDs(resultID);
}

////////////////////////////////////////////////////////////////////
//! Checks if a Function can be linked to a Result.
/*!
 * True if the Function with functionID can be linked to the Result
 * with resultID, else False
 *
 * @param[in]   resultID
 *              ID of the result.
 * @param[in]   functionID
 *              ID of the function.
 */
bool
isResultFunction(const string resultID, const string functionID)
{
  return ModelLib::getInstance()->isProviderFunction(resultID, functionID);
}

////////////////////////////////////////////////////////////////////
//! The IDs of all parameters registered for a function.
/*!
 * Returns the IDs of all registered parameters for function.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @exception   invalid_argument
 *              if no function with functionID registered.
 */
set<string>
getFunctionParamIDs(const string functionID)
{
  return ModelLib::getInstance()->getFunctionParamIDs(functionID);
}

////////////////////////////////////////////////////////////////////
//! Checks if a function parameter is registered with the model library.
/*!
 * Returns true if parameter with paramID is registered for function
 * with functionID, else false.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[out]  paramID
 *              ID of the parameter.
 */
bool
isRegisteredFunctionParam(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->isRegisteredFunctionParam(string(functionID),
      string(paramID));
}

////////////////////////////////////////////////////////////////////
//! Name of a function parameter.
/*!
 * Returns the name of the function parameter.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no function with functionID
 *              or parameter with paramID registered.
 */
string
getFunctionParamName(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamName(functionID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Description of a function parameter.
/*!
 * Returns the description of the function parameter.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no function with functionID
 *              or parameter with paramID registered.
 */
string
getFunctionParamDesc(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamDesc(functionID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Datatype of a function parameter.
/*!
 * Returns the datatype of the function parameter.
 * Valid datatypes are "double", "int", and "string".
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no function with functionID
 *              or parameter with paramID registered.
 */
string
getFunctionParamType(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamType(functionID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Default value of an integer function parameter.
/*!
 * Returns the integer default value of the function parameter.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no function with functionID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type int.
 */
int
getFunctionParamDefValInt(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamDefValInt(functionID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Default value of a double function parameter.
/*!
 * Returns the double default value of the function parameter.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no function with functionID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type double.
 */
double
getFunctionParamDefValDouble(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamDefValDouble(functionID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Default value of a string function parameter.
/*!
 * Returns the string default value of the function parameter.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no function with functionID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type string.
 */
string
getFunctionParamDefValString(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamDefValString(functionID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Values of an enum function parameter.
/*!
 * Returns the enum values of the function parameter.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no function with functionID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type enum.
 */
vector<double>
getFunctionParamEnumValues(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamEnumValues(functionID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Default index of an enum function parameter.
/*!
 * Returns the default index of the function parameter.
 *
 * @param[in]   functionID
 *              ID of the function.
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no function with functionID
 *              or parameter with paramID registered.
 * @exception   domain_error
 *              if parameter with paramID is not of type enum.
 */
unsigned getFunctionParamEnumDefIndex(const string functionID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamEnumDefIndex(functionID, paramID);
}

//@}

////////////////////////////////////////////////////////////////////
//! Functions for setting up of the model library.
/**
 * @name Setup
 */
//@{
// Setup:

// Model:

////////////////////////////////////////////////////////////////////
//! Select a Current Model.
/*!
 * All Parameters of the selected Model are set to their default values.
 *
 * @param[in]   modelID
 *              ID of the model.
 * @exception   invalid_argument
 *              if no model with modelID
 *              or parameter with paramID registered.
 */
void
setCurrentModel(const string modelID)
{
  ModelLib::getInstance()->setCurrentModel(modelID);
}

////////////////////////////////////////////////////////////////////
//! Unset the Current Model.
/*!
 */
void
unsetCurrentModel()
{
  ModelLib::getInstance()->unsetCurrentModel();
}

////////////////////////////////////////////////////////////////////
//! Checks if a Current Model is selected.
/*!
 */
bool
isSetCurrentModel()
{
  return ModelLib::getInstance()->isSetCurrentModel();
}

////////////////////////////////////////////////////////////////////
//! The ID of the Current Model.
/*!
 * Returns the ID of the Current Model; if no Current Model is selected,
 * the empty string is returned.
 */
string
getCurrentModelID()
{
  return ModelLib::getInstance()->getCurrentModelID();
}

////////////////////////////////////////////////////////////////////
//! Set the value of a Parameter of data type 'double' of the Current Model.
/*!
 * @param[in]   paramID
 *              ID of the parameter.
 * @param[in]   val
 * 	            The value of the parameter
 *
 * @exception   invalid_argument
 *              if no double parameter with paramID
 *              is registered for the Current Model.
 */
void
setParamDouble(const string paramID, double val)
{
  ModelLib::getInstance()->setParamDouble(paramID, val);
}

////////////////////////////////////////////////////////////////////
//! Set the value of a Parameter of data type 'int' of the Current Model.
/*!
 * @param[in]   paramID
 *              ID of the parameter.
 * @param[in]   val
 * 	            The value of the parameter
 *
 * @exception   invalid_argument
 *              if no int parameter with paramID
 *              is registered for the Current Model.
 */
void
setParamInt(const string paramID, int val)
{
  ModelLib::getInstance()->setParamInt(paramID, val);
}

////////////////////////////////////////////////////////////////////
//! Set the value of a Parameter of data type 'string' of the Current Model.
/*!
 * @param[in]   paramID
 *              ID of the parameter.
 * @param[in]   val
 * 	            The value of the parameter
 *
 * @exception   invalid_argument
 *              if no string parameter with paramID
 *              is registered for the Current Model.
 */
void
setParamString(const string paramID, const string val)
{
  ModelLib::getInstance()->setParamString(paramID, val);
}

////////////////////////////////////////////////////////////////////
//! Set the enumeration index of a Parameter of data type 'enum' of the Current Model.
/*!
 * @param[in]   paramID
 *              ID of the parameter.
 * @param[in]   index
 * 	            The enumeration index of the parameter
 *
 * @exception   invalid_argument
 *              if no enum parameter with paramID
 *              is registered for the Current Model.
 * @exception   domain_error
 *              if invalid index is specified.
 */
void
setParamEnumIndex(const string paramID, const unsigned index)
{
  ModelLib::getInstance()->setParamEnumIndex(paramID, index);
}

////////////////////////////////////////////////////////////////////
//! Get the value of a Parameter of data type 'double' of the Current Model.
/*!
 * Returns the value of the parameter.
 *
 * @param[in]   paramID
 *              ID of the parameter.
 *
 * @exception   invalid_argument
 *              if no double parameter with paramID
 *              is registered for the Current Model.
 */
double
getParamDouble(const string paramID)
{
  return ModelLib::getInstance()->getParamDouble(paramID);
}

////////////////////////////////////////////////////////////////////
//! Get the value of a Parameter of data type 'int' of the Current Model.
/*!
 * Returns the value of the parameter.
 *
 * @param[in]   paramID
 *              ID of the parameter.
 *
 * @exception   invalid_argument
 *              if no int parameter with paramID
 *              is registered for the Current Model.
 */
int
getParamInt(const string paramID)
{
  return ModelLib::getInstance()->getParamInt(paramID);
}

////////////////////////////////////////////////////////////////////
//! Get the value of a Parameter of data type 'string' of the Current Model.
/*!
 * Returns the value of the parameter.
 *
 * @param[in]   paramID
 *              ID of the parameter.
 *
 * @exception   invalid_argument
 *              if no string parameter with paramID
 *              is registered for the Current Model.
 */
string
getParamString(const string paramID)
{
  return ModelLib::getInstance()->getParamString(paramID);
}

////////////////////////////////////////////////////////////////////
//! Get the enumeration index of a Parameter of data type 'enum' of the Current Model.
/*!
 * Returns the enumeration index of the parameter.
 *
 * @param[in]   paramID
 *              ID of the parameter.
 * @exception   invalid_argument
 *              if no enum parameter with paramID
 *              is registered for the Current Model.
 */
unsigned
getParamEnumIndex(const string paramID)
{
  return ModelLib::getInstance()->getParamEnumIndex(paramID);
}

////////////////////////////////////////////////////////////////////
//! Link a Function to a Result.
/*!
 * All Function Parameters of the Function are set to their default values.
 *
 * @param[in]   resultID
 *              ID of the result.
 * @param[in]   functionID
 *              ID of the function.
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no function with functionID is registered.
 * @exception   invalid_argument
 *              if no function and parameter are not of identical type.
 */
void
setFunction(const string resultID, const string functionID)
{
  ModelLib::getInstance()->setFunction(resultID, functionID);
}

////////////////////////////////////////////////////////////////////
//! Unlink any Function that is linked to a Result.
/*!
 * @param[in]   resultID
 *              ID of the result.
 *
 * @exception   domain_error
 *              if no function set for specified result.
 */
void
unsetFunction(const string resultID)
{
  ModelLib::getInstance()->unsetFunction(resultID);
}

////////////////////////////////////////////////////////////////////
//! Get the ID of the Function linked to a Result.
/*!
 * If no function is linked to the result, the empty string is returned.
 *
 * @param[in]   resultID
 *              ID of the result.
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 */
const string
getFunctionID(const string resultID)
{
  return ModelLib::getInstance()->getFunctionID(resultID);
}

////////////////////////////////////////////////////////////////////
//! Set the value of a Function Parameter of data type 'double'.
/*!
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 * @param[in]   paramID
 *              ID of the function parameter.
 * @param[in]   val
 * 	            The value of the function parameter
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no double function parameter with paramID
 *              is registered for function linked to the result.
 */
void
setFunctionParamDouble(const string resultID, const string paramID,
    double val)
{
  ModelLib::getInstance()->setFunctionParamDouble(resultID, paramID, val);
}

////////////////////////////////////////////////////////////////////
//! Set the value of a Function Parameter of data type 'int'.
/*!
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 * @param[in]   paramID
 *              ID of the function parameter.
 * @param[in]   val
 * 	            The value of the function parameter
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no int function parameter with paramID
 *              is registered for function linked to the result.
 */
void
setFunctionParamInt(const string resultID, const string paramID, int val)
{
  ModelLib::getInstance()->setFunctionParamInt(resultID, paramID, val);
}

////////////////////////////////////////////////////////////////////
//! Set the value of a Function Parameter of data type 'string'.
/*!
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 * @param[in]   paramID
 *              ID of the function parameter.
 * @param[in]   val
 * 	            The value of the function parameter
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no string function parameter with paramID
 *              is registered for function linked to the result.
 */
void
setFunctionParamString(const string resultID, const string paramID,
    const string val)
{
  ModelLib::getInstance()->setFunctionParamString(resultID, paramID, val);
}

////////////////////////////////////////////////////////////////////
//! Set the enumeration index of a Function Parameter of data type 'enum'.
/*!
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 * @param[in]   paramID
 *              ID of the function parameter.
 * @param[in]   index
 * 	            The enumeration index of the function parameter
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no enum function parameter with paramID
 *              is registered for function linked to the result.
 * @exception   domain_error
 *              if invalid index is specified.
 */
void
setFunctionParamEnumIndex(const string resultID, const string paramID, unsigned index)
{
  ModelLib::getInstance()->setFunctionParamEnumIndex(resultID, paramID, index);
}

////////////////////////////////////////////////////////////////////
//! Get the value of a Function Parameter of data type 'double'.
/*!
 * Returns the value of the parameter.
 *
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 * @param[in]   paramID
 *              ID of the parameter.
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no double parameter with paramID
 *              is registered for function linked to the result.
 */
double
getFunctionParamDouble(const string resultID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamDouble(resultID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Get the value of a Function Parameter of data type 'int'.
/*!
 * Returns the value of the parameter.
 *
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 * @param[in]   paramID
 *              ID of the parameter.
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no int parameter with paramID
 *              is registered for function linked to the result.
 */
int
getFunctionParamInt(const string resultID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamInt(resultID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Get the value of a Function Parameter of data type 'string'.
/*!
 * Returns the value of the parameter.
 *
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 * @param[in]   paramID
 *              ID of the parameter.
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no string parameter with paramID
 *              is registered for function linked to the result.
 */
string
getFunctionParamString(const string resultID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamString(resultID, paramID);
}

////////////////////////////////////////////////////////////////////
//! Get the enumeration index of a Function Parameter of data type 'enum'.
/*!
 * Returns the enumeration index of the Function Parameter.
 *
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 * @param[in]   paramID
 *              ID of the function parameter.
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 * @exception   invalid_argument
 *              if no enum function parameter with paramID
 *              is registered for function linked to the result.
 */
unsigned
getFunctionParamEnumIndex(const string resultID, const string paramID)
{
  return ModelLib::getInstance()->getFunctionParamEnumIndex(resultID, paramID);
}

// Identification of Tdust with Temperature and vice versa:

////////////////////////////////////////////////////////////////////
//! Identifies the Result "tdust" with the Result "temperature".
/*!
 * @exception   domain_error
 *              if "temperature" already identified with "tdust".
 */
void
setTdustIdentTemp()
{
	ModelLib::getInstance()->setTdustIdentTemp();
}

////////////////////////////////////////////////////////////////////
//! Unset the identification of the Result "tdust" with the Result "temperature".
/*!
 * @exception   domain_error
 *              if "tdust" not identified with "temperature".
 */
void
unsetTdustIdentTemp()
{
	ModelLib::getInstance()->unsetTdustIdentTemp();
}

////////////////////////////////////////////////////////////////////
//! Checks if the Result "tdust" is identified with the Result "temperature".
/*!
 * Returns True if  the Result "tdust" is identified with the Result "temperature",
 * else False.
 */
const bool
isTdustIdentTemp()
{
	return ModelLib::getInstance()->isTdustIdentTemp();
}

////////////////////////////////////////////////////////////////////
//! Identifies the Result "temperature" with the Result "tdust".
/*!
 * @exception   domain_error
 *              if "tdust" already identified with "temperature".
 */
void
setTempIdentTdust()
{
	ModelLib::getInstance()->setTempIdentTdust();
}

////////////////////////////////////////////////////////////////////
//! Unset the identification of the Result "temperature" with the Result "tdust".
/*!
 * @exception   domain_error
 *              if "temperature" not identified with "tdust".
 */
void
unsetTempIdentTdust()
{
	ModelLib::getInstance()->unsetTempIdentTdust();
}

////////////////////////////////////////////////////////////////////
//! Checks if the Result "temperature" is identified with the Result "tdust".
/*!
 * Returns True if  the Result "temperature" is identified with the Result "tdust",
 * else False.
 */
const bool
isTempIdentTdust()
{
	return ModelLib::getInstance()->isTempIdentTdust();
}


// Results:

////////////////////////////////////////////////////////////////////
//! Checks if a Result is provided by the Current Model.
/*!
 * Returns True if the Result is provided by the Current Model, else False.
 *
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 */
const bool
isCurrentResultModel(const string resultID)
{
  return ModelLib::getInstance()->isCurrentProviderModel(resultID);
}

////////////////////////////////////////////////////////////////////
//! Checks if a Result is linked to a Function.
/*!
 * Returns True if the Result is linked to a Function, else False.
 *
 * @param[in]   resultID
 *              ID of the result to which the function is linked.
 *
 * @exception   invalid_argument
 *              if no result with resultID is registered.
 */
const bool
isCurrentResultFunction(const string resultID)
{
  return ModelLib::getInstance()->isCurrentProviderFunction(resultID);
}

////////////////////////////////////////////////////////////////////
//! Checks if all Results are either provided by the Current Model or linked to a Function.
/*!
 * Returns True if configuration is complete, else False.
 */
const bool
isConfigurationComplete()
{
	return ModelLib::getInstance()->isConfigurationComplete();
}


////////////////////////////////////////////////////////////////////
//! Finalize the configuration.
/*!
 * Depending on the Current Model and Functions linked to Results,
 * internal variables of the Current Model are set, data from files
 * may be read in, etc.
 */
void
finalizeConfiguration()
{
  ModelLib::getInstance()->finalizeConfiguration();
}

////////////////////////////////////////////////////////////////////
//! Checks if the configuration has been finalized.
/*!
 * Returns True if configuration has been finalized, else False.
 */
bool
isFinalizedConfiguration()
{
  return ModelLib::getInstance()->isFinalizedConfiguration();
}

//@}

////////////////////////////////////////////////////////////////////
//! Functions for using the model library.
/**
 * @name Usage
 */
//@{

// Usage:

////////////////////////////////////////////////////////////////////
//! Particle density of current model at single point.
/*!
 * Returns the particle density at x,y,z (m^-3).
 *
 * @param[in]   x,y,z
 *              Cartesian coordinates (m).
 */
double
density(double x, double y, double z)
{
  return ModelLib::getInstance()->density(x, y, z);
}

////////////////////////////////////////////////////////////////////
//! Gas temperature of current model at single point.
/*!
 * Returns the gas temperature at x,y,z (K).
 *
 * @param[in]   x,y,z
 *              Cartesian coordinates (m).
 */
double
temperature(double x, double y, double z)
{
  return ModelLib::getInstance()->temperature(x, y, z);
}

////////////////////////////////////////////////////////////////////
//! Abundance of current model at single point.
/*!
 * Returns the abundance at x,y,z.
 *
 * @param[in]   x,y,z
 *              Cartesian coordinates (m).
 */
double
abundance(double x, double y, double z)
{
  return ModelLib::getInstance()->abundance(x, y, z);
}

////////////////////////////////////////////////////////////////////
//! Dust temperature of current model at single point.
/*!
 * Returns the dust temperature at x,y,z (K).
 *
 * @param[in]   x,y,z
 *              Cartesian coordinates (m).
 */
double
t_dust(double x, double y, double z)
{
  return ModelLib::getInstance()->t_dust(x, y, z);
}

////////////////////////////////////////////////////////////////////
//! Doppler b-parameter of current model at single point.
/*!
 * Returns the Doppler b-parameter (1/e half-width of line profile) at x,y,z (m/s).
 *
 * @param[in]   x,y,z
 *              Cartesian coordinates (m).
 */
double
doppler(double x, double y, double z)
{
  return ModelLib::getInstance()->doppler(x, y, z);
}

////////////////////////////////////////////////////////////////////
//! Velocity  of current model at single point.
/*!
 * Returns the velocity (cartesian components) at x,y,z (m/s).
 *
 * @param[in]   x,y,z
 *              Cartesian coordinates (m).
 */
vector<double>
velocity(double x, double y, double z)
{
  double v[3];
  vector<double> vv(3);

  ModelLib::getInstance()->velocity(x, y, z, v);

  for (int i = 0; i < 3; i++)
    vv[i] = v[i];
  return vv;
}

////////////////////////////////////////////////////////////////////
//! Magnetic field  of current model at single point.
/*!
 * Returns the magnetic field (cartesian components) at x,y,z (Gauss).
 *
 * @param[in]   x,y,z
 *              Cartesian coordinates (m).
 */
vector<double>
bmag(double x, double y, double z)
{
  double b[3];
  vector<double> vb(3);

  ModelLib::getInstance()->bmag(x, y, z, b);

  for (int i = 0; i < 3; i++)
    vb[i] = b[i];
  return vb;
}

//@}
