////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <stdexcept>

#include "model.h"
#include "parameterregistry.h"

using namespace std;

Model::Model()
: m_isFinalized(false)
{}

void
Model::initializeParams()
{
  set<string> paramIDs = getParamIDs();
  set<string>::iterator paramID;
  for (paramID = paramIDs.begin(); paramID != paramIDs.end(); ++paramID) {
      const ParamReg* paramReg = getParamReg(*paramID);

      bool foundType = false;

      const ParamRegInt* iParamReg = dynamic_cast<const ParamRegInt*>(paramReg);
      if (iParamReg) {
         setParamInt(*paramID, iParamReg->getDefValue());
         foundType = true;
      }
      const ParamRegDouble* dParamReg = dynamic_cast<const ParamRegDouble*>(paramReg);
      if (dParamReg) {
        setParamDouble(*paramID, dParamReg->getDefValue());
        foundType = true;
      }
      const ParamRegString* sParamReg = dynamic_cast<const ParamRegString*>(paramReg);
      if (sParamReg) {
        setParamString(*paramID, sParamReg->getDefValue());
        foundType = true;
      }
      const ParamRegEnum* eParamReg = dynamic_cast<const ParamRegEnum*>(paramReg);
      if (eParamReg) {
        setParamEnumIndex(*paramID, eParamReg->getDefIndex());
        foundType = true;
      }
      if (!foundType)
        throw domain_error("Invalid parameter type for paramID " + *paramID);
  }
}


void
Model::setParamDouble(string paramID, double val)
{
  m_paramDouble[paramID] = val;
  m_isFinalized = false;
}

void
Model::setParamInt(string paramID, int val)
{
  m_paramInt[paramID] = val;
  m_isFinalized = false;
}

void
Model::setParamString(string paramID, string val)
{
  m_paramString[paramID] = val;
  m_isFinalized = false;
}

void
Model::setParamEnumIndex(string paramID, unsigned index)
{
  const ParamRegEnum* eParamReg = dynamic_cast<const ParamRegEnum*>(getParamReg(paramID));
  if (!eParamReg)
    throw domain_error("Invalid parameter type for paramID " + paramID);

  vector<double> values = eParamReg->getValues();

  if (index >= values.size())
    throw invalid_argument("Invalid index for enum parameter " + paramID);

  m_paramEnumIndex[paramID] = index;
  m_paramEnumValue[paramID] = values[index];
}

const double
Model::getParamDouble(const string paramID)
{
  if (m_paramDouble.find(paramID) == m_paramDouble.end())
    throw invalid_argument("Invalid paramID " + paramID);
  return m_paramDouble[paramID];
}

const int
Model::getParamInt(const string paramID)
{
  if (m_paramInt.find(paramID) == m_paramInt.end())
    throw invalid_argument("Invalid paramID " + paramID);
  return m_paramInt[paramID];
}

const string
Model::getParamString(const string paramID)
{
  if (m_paramString.find(paramID) == m_paramString.end())
    throw invalid_argument("Invalid paramID " + paramID);
  return m_paramString[paramID];
}

const unsigned
Model::getParamEnumIndex(const string paramID)
{
  if (m_paramInt.find(paramID) == m_paramEnumIndex.end())
    throw invalid_argument("Invalid paramID " + paramID);
  return m_paramEnumIndex[paramID];
}

void
Model::finalizeConfiguration()
{
  onFinalizeConfiguration();
  m_isFinalized = true;
}

const bool
Model::isFinalizedConfiguration()
{
  return m_isFinalized;
}
