////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <math.h>

#include "modelAllen.h"

#include <stdexcept>

string
ModelAllen::modelID = "allen03a";

const ModelReg
ModelAllen::getModelReg()
{
  ModelReg modelReg;
  modelReg.name = "Allen 03a";
  modelReg.description = "Collapse of magnetized singular isothermal toroid (pivotal states)";
  modelReg.bibref = "Allen et al. 2003, ApJ, 599, 351";
  return modelReg;
}

const set<string>
ModelAllen::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("cs");
  paramIDs.insert("age");
  paramIDs.insert("Rn");
  paramIDs.insert("T");
  return paramIDs;
}

const ParamReg*
ModelAllen::getParamReg(string paramID)
{
  if (paramID == "cs")
    return new ParamRegDouble("cs",
                              "Velocity of sound",
                              1000.,
                              "m/s");
  else if (paramID == "age")
    return new ParamRegDouble("Age",
                              "Age of source",
                              1e+5, // 1e5*year
                              "year");
  else if (paramID == "Rn")
    return new ParamRegDouble("Rn",
                              "Validity radius of kinematic approximation",
                               1.0,// 100AU
                              "AU");
  else if (paramID == "T")
    return new ParamRegDouble("T",
                              "Temperature - constant",
                              30.,
                              "K");
  else
    throw invalid_argument("Invalid paramID " + paramID + " for modelID " + this->modelID);
}



ModelAllen::ModelAllen()
{}

ModelAllen::~ModelAllen()
{}


void 
ModelAllen::onFinalizeConfiguration()
{
  double ggrav = 6.674e-11; // m^3 kg^-1 s^-2

  m_v0 = m_paramDouble["cs"];
  m_t0 = m_paramDouble["age"]*31536000e+0;
  m_rn = m_paramDouble["Rn"]*1.496e+11;
  m_T  = m_paramDouble["T"];

  m_r0 = m_v0 * m_t0;
  m_n0 = 1. / (4. * M_PI * ggrav * m_t0 * m_t0 * 2.e-20); // mH2
  m_b0 = m_v0 / (sqrt(ggrav) * m_t0);
}

const double 
ModelAllen::density(const double x, const double y, const double z)
{
  double r, sinTheta;

  r = sqrt(x*x + y*y + z*z);
  sinTheta = sqrt(x*x + y*y) / r;

  if ( r < m_rn )
    r = m_rn;

  return m_n0 * pow(1.8 * sinTheta / (r / m_r0 ), 1.5);
}

const double
ModelAllen::temperature(const double x, const double y, const double z)
{
  return m_T;
}


const void
ModelAllen::velocity(const double x, const double y, const double z, double* v)
{
  double r, cosTheta, sinTheta, cosPhi, sinPhi;
  double vr;

  r = sqrt(x*x + y*y + z*z);  

  cosTheta = z / r;
  sinTheta = sqrt(x*x + y*y) / r;
  cosPhi = x / (r * sinTheta);
  sinPhi = y / (r * sinTheta);

  // Spherical components:
  vr = - m_v0 * exp(-(r / m_r0) * (r / m_r0)) / sqrt(r / m_r0);

  // Cartesian components:
  v[0] = vr * sinTheta * cosPhi;
  v[1] = vr * sinTheta * sinPhi;
  v[2] = vr * cosTheta;
}

const void
ModelAllen::bmag(const double x, const double y, const double z, double* b)
{
  double r, cosTheta, sinTheta, cosPhi, sinPhi;
  double r0975;
  double dphir, dphit;
  double br, btheta; 

  r = sqrt(x*x + y*y + z*z);    
  cosTheta = z / r;
  sinTheta = sqrt(x*x + y*y) / r;
  cosPhi = x / (r * sinTheta);
  sinPhi = y / (r * sinTheta);

  // Spherical components:
  r0975 = (r / m_r0) + 0.975;

  dphir = r0975 * M_PI * M_PI * sinTheta * sinTheta / 2; //  / ( (r / r0 ) * sinTheta ) ? 
  dphit = r0975 * r0975 * M_PI * M_PI * sinTheta * cosTheta / 2; //  / ( (r / r0 ) * sinTheta ) ? 

  br = m_b0 * dphit / r;
  btheta = - m_b0 * dphir;

  // Cartesian components:
  b[0] = br * sinTheta * cosPhi + btheta * cosTheta * cosPhi;
  b[1] = br * sinTheta * sinPhi + btheta * cosTheta * sinPhi;
  b[2] = br * cosTheta - btheta * sinPhi;
}


