////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Attila Juhasz
// Sterrewacht Leiden
// Universiteit Leiden
// juhasz@strw.leidenuniv.nl
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <math.h>

#include "modelBonnorEbert56.h"
#include <stdexcept>


string
ModelBonnorEbert56::modelID = "BonnorEbert56";

const ModelReg
ModelBonnorEbert56::getModelReg()
{
  ModelReg modelReg;
  modelReg.name = "Bonnor Ebert sphere";
  modelReg.description = "Isothermal hydrostatic molecular cloud core by Bonnor 1956 & Ebert 1955";
  modelReg.bibref = "Bonnor 1956, MNRAS, 116, 351 ; Ebert 1955, ZA (Zeitschrift fuer Astrophysik), 37, 217";
  return modelReg;
}

const set<string>
ModelBonnorEbert56::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("rhoc");
  paramIDs.insert("T");
  return paramIDs;
}

const ParamReg*
ModelBonnorEbert56::getParamReg(string paramID)
{
  if (paramID == "rhoc")
    return new ParamRegDouble("rhoc",
                              "Central volume density of the core",
                              1e6, // 1e3*mu*mp*1e3
                              "1/cm^3");
  else if (paramID == "T")
    return new ParamRegDouble("T",
                              "Temperature of the core",
                              10.,
                              "K");
  else
    throw invalid_argument("Invalid paramID " + paramID + " for modelID " + this->modelID);
}



ModelBonnorEbert56::ModelBonnorEbert56()
{}

ModelBonnorEbert56::~ModelBonnorEbert56()
{}


void 
ModelBonnorEbert56::onFinalizeConfiguration()
{
#define mp 1.6726000e-27 // Proton mass
#define mu 2.3 // mean molecular weight 

  m_rhoc   = m_paramDouble["rhoc"]*mp*mu*1e6;
  m_T      = m_paramDouble["T"];

}

const double 
ModelBonnorEbert56::density(const double x, const double y, const double z)
{
#define pi 3.14159265358979323846264338328
#define q 1.03 // Relative stepsize in the integration (r_i+1 / r_i)
#define mp 1.6726000e-27 // Proton mass (SI)
#define gg 6.674e-11 // gravitational constant (SI)
#define kk 1.3807e-23 // Boltzman constant (SI)
#define au 1.496e11 // Astronomical unit (SI)
#define mu 2.3 // mean molecular weight 


  double rho, r0, or1, r1, ri, r, dr, mr05, cs2, dum;
  int nstep, i;

// Set the boundary condition and pre-calculate some variables 

  r0 = 0e0;
  or1 = 10.*au;
  r1 = or1;
  rho = m_rhoc;
  r     = sqrt(x*x + y*y + z*z);
  
  if (r<=or1){
	return rho/(mu*mp);
  }


  cs2   = kk*m_T / (mu*mp);
  dum   = log(r/r1) / log(q);
  nstep = int(ceil(dum));
  mr05  = 0.0;
  for (i=0;i<nstep-1;i++){
		dr    = r1 - r0;
	    ri    = r0 + 0.5 * dr;
		mr05  = mr05 + 4. * pi * ri*ri * rho * dr;
		rho   = exp(log(rho) - gg*mr05/cs2/ri/ri*dr);
		r0    = r1;
		r1    = r0 * q;
  }

  r1   = or1 * pow(q, dum);
  ri   = r0 + 0.5*dr;
  dr   = r1 - r0;
  mr05 = mr05 + 4. * pi * ri*ri * rho * dr;
  rho  = exp(log(rho) - gg*mr05/cs2/ri/ri*dr) ;

   return rho/(mu*mp);
}
const double
ModelBonnorEbert56::temperature(const double x, const double y, const double z)
{
  return m_T; 
}

const void
ModelBonnorEbert56::velocity(const double x, const double y, const double z, double* v)
{
  
	v[0] = 0.0;
	v[1] = 0.0;
	v[2] = 0.0;	
}




