////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Attila Juhasz
// Sterrewacht Leiden
// Universiteit Leiden
// juhasz@strw.leidenuniv.nl
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <math.h>

#include "modelCG97.h"

#include <stdexcept>
#include <iostream>
#include <gsl/gsl_interp.h>


string
ModelCG97::modelID = "CG97";

const ModelReg
ModelCG97::getModelReg()
{
  ModelReg modelReg;
  modelReg.name = "Chiang & Goldreich 1997";
  modelReg.description = "Passive irradiated disk";
  modelReg.bibref = "Chiang & Goldreich 1997, ApJ, 490, 368";
  return modelReg;
}

const set<string>
ModelCG97::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("rin");
  paramIDs.insert("rout");
  paramIDs.insert("plsig1");
  paramIDs.insert("hph");
  paramIDs.insert("sig0");
  paramIDs.insert("Mstar");
  paramIDs.insert("Rstar");
  paramIDs.insert("Tstar");
  paramIDs.insert("bgdens");
  return paramIDs;
}

const ParamReg*
ModelCG97::getParamReg(string paramID)
{
  if (paramID == "rin")
    return new ParamRegDouble("rin",
                              "Inner Radius of the disk",
                              1.0, // AU
                              "AU");
  else if (paramID == "rout")
    return new ParamRegDouble("rout",
                              "Outer radius of the disk",
                              100.0,
                              "AU");
  else if (paramID == "plsig1")
    return new ParamRegDouble("plsig1",
                              "Power exponent of the radial surface density distribution",
                              -1.0,
                              "");
  else if (paramID == "hph")
    return new ParamRegDouble("hph",
                              "Ratio of the height of the disk atmosphere above the midplane  and the pressure scale height (at all radius)",
                              4.0,
                              "");
  else if (paramID == "sig0")
    return new ParamRegDouble("sig0",
                              "Surface density at 'rout'",
                              1.0E-2,
                              "g/cm^2");
  else if (paramID == "Tstar")
    return new ParamRegDouble("Tstar",
                              "Effective temperature of the star",
                              4000.0,
                              "K");
  else if (paramID == "Rstar")
    return new ParamRegDouble("Rstar",
                              "Radius of the central star",
                              2.0,
                              "Rsun");
  else if (paramID == "Mstar")
    return new ParamRegDouble("Mstar",
                              "Mass of the central star",
                              0.5,
                              "Msun");
  else if (paramID == "bgdens")
    return new ParamRegDouble("bgdens",
                              "Background number density (floor value for the disk density)",
                              1.0e-4, // 2.6e-7 1/cm^3
                              "1/cm^3");
  else
    throw invalid_argument("Invalid paramID " + paramID + " for modelID " + this->modelID);
}

#define ms 1.99e+33
#define rs 69600000000.0
#define au 14960000000000.0
#define pi 3.14159265358979323846264338328
#define gsize 1e-5
#define kk 1.3807e-16
#define hh 6.6262e-27
#define cc 29979245800.0
#define gg 6.674e-8 //
#define mp 1.6726e-24  // Proton mass	

//*****************************************************************************/
//*****************************************************************************/
const double get_kappav(){
    return 400.0;
}
//-----------------------------------------------------------------------------*/
const double get_eps(const double temp, const double beta){

    return (8.0*pi*gsize*kk*temp/(hh*cc))*beta;
}
//-----------------------------------------------------------------------------*/
const double get_ts(const double r, const double rstar, const double tstar){

    double tds_old, tds, diff, cconv, dummy_const, epsilon;

    tds_old = sqrt(rstar/r) * tstar;
    tds = 2000.0;
    diff = (tds_old - tds)/tds;
    cconv = 1e-2;

    dummy_const = sqrt(rstar/r/2.) * tstar;


    while(fabs(diff)>cconv){
        tds_old = tds;
        epsilon = get_eps(tds, 1.0);
        tds = dummy_const / sqrt(sqrt(epsilon));
        diff = (tds_old - tds) / tds;
    }

    return tds;
}
//-----------------------------------------------------------------------------*/
const double get_ti_thick(const double alpha, const double r, const double rstar, const double tstar){
    
    double ti;

    ti = sqrt(sqrt(alpha/4.)) * sqrt(rstar/r) * tstar;

    return ti;
    
}
//-----------------------------------------------------------------------------*/
const double get_ti_thincool(const double alpha, const double r, const double rstar, const double tstar, const double sigma){

    double ti_old, ti, diff, cconv, kappa_v, dummy_const, epsilon;

    ti_old = sqrt(rstar/r) * tstar;
    ti = 2000.0;
    diff = (ti_old-ti)/ ti;
    cconv = 1e-2;

    kappa_v = get_kappav();

    dummy_const = sqrt(sqrt(alpha/4./sigma/kappa_v)) *sqrt(rstar/r) * tstar;

    while(fabs(diff)>cconv){
        ti_old = ti;
        epsilon = get_eps(ti,1.0);
        ti = dummy_const / sqrt(sqrt(epsilon));
        diff = (ti_old - ti)/ti;
    }
    return ti;
}
//-----------------------------------------------------------------------------*/
double get_ti_thin(double alpha, double r, double rstar, double tstar, double tds){

    double ti_old, ti, diff, cconv, eps_s, dummy_const, epsilon;

    ti_old = sqrt(rstar/r) * tstar;
    ti = 2000.0;
    diff = (ti_old-ti)/ ti;
    cconv = 1e-2;

    eps_s = get_eps(tds,1.0);

    dummy_const = sqrt(sqrt(alpha*eps_s*eps_s)) *sqrt(rstar/r) * tstar;

    while(fabs(diff)>cconv){
        ti_old = ti;
        epsilon = get_eps(ti,1.0);
        ti = dummy_const / sqrt(sqrt(epsilon));
        diff = (ti_old - ti)/ti;
    }

    return ti;
}
//-----------------------------------------------------------------------------*/
void get_hpcg01(double r[], const double mstar, const double rstar, const double tstar, const double mu, const double hph, \
        double alpha[], double hpr[]){


    double tc, gamma[npgrid], ti[npgrid];
    int niter, ir;
    double ti_old, ti_new, diff, diff_max;


    tc = gg * mstar * mu * mp / kk / rstar;
    for(niter=0;niter<2;niter++){
        for(ir=0;ir<npgrid/2;ir++){
            if(ir==0){
                if (niter==0){
                    gamma[0] = 1.2;
                    gamma[1] = 1.2;
                } else{
                    gamma[2*ir] = 1.5 + 0.5 * log(ti[1]/ti[0]) / log(r[1]/r[0]);
                    gamma[2*ir+1] = gamma[2*ir];
                } 
            } else{
                gamma[2*ir] = 1.5 + 0.5 * log(ti[2*ir-1] / ti[2*ir-2]) / log(r[2*ir-1] / r[2*ir-2]);
                gamma[2*ir+1] = gamma[2*ir];

            }
            
        
            ti_old = 0.;
            ti_new = 10.;
            diff = fabs(ti_new - ti_old);
            diff_max = 0.05;

            while(diff>diff_max){
                ti_old = ti_new;
                ti_new = sqrt(sqrt( ((rstar / (3.*pi*r[2*ir])) + \
                        0.25 * (gamma[2*ir]-1.0) * hph * sqrt(ti_old/tc) * sqrt(r[2*ir]/rstar)) ))*\
                         sqrt(rstar/r[2*ir]) * tstar;
                diff = fabs(ti_new-ti_old);
            }
            ti[2*ir] = ti_new;

            ti_old = 0.; 
            ti_new = 10.;
            diff = fabs(ti_new - ti_old);
            diff_max = 0.05;
            
            while(diff>diff_max){
                ti_old = ti_new;
                ti_new = sqrt(sqrt( ((rstar / (3.*pi*r[2*ir+1])) + \
                        0.25 * (gamma[2*ir+1]-1.0) * hph * sqrt(ti_old/tc) * sqrt(r[2*ir+1]/rstar)) ))*\
                         sqrt(rstar/r[2*ir+1]) * tstar;
                diff = fabs(ti_new-ti_old);
            }
           ti[2*ir+1] = ti_new ;

           alpha[2*ir] = (4./3./pi) * rstar / r[2*ir] + (gamma[2*ir]-1.) * hph * sqrt(ti[2*ir]/tc) * \
                         sqrt(r[2*ir]/rstar);

           alpha[2*ir+1] = (4./3./pi) * rstar / r[2*ir+1] + (gamma[2*ir+1]-1.) * hph * sqrt(ti[2*ir+1]/tc) * \
                         sqrt(r[2*ir+1]/rstar);


           hpr[2*ir] = sqrt(ti[2*ir]/tc) * sqrt(r[2*ir] / rstar);
           hpr[2*ir+1] = sqrt(ti[2*ir+1]/tc) * sqrt(r[2*ir+1] / rstar);
        }
    }

}
//*****************************************************************************/

ModelCG97::ModelCG97()
{}

ModelCG97::~ModelCG97()
{}


void 
ModelCG97::onFinalizeConfiguration()
{
    int ir;

// I expect the input to be cgs and here I convert everything to SI for Lime
  m_rin    = m_paramDouble["rin"]*1.496e+11;
  m_rout   = m_paramDouble["rout"]*1.496e+11;
  m_plsig1 = m_paramDouble["plsig1"];
  m_hph    = m_paramDouble["hph"];
  m_sig0   = m_paramDouble["sig0"]*10.0; //
  m_mstar  = m_paramDouble["Mstar"]*1.99e+30;
  m_rstar  = m_paramDouble["Rstar"]*6.9600000e+8;
  m_tstar  = m_paramDouble["Tstar"];
  m_bgdens = m_paramDouble["bgdens"]*1e6;

    my_ri[0] = 0.;
    for (ir=0; ir<npgrid+1; ir++){
        my_ri[ir+1] = m_rin*0.9 * pow((m_rout*1.1/(m_rin*0.9)), float(ir) / float(npgrid-1)) * 1e2;
    }
    for (ir=0; ir<npgrid; ir++){
        my_r[ir] = sqrt(my_ri[ir+1] * my_ri[ir+2]);
    }

// The CG97 functions are written in cgs so use the cgs units for there
//    get_hpcg01(my_r, m_mstar, m_rstar, m_tstar, 2.3, m_hph, my_alpha, my_hp);
    get_hpcg01(my_r, m_mstar*1e3, m_rstar*1e2, m_tstar, 2.3, m_hph, my_alpha, my_hp);

    for (ir=0; ir<npgrid+1; ir++){
        my_ri[ir] = my_ri[ir]*1e-2;
    }

    for (ir=0; ir<npgrid; ir++){

        my_r[ir]  = my_r[ir]*1e-2;
        my_hp[ir] = my_hp[ir] * my_r[ir];
        my_ts[ir] = get_ts(my_r[ir]*1e2, m_rstar*1e2, m_tstar);
        my_ti[ir] = get_ti_thick(my_alpha[ir], my_r[ir]*1e2, m_rstar*1e2, m_tstar);
  
              

    }
}


const double
ModelCG97::t_dust(const double x, const double y, const double z)
{
  int id;
  double rc, temp;

  rc    = sqrt(x*x + y*y);

  if (rc > m_rin){
    if (rc < m_rout){
      id = gsl_interp_bsearch(my_ri, rc, 0, npgrid-1);

      if (fabs(z)>m_hph*my_hp[id]){
        temp = my_ts[id];
      } else {
        temp = my_ti[id];
      } 
    }
    else {
      temp   = 2.73;
    }    
  }
  else { 
    temp = 2.73;
  }

  return temp;
}



const double 
ModelCG97::density(const double x, const double y, const double z)
{
#define pi 3.14159265358979323846264338328
#define mmw 2.3       // Mean molecular weight
#define mpsi 1.6726e-27  // Proton mass	
  int id;
  double rc, rho;


  rc    = sqrt(x*x + y*y);

  if (rc > m_rin){
    if (rc < m_rout){
      id = gsl_interp_bsearch(my_ri, rc, 0, npgrid-1);
      rho = m_sig0 * pow((rc / m_rout), m_plsig1)/ (my_hp[id] * sqrt(2.*pi)) * exp(-0.5* (z / my_hp[id]) * (z/my_hp[id]));
    }
    else {
      rho   = 0.0;
    }    
  }
  else { 
    rho = 0.0;
  }
  
return (rho)/(mmw*mpsi) + m_bgdens;
}


const void
ModelCG97::velocity(const double x, const double y, const double z, double* v)
{
#define ggsi 6.674e-11 //
  
  double rc, vkep, sinp, cosp;

  rc   = sqrt(x*x + y*y);
  vkep = sqrt(ggsi*m_mstar/rc);
  sinp = y / (rc + 1e-90);
  cosp = x / (rc + 1e-90);
  
  v[0] = vkep * sinp;
  v[1] = -vkep * cosp;
  v[2] = 0.0e+0;

  
}




