////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Attila Juhasz
// Sterrewacht Leiden
// Universiteit Leiden
// juhasz@strw.leidenuniv.nl
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include "modelDDN01.h"
#include <stdexcept>
#include "gsl/gsl_spline.h"
#include "gsl/gsl_sf_erf.h"
#include <iostream>
#include <stdio.h>

string
ModelDDN01::modelID = "DDN01";

const ModelReg
ModelDDN01::getModelReg()
{
  ModelReg modelReg;
  modelReg.name = "DullemondDominik01";
  modelReg.description = "Passive irradiated circumstellar disk with an inner hole";
  modelReg.bibref = "Dullemond & Dominik 2001, ApJ, 560, 957";
  return modelReg;
}

const set<string>
ModelDDN01::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("rin");
  paramIDs.insert("rout");
  paramIDs.insert("mdisk");
  paramIDs.insert("plsig1");
  paramIDs.insert("Mstar");
  paramIDs.insert("Rstar");
  paramIDs.insert("Tstar");
  paramIDs.insert("bgdens");
  paramIDs.insert("dustopac_fname");
  return paramIDs;
}

const ParamReg*
ModelDDN01::getParamReg(string paramID)
{
  if (paramID == "rin")
    return new ParamRegDouble("rin",
                              "Inner Radius of the disk",
                              1.0, // AU
                              "AU");
  else if (paramID == "rout")
    return new ParamRegDouble("rout",
                              "Outer radius of the disk",
                              100.0,
                              "AU");
  else if (paramID == "plsig1")
    return new ParamRegDouble("plsig1",
                              "Power exponent of the radial surface density distribution",
                              -1.0,
                              "");
  else if (paramID == "mdisk")
    return new ParamRegDouble("mdisk",
                              "Total disk mass (gas + dust ; gas-to-dust ratio of 100 is used)",
                              0.01,
                              "Msun");
  else if (paramID == "Mstar")
    return new ParamRegDouble("Mstar",
                              "Mass of the central star",
                              1.0,
                              "Msun");
  else if (paramID == "Rstar")
    return new ParamRegDouble("Rstar",
                              "Radius of the central star",
                              1.0,
                              "Rsun");
  else if (paramID == "Tstar")
    return new ParamRegDouble("Mstar",
                              "Effective temperature of the central star",
                              4000.,
                              "K");
  else if (paramID == "bgdens")
    return new ParamRegDouble("bgdens",
                              "Background number density (floor value for the disk density)",
                              1.0e-4, // 2.6e-7 1/cm^3
                              "1/cm^3");
  
  else if (paramID == "dustopac_fname")
    return new ParamRegString("dustopac_fname",
                              "File name (with full path) containing dust opacity table in two column (wavelength [um], kappa_abs [cm^2/g]) format",
                              "jena_thin_e6.tab");
  else
    throw invalid_argument("Invalid paramID " + paramID + " for modelID " + this->modelID);
}



ModelDDN01::ModelDDN01()
{}

ModelDDN01::~ModelDDN01()
{}


///****************************************************************************/
///****************************************************************************/
#define ms 1.99e+33
#define rs 69600000000.0
#define au 14960000000000.0
#define pi 3.14159265358979323846264338328
#define gsize 1e-5
#define kk 1.3807e-16
#define hh 6.6262e-27
#define cc 29979245800.0
#define gg 6.674e-8 //
#define mp 1.6726e-24  // Proton mass	
#define ss 5.6703e-05 
#define mu 2.3
// ---------------------------------------------------------------------------------------------------------------------
//  Function to read the opacities
// --------------------------------------------------------------------------------------------------------------------- */
void read_kappa(const char *fname, double wav[], double kappa[]){
   
    double wmin=0.01, wmax=1e5; 
    FILE *rfile;
    char str[80];
    int nlines, i,  j;
    double *wav_tmp, *kappa_tmp, dum;
	gsl_spline *spline;
	gsl_interp_accel *acc=gsl_interp_accel_alloc();

    // Create the wavelength array */
    for (i=0; i<nwav; i++){
        wav[i] = wmin * pow(wmax/wmin, (i*1.0)/(1.0*nwav-1.));
    }
   
    // Read the opacity file */
    rfile = fopen(fname, "r");
    // Check the line numbers */
    nlines = 0;
    while(fgetc(rfile)!=EOF){
        fgets(str, 80, rfile);
        nlines++; 
    }

    rewind(rfile);
    wav_tmp = (double *)malloc(sizeof(double)*nlines);
    kappa_tmp = (double *)malloc(sizeof(double)*nlines);
    for (i=0;i<nlines;i++){
        fscanf(rfile, "%lf %lf\n", &wav_tmp[i], &kappa_tmp[i]);

        // In the Lime dust opacity files the wavelength should be in cm, so stick to that and convert that to micron */
        wav_tmp[i] = log10(wav_tmp[i]);
        kappa_tmp[i] = log10(kappa_tmp[i]);
    }
    fclose(rfile);

// Order the arrays in wavelength if necessary*/
    if(wav_tmp[1]<wav_tmp[0]){
        i=0;
        j=nlines-1;
        while(i<j){

            dum = wav_tmp[i];
            wav_tmp[i] = wav_tmp[j];
            wav_tmp[j] = dum;

            dum = kappa_tmp[i];
            kappa_tmp[i] = kappa_tmp[j];
            kappa_tmp[j] = dum;

            i++;
            j--;
        }
    }


//  OK, the file is read now check if there is wide enough wavelenght coverage for both the 
//   Heating and the cooling radiation fields */

    spline = gsl_spline_alloc(gsl_interp_cspline, nlines);
	gsl_spline_init(spline,wav_tmp, kappa_tmp, nlines);
   
    dum = (kappa_tmp[nlines-1] - kappa_tmp[nlines-2]) / (wav_tmp[nlines-1] - wav_tmp[nlines-2]);

    for(i=0;i<nwav;i++){
        if (log10(wav[i])<=wav_tmp[0]){
            kappa[i] = pow(10.,kappa_tmp[0]);
        } else if (log10(wav[i])>=wav_tmp[nlines-1]){
            kappa[i] = pow(10,kappa_tmp[nlines-1]) * pow(10., (log10(wav[i])-wav_tmp[nlines-1]) * dum); 
        } else {
            kappa[i] = pow(10., gsl_spline_eval(spline, log10(wav[i]), acc));
        }
                
    }
    free(wav_tmp);
    free(kappa_tmp);
}
// ---------------------------------------------------------------------------------------------------------------------
//  Simple integrator *
// --------------------------------------------------------------------------------------------------------------------- */
double integr(int n, double x[], double y[]){
    
    double dum;
    int i;

    // NOTE:
    // I use this function only to integrate in frequency and since all frequency-dependent quantities are ordered in 
    // wavelength instead of frequency the last term in the right-hand side is multiplied by -1
    dum = 0.0;
    for (i=0;i<n-1;i++){
        dum = dum + 0.5 * (y[i+1] + y[i]) * (x[i] - x[i+1]);
    } 

    return dum;
}


// Planck function */
void bnu_t(double nu[], double T, double bnut[]){
    int i;
    
    for(i=0;i<nwav;i++){
        bnut[i] = 2.0*hh*pow(nu[i],3)/cc/cc / (exp(hh*nu[i]/kk/T)-1.0);
    }


    return;
}

// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the ratio of the surface height over pressure scale height
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_chicg(double alpha, double sigma, double kappa_p){

    double chi_old, dum, threshold, chi, diff;

    chi_old = 0.;
    dum = 2.*alpha /sigma/kappa_p;
    threshold = 1e-4;
    chi = 4.0;
    diff = fabs(chi_old - chi)/chi;

    while(diff>threshold){  
        chi_old = chi;
        chi = sqrt(fabs(-2.0 * log(dum * exp(-0.5*chi_old*chi_old) / (1.0 - gsl_sf_erf(chi_old*sqrt(0.5))))));
        diff = fabs(chi_old-chi)/chi;
    }

    return chi;
}
// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the planck mean opacity
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_kappa_planck(double nu[], double kappa[], double T){

    double bb[nwav], dum[nwav];
    int i;

    bnu_t(nu, T, bb);

    for (i=0;i<nwav;i++){
        dum[i] = bb[i]*kappa[i];
    }
    //return integr(nwav, nu, dum) / integr(nwav, nu, bb);*/
    return integr(nwav, nu, dum) / (ss*pow(T,4)/pi);
}   

// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the planck mean opacity (from an imput frequency depencent flux array instead of the temperature only)
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_kappa_planck_Firr(double nu[], double kappa[], double Firr[]){

    double dum[nwav];
    int i;

    for (i=0;i<nwav;i++){
        dum[i] = Firr[i]*kappa[i];
    }

    return integr(nwav, nu, dum) / integr(nwav, nu, Firr); 
}   

// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the psi parameter for the disk interior
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_psii(double kappa[], double nu[], double Ti, double sigma){
    
    double bnut[nwav], fnu[nwav];
    int i;

    bnu_t(nu, Ti, bnut);
    for (i=0;i<nwav;i++){
        fnu[i] = bnut[i] * (1.0 - exp(-sigma * kappa[i]));
    }

    return integr(nwav, nu, fnu) / (ss*pow(Ti,4) / pi);
    //return integr(nwav, nu, fnu) / integr(nwav, nu, bnut);*/
}

// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the psi parameter for the disk surface layer
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_psis(double kappa[], double nu[], double Tsurf, double sigma){
    
    double bnut[nwav], fnu[nwav];
    int i;

    bnu_t(nu, Tsurf, bnut);
    for (i=0;i<nwav;i++){
        bnut[i] = bnut[i] * kappa[i];
        fnu[i] = bnut[i] * (1.0 - exp(-sigma * kappa[i]));
    }

    return integr(nwav, nu, fnu) / integr(nwav, nu, bnut);
}

// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the surface density at the outer radius from the disk mass
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_sig0(double rin, double rout, double mdisk, double plsig){

    double dum;

    if (plsig==-2.){
        dum = 2.0 * pi * pow(rout, -plsig) * (log(rout) - log(rin));
    } else {
        dum = 2.0 * pi * pow(rout, -plsig) / (2.0 + plsig) * (pow(rout, (2.+plsig)) - pow(rin, (2.+plsig)));
    }

    return mdisk/dum;
}
// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the disk interior temperature 
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_ti(double alpha, double r, double rstar, double tstar, double kappa[], double nu[], double sigma, double tsurf, double ti_init, double Firr){

    double psi_s, psi_i, ti, ti_old, diff, threshold;

    ti = ti_init; 
    psi_s = get_psis(kappa, nu, tsurf, sigma); 
    psi_i = get_psis(kappa, nu, ti, sigma); 

    ti_old = ti;
    ti = sqrt(sqrt(Firr * 0.5 * psi_s / (psi_i * ss)));
    diff = fabs(ti_old - ti) / ti;
    threshold = 0.01;

    while(diff>threshold){
        ti_old = ti;
        psi_i = get_psis(kappa, nu, ti, sigma); 
        ti = sqrt(sqrt(Firr * 0.5 * psi_s / (psi_i * ss)));
        diff = fabs(ti_old - ti) / ti;
    }

    return ti;
}

// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the rim temperature including self-irradiation
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_trim(double rstar, double rin, double mstar, double sigma, double kappa_p, double tstar){

    double chi, trim, trim_old, h, threshold, diff;

    chi = get_chicg(1.0, sigma, kappa_p*8.0);
    trim_old = 2000.;
    h = sqrt(kk*trim_old*pow(rin,3) / (mu*mp*gg*mstar));
    trim = sqrt(rstar/rin) * tstar * sqrt(sqrt((1. + h*chi/rin)));

    diff = fabs(trim_old -trim)/trim;
    threshold = 0.01;

    while(diff>threshold){
        trim_old = trim;
        h = sqrt(kk*trim_old*pow(rin,3) / (mu*mp*gg*mstar));
        trim = sqrt(rstar/rin) * tstar * sqrt(sqrt((1. + h*chi/rin)));
        diff = fabs(trim_old -trim)/trim;
    }


    return trim;
}

// ---------------------------------------------------------------------------------------------------------------------
 // Calculates the temperature in the surface layer
 // --------------------------------------------------------------------------------------------------------------------- ///
double get_tsurf(double r, double rstar, double tstar, double kappa[], double nu[], double Firr[]){

    double ts, ts_old, diff, threshold, c1, eps;

    ts_old = sqrt(rstar/r) * tstar;
    ts = 2000.;
    diff = fabs(ts_old - ts)/ts;
    threshold = 0.01;

    c1 = sqrt(rstar/r/2.) * tstar;


    while(diff>threshold){
        ts_old = ts;
        eps = get_kappa_planck(nu, kappa, ts) / get_kappa_planck_Firr(nu, kappa, Firr);
        ts = c1 /sqrt(sqrt(eps));
        diff = fabs(ts_old - ts) / ts;
        
    }
   return ts; 
}

// ---------------------------------------------------------------------------------------------------------------------
 // Function to calculate the disk structure  (r, Hs, Hp, ti, ts)
 // --------------------------------------------------------------------------------------------------------------------- ///
void get_disk_struct(double rin, double rdisk, double mdisk, double plsig1, const char *opac_fname, double mstar, double rstar, double tstar, 
    double r[], double ri[], double Hp[], double Hs[], double ti[], double ts[]){

    double wav[nwav], nu[nwav], kappa[nwav], Firr[nwav], Firr_int, Firr_star[nwav][nrmod], Firr_star_int[nrmod], bnut[nwav], dum[nwav];
    double alpha[nrmod], gamma[nrmod], chi[nrmod], sigma[nrmod];
    double chirim[nrmod], Hsrim[nrmod], Hprim[nrmod], trim_diff[nrmod], alpha_iter[nrmod], dum_arwav[nwav], bnut_rim[nwav]; 
    double sig0, kappa_planck_star, trim, tc, alpha_old, delta, theta, diff1, diff2;
    int ir,inu,iiter, iish, iicg;
    int cg97_converged, converged;

    // Read the opacities */ 
    read_kappa(opac_fname, wav, kappa);
    for (inu=0; inu<nwav; inu++){
        // The wavelength should be in micron */
        nu[inu] = cc/wav[inu]*1e4;
    }
    
    // Set up radial grid */
    sig0 = get_sig0(rin, rdisk, mdisk, plsig1)*0.01;

    for (ir=0; ir<nrmod+1; ir++){
        ri[ir] = rin*0.9 * pow((rdisk*1.1/(rin*0.9)), float(ir) / float(nrmod-1));
    }
    for (ir=0; ir<nrmod; ir++){
        r[ir] = sqrt(ri[ir] * ri[ir+1]);
    }
    for (ir=0; ir<nrmod+1; ir++){
        ri[ir] = ri[ir] * 1e-2; 
    }

    for (ir=0; ir<nrmod; ir++){
        r[ir] = rin * pow(rdisk/rin, (1.0*ir) / (1.0*nrmod -1.));
        sigma[ir] = sig0 * pow(r[ir]/rdisk, plsig1);
        alpha[ir] = 0.05;
    }
    

    kappa_planck_star = get_kappa_planck(nu, kappa, tstar);

    trim = get_trim(rstar, rin, mstar, sigma[0], kappa_planck_star, tstar);
    Hprim[0] = sqrt(kk*trim*pow(rin,3) / (gg*mstar*mu*mp));

    for (ir=0;ir<nrmod;ir++){
        chirim[ir]    = get_chicg(1.0, sigma[ir], kappa_planck_star*8.0);
        Hsrim[ir]     = chirim[0] * Hprim[0] - (r[ir] - r[0])/8.0;
        Hsrim[ir]     = (Hsrim[ir] + fabs(Hsrim[ir]))*0.5;
        Hprim[ir]     = Hsrim[ir]/chirim[ir];
        trim_diff[ir] = trim * (Hsrim[ir]/Hsrim[0]) * (Hsrim[ir]/Hsrim[0]);
        if (Hsrim[ir]>0)
            iish = ir+1;
        ti[ir] = 100.;
    }

    
// First go for a CG97 disk */
    tc = gg*mstar*mu*mp/kk/rstar;

    gamma[0] = 1.28;
    gamma[1] = 1.28;
    alpha[0] = 0.05;


    for (ir=0; ir<nrmod; ir++){
        bnu_t(nu, tstar, bnut);
        for (inu=0; inu<nwav; inu++){
            Firr_star[inu][ir] = bnut[inu] * pi * pow((rstar/r[ir]), 2);
            dum[inu] = Firr_star[inu][ir] * alpha[ir];
        }
        Firr_star_int[ir] = integr(nwav, nu, dum);
    }


    iiter=0;
    cg97_converged = 0;

    while(cg97_converged!=1){
        iiter++;
        diff2 = 0.;
        for (ir=0; ir<nrmod; ir++){

            alpha_iter[ir] = alpha[ir];
            converged = 0;

            if ((ir%2)==0){
                if (ir>1){
                    gamma[ir] = (r[ir-1] + r[ir-2]) / (Hs[ir-1] + Hs[ir-2]) * (Hs[ir-2] - Hs[ir-1]) / (r[ir-2] - r[ir-1]);
                    gamma[ir+1] = gamma[ir];
                }
            }
            if (ir==0){
                if (iiter>1){
                    gamma[ir]   = (r[ir+2] + r[ir+3]) / (Hs[ir+2] + Hs[ir+3]) * (Hs[ir+3] - Hs[ir+2]) / (r[ir+3] - r[ir+2]);
                    gamma[ir+1] = gamma[ir];
                }
            }

            for (inu=0;inu<nwav;inu++){
                dum_arwav[inu] = Firr_star[inu][ir];
            }

            while (converged!=1){
                alpha_old = alpha[ir];

                ts[ir] = get_tsurf(r[ir], rstar, tstar, kappa, nu, dum_arwav);
                ti[ir] = sqrt(sqrt(0.5*Firr_star_int[ir])) * trim;
                ti[ir] = get_ti(alpha[ir], r[ir], rstar, tstar, kappa, nu, sigma[ir], ts[ir], ti[ir], Firr_star_int[ir]);
                chi[ir] = get_chicg(alpha[ir], sigma[ir], kappa_planck_star);
                Hp[ir] = sqrt(ti[ir]/tc) * sqrt(r[ir]/rstar) * r[ir];
                Hs[ir] = Hp[ir] * chi[ir];
                

                alpha[ir] = (4./3./pi) * rstar / r[ir] + (gamma[ir]-1.) * chi[ir] * sqrt(ti[ir]/tc) * sqrt(r[ir]/rstar);
                //alpha[ir] = (gamma[ir]-1.) * Hs[ir]/r[ir];

                // Check the convergence criterion for alpha at r[ir] */
                diff1 = fabs((alpha[ir] - alpha_old) / alpha[ir]);
                if (diff1<0.001){
                    converged = 1;
                }
                // Check the deviation of alpha[ir] in this iteration compared to the previous big iteration */
                diff1 = fabs((alpha[ir] - alpha_iter[ir])/alpha[ir]);
                if (diff1>diff2){
                    diff2 = diff1;
                }

            }
        if (diff2<0.001)
            cg97_converged=1;
        }
    }
    
    // Now we have a CG97 disk, let's include the rim and the self-irradiation */
    
    bnu_t(nu, trim, bnut_rim);

    for (ir=iish;ir<nrmod;ir++){
        converged = 0;
        delta = -1.;
        theta = -1.;

        while (converged!=1){
            delta = (r[ir] / rin - 1.0) / (Hs[ir] / Hsrim[0] - 1.);
            theta = atan( (r[ir]-rin) / (Hs[ir] - Hsrim[0]) );
                
            for (inu=0;inu<nwav;inu++){
                 Firr[inu] = 0.;
            }

            if ((Hs[ir]/r[ir])<(Hsrim[0]/r[0])){
                iicg = ir+1;
            } else {
                for (inu=0;inu<nwav;inu++){
                    Firr[inu] = Firr_star[inu][ir];
                }
            }

            alpha_old = alpha[ir];
            if(delta>=0){
                if (delta<1.){
                    for (inu=0;inu<nwav;inu++){
                        Firr[inu] = Firr[inu] + 2. /pi * bnut_rim[inu] * pow(rin/r[ir], 2) * \
                                    cos(theta) * (delta * sqrt(1.0 - delta*delta) + asin(delta));
                        dum[inu] = Firr[inu] * alpha[ir];
                    }
                } else {
                    for (inu=0;inu<nwav;inu++){
                        Firr[inu] = Firr[inu] + bnut_rim[inu] * pow(rin/r[ir], 2) * cos(theta);
                        dum[inu] = Firr[inu] * alpha[ir];
                    }
                }
                    
                Firr_int = integr(nwav, nu, dum);
                
                ts[ir] = get_tsurf(r[ir], rstar, tstar, kappa, nu, Firr);
                ti[ir] = sqrt(sqrt(0.5*Firr_int)) * trim;
                ti[ir] = get_ti(alpha[ir], r[ir], rstar, tstar, kappa, nu, sigma[ir], ts[ir], ti[ir], Firr_int);
                chi[ir] = get_chicg(alpha[ir], sigma[ir], kappa_planck_star);
                Hp[ir] = sqrt(ti[ir]/tc) * sqrt(r[ir]/rstar) * r[ir];
                Hs[ir] = Hp[ir] * chi[ir];


                alpha[ir] = (gamma[ir]-1.) * Hs[ir]/r[ir];
                diff1 = fabs( (alpha[ir]-alpha_old) / alpha[ir]);
                if (diff1<1e-4)
                    converged = 1;
            } else {
                Hs[ir] = 0.;
                Hp[ir] = 0.;
                ti[ir] = 0.;
                ts[ir] = 0.;
                chi[ir] = 0.;
                iish = ir-1;
                converged = 1;
            }
        }

    }

    // Now we have a CG97 disk + puffed-up rim + self-irradiation, let's glue them together */

    for (ir=0;ir<iish+1;ir++){
        Hs[ir] = Hsrim[ir];
        Hp[ir] = Hprim[ir];
        chi[ir] = chirim[ir];
        ti[ir] = Hprim[ir]*Hprim[ir] * tc * rstar/pow(r[ir], 3);
        ts[ir] = ti[ir];
    }

    if (iish<iicg){
        if (iicg>0.){
            for (ir=iish;ir<iicg;ir++){
                diff1 = ( (Hs[ir+1]/r[ir+1]) - ( log(r[ir+1]) - log(r[ir]) ) / 8.) * r[ir];
                Hs[ir] = pow((pow(diff1, 8) + pow(Hs[ir],8)), 1./8.);
                chi[ir] = get_chicg(alpha[ir], sigma[ir], kappa_planck_star);
                Hp[ir] = Hs[ir] / chi[ir];
                ti[ir] = Hp[ir]*Hp[ir] * (gg*mstar*mu*mp) / kk / pow(r[ir],3);
                if (ti[ir]>ts[ir])
                    ts[ir] = ti[ir];
            }
         }
    }


    for (ir=0; ir<nrmod; ir++){
        if (Hs[ir]/r[ir]<=0.04){
            Hs[ir] = 0.04*r[ir];
            Hp[ir] = 0.04*r[ir]/4.;
            ti[ir] = Hp[ir]*Hp[ir] * (gg*mstar*mu*mp) / kk / pow(r[ir],3);
            ts[ir] = ti[ir];
        }
        Hs[ir] = Hs[ir] * 1e-2;
        Hp[ir] = Hp[ir] * 1e-2;

    }

} 


//*****************************************************************************/
//*****************************************************************************/
void 
ModelDDN01::onFinalizeConfiguration()
{

// I expect the input to be cgs and here I convert everything to SI for Lime
  m_rin    = m_paramDouble["rin"]*1.496e+11;
  m_rout   = m_paramDouble["rout"]*1.496e+11;
  m_mdisk  = m_paramDouble["mdisk"]*1.99e+30;
  m_plsig1 = m_paramDouble["plsig1"];
  m_mstar  = m_paramDouble["Mstar"]*1.99e+30;
  m_rstar  = m_paramDouble["Rstar"]*6.96e+8;
  m_tstar  = m_paramDouble["Tstar"];
  m_bgdens = m_paramDouble["bgdens"]*1e6;
  m_dustopac_fname = (m_paramString["dustopac_fname"]).c_str();
  

//  kappa_fname = "testopac_micron.dat";
//  read_kappa(m_dustopac_fname, wav, kappa);

  my_sig0 = get_sig0(m_rin, m_rout, m_mdisk, m_plsig1);
  get_disk_struct(m_rin*1e2, m_rout*1e2, m_mdisk*1e3, m_plsig1, m_dustopac_fname, m_mstar*1e3, m_rstar*1e2, m_tstar, my_r, my_ri, my_hp, my_hs, my_ti, my_ts);
}


const double
ModelDDN01::t_dust(const double x, const double y, const double z)
{
  int id;
  double rc, temp;

  rc    = sqrt(x*x + y*y);

  if (rc > m_rin){
    if (rc < m_rout){
      id = gsl_interp_bsearch(my_ri, rc, 0, nrmod-1);

      if (fabs(z)>my_hs[id]){
        temp = my_ts[id];
      } else {
        temp = my_ti[id];
      } 
    }
    else {
      temp   = 2.73;
    }    
  }
  else { 
    temp = 2.73;
  }

  return temp;
}
const double
ModelDDN01::density(const double x, const double y, const double z)
{
#define pi 3.14159265358979323846264338328
#define mmw 2.3       // Mean molecular weight
#define mpsi 1.6726e-27  // Proton mass	
  int id;
  double rc, rho;


  rc    = sqrt(x*x + y*y);

  if (rc > m_rin){
    if (rc < m_rout){
      id = gsl_interp_bsearch(my_ri, rc, 0, nrmod-1);
      rho = my_sig0 * pow((rc/m_rout), m_plsig1)/ (my_hp[id] * sqrt(2.*pi)) * exp(-0.5* (z / my_hp[id]) * (z/my_hp[id])) ;
    }
    else {
      rho   = 0.0;
    }    
  }
  else { 
    rho = 0.0;
  }

  return (rho)/(mmw*mpsi) + m_bgdens;
}
const void
ModelDDN01::velocity(const double x, const double y, const double z, double* v)
{
#define ggsi 6.674e-11 //

  double rc, vkep, sinp, cosp;

  rc   = sqrt(x*x + y*y);
  vkep = sqrt(ggsi*m_mstar/rc);
  sinp = y / (rc + 1e-90);
  cosp = x / (rc + 1e-90);
  
  v[0] = vkep * sinp;
  v[1] = -vkep * cosp;
  v[2] = 0.0e+0;

  
}




