////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Wouter Vlemmings
// Chalmers University of Technology
// Onsala Space Observatory
// wouter.vlemmings.at.chalmers.se	
//
// and
// Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <stdexcept>
// #include <iostream>

#include <math.h>

#include "modelMamon88.h"
#include "modelMamonInterpol.h" // For interpolation of tabulated values

using namespace std;


// Physical constants:
#define PI      3.14159265358979323846
#define AMU     1.6605402e-27 // kg
#define AU		1.49598e11 // m

ModelMamon88::ModelMamon88()
{}

ModelMamon88::~ModelMamon88()
{}


string
ModelMamon88::modelID = "Mamon88";

const ModelReg
ModelMamon88::getModelReg()
{
  ModelReg modelReg;

  modelReg.name = "Mamon et al. 1988";
  modelReg.description = "Expanding circumstellar envelope";
  modelReg.bibref = "Mamon et al. 1988, ApJ 328, 797";

  return modelReg;
}

const set<string>
ModelMamon88::getParamIDs()
{
  set<string> paramIDs;

  // ADJUST: Insert IDs for all needed parameters
  paramIDs.insert("mdot");
  paramIDs.insert("rin");
  paramIDs.insert("ve");
  paramIDs.insert("tin");
  paramIDs.insert("ab0");

  return paramIDs;
}

const ParamReg*
ModelMamon88::getParamReg(string paramID)
{
  // ADJUST: Create and return a new ParamReg object of the correct type
  //         for each paramID specified in getParamIDs
  if (paramID == "mdot")
    return new ParamRegDouble("Mdot",
                              "Mass loss rate (10^-8 .. 10^-4 Msun/yr)",
                              1.e-5,
                              "Msun/year");
  if (paramID == "rin")
    return new ParamRegDouble("Rin",
                           	  "Inner radius",
                           	  2.,
                           	  "AU");
  if (paramID == "ve")
    return new ParamRegDouble("ve",
                           	  "Expansion velocity (7.5 .. 30 km/s)",
                           	  15.,
                           	  "km/s");
  if (paramID == "tin")
    return new ParamRegDouble("Tin",
                           	  "Temperature at inner edge",
                           	  1100.,
                           	  "K");
  if (paramID == "ab0")
    return new ParamRegDouble("ab0",
                           	  "Abundance at inner edge",
                           	  4.e-4,
                           	  "");

  throw invalid_argument("Invalid paramID " + paramID + " for modelID " + this->modelID);
}


void
ModelMamon88::onFinalizeConfiguration()
{
	m_mdot = m_paramDouble["mdot"] * 6.3102486e+22; // Msun/yr to kg/s
	m_rin  = m_paramDouble["rin"]  * AU;
	m_ve   = m_paramDouble["ve"]  * 1.e3;
	m_tin  = m_paramDouble["tin"];
	m_ab0  = m_paramDouble["ab0"];

	// Alpha and r1/2 are interpolated from table 3 of Mamon et al. 1988
	// The interpolation routines take mdot in Msun/yr and ve in km/s
	m_alpha = getAlpha(m_paramDouble["mdot"], m_paramDouble["ve"]);
	m_rhalf= getRhalf(m_paramDouble["mdot"], m_paramDouble["ve"]);

	// cout << "RS_DEBUG: alpha = " << m_alpha << " r1/2 = " << m_rhalf << endl;
}

const double
ModelMamon88::density(const double x, const double y, const double z)
{
  double r = sqrt(x*x + y*y + z*z);

  double density = 0.;
  if(r > m_rin){
	density = m_mdot / (4.0 * PI * r * r * m_ve);
  }

  // convert to number density:
  // TODO: set correct molar mass for each collision partner
  density /= 2.0 * AMU;

  return density;
}

const double
ModelMamon88::temperature(const double x, const double y, const double z)
{
  double r =sqrt(x*x+y*y+z*z);

  double temperature = 0.;
  if(r > m_rin && r<6000 * AU){
	temperature = m_tin * pow(r / m_rin, -0.54);
  }

  if(r > 6000 * AU){
	temperature = m_tin * pow((6000. * AU) / m_rin, -0.54) * pow(r / (6000 * AU), -0.72);
  }

  return temperature;
}

const double
ModelMamon88::abundance(const double x, const double y, const double z)
{
  double r = sqrt(x*x+y*y+z*z);

  double t1 = pow((r / m_rhalf), m_alpha);
  return m_ab0 * exp(-1 * log(2) * t1);
}

const void
ModelMamon88::velocity(const double x, const double y, const double z, double* v)
{
  double theta=atan2(sqrt(x*x+y*y),z);
  double phi=atan2(y,x);

  //Vector transformation back into Cartesian basis
  v[0] = m_ve*sin(theta)*cos(phi);
  v[1] = m_ve*sin(theta)*sin(phi);
  v[2] = m_ve*cos(theta);
}


// Methods to determine values for alpha and r1/2 by interpolation
// of values from table 3 of Mamon et al. 1988

