////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <math.h>
#include "modelMamonInterpol.h"

using namespace std;


const double
getAlpha(const double mdot, const double ve)
{
	// mdot in Msun/yr
	// ve in km/s !!!

	// Interpolation is done for alpha
	// on logMdot-ve grid

	double alphaTable[13][3] =
           {{1.71, 1.39, 1.20},
			{1.81, 1.46, 1.23},
			{1.96, 1.60, 1.31},
			{2.09, 1.74, 1.39},
			{2.22, 1.89, 1.51},
			{2.38, 2.09, 1.71},
			{2.51, 2.24, 1.88},
			{2.66, 2.39, 2.05},
			{2.90, 2.61, 2.29},
			{3.07, 2.79, 2.47},
			{3.26, 2.96, 2.66},
			{3.51, 3.20, 2.89},
			{3.71, 3.39, 3.07}
	};

	vector< vector<double> > vAlphaTable;
	vAlphaTable.resize(13, vector<double>(3, 0.));

	for (int i=0; i<13; i++)
		for (int j=0; j<3; j++)
			vAlphaTable[i][j] = alphaTable[i][j];

	return bilinearInterpol(log10(mdot), ve , logMdotTable(), veTable(), vAlphaTable);
}

const double
getRhalf(const double mdot, const double ve)
{
	// mdot in Msun/yr
	// ve in km/s !!!

	// Unit of output is m

	// Interpolation is done for log(r1/2)
	// on logMdot-ve grid

	double rhalfTable[13][3] =
		   {{7.50e15, 9.01e15, 1.39e16},
			{9.79e15, 1.05e16, 1.48e16},
			{1.49e16, 1.40e16, 1.71e16},
			{2.12e16, 1.85e16, 2.01e16},
			{3.10e16, 2.54e16, 2.49e16},
			{5.23e16, 4.05e16, 3.55e16},
			{7.91e16, 5.95e16, 4.88e16},
			{1.21e17, 8.88e16, 6.94e16},
			{2.14e17, 1.54e17, 1.15e17},
			{3.35e17, 2.35e17, 1.72e17},
			{5.31e17, 3.65e17, 2.61e17},
			{9.99e17, 6.67e17, 4.63e17},
			{1.64e18, 1.07e18, 7.26e17}
	};
	// Unit is cm

	vector< vector<double> > logRhalfTable;
	logRhalfTable.resize(13, vector<double>(3, 0.));

	for (int i=0; i<13; i++)
		for (int j=0; j<3; j++) {
			logRhalfTable[i][j] = log10(rhalfTable[i][j] / 100.); // Convert to m
		}

	double logRhalf = bilinearInterpol(log10(mdot), ve , logMdotTable(), veTable(), logRhalfTable);
	return pow(10., logRhalf);
}

const vector<double>
logMdotTable()
{
	// Values of Mdot as in table 3 of Mamon et al. 1988
	// log(Mdot / (Msun/yr)) is returned

	double mdots[13] =
	{1.e-8, 2.e-8, 5.e-8, 1.e-7, 2.e-7, 5.e-7, 1.e-6, 2.e-6, 5.e-6, 1.e-5, 2.e-5, 5.e-5, 1.e-4};
	// Unit is Msun/year

	vector<double> logMdots(13, 0.);
	for (int i=0; i<13; i++)
		logMdots[i] = log10(mdots[i]);

	return logMdots;
}

const vector<double>
veTable()
{
	// Values of Mve as in table 3 of Mamon et al. 1988
	// ve / (km/s) is returned

	vector<double> ves(3, 0.);
	ves[0] = 7.5;  // km/s
	ves[1] = 15.;
	ves[2] = 30.;

	return ves;
}



const double
bilinearInterpol(const double& x,
		         const double& y,
		         const vector<double>& xgrid,
		         const vector<double>& ygrid,
		         const vector< vector<double> > zgrid)
{
	// Do a bi-linear interpolation to determine the value z at (x, y)
	// with zgrid given at (xgrid, ygrid)
	//
	// For values of (x, y) outside the grid, an extrapolation is
	// performed with the interpolation parameters of the related
	// grid cell.

	// Find i with xgrid[i] <= x < xgrid[i+1]:
	int i = -1;
	int imax = xgrid.size() - 1;
	while (i < imax && x >= xgrid[i+1])
		i++;

	// Same with j and y:
	int j = -1;
	int jmax = ygrid.size() - 1;
	while (j < jmax && y >= ygrid[j+1])
		j++;

	// Define grid rectangle:
	double il, iu;
	if (i < 0) {
		// Extrapolation:
		il = 0;
		iu = 1;
	} else if (i == imax) {
		// Extrapolation:
		il = imax - 1;
		iu = imax;
	} else {
		// Interpolation:
		il = i;
		iu = i + 1;
	}

	double jl, ju;
	if (j < 0) {
		// Extrapolation:
		jl = 0;
		ju = 1;
	} else if (j == jmax) {
		// Extrapolation:
		jl = jmax - 1;
		ju = jmax;
	} else {
		// Interpolation:
		jl = j;
		ju = j + 1;
	}

	// Interpolate or extrapolate:
	double t = (x - xgrid[il]) / (xgrid[iu] - xgrid[il]);
	double u = (y - ygrid[jl]) / (ygrid[ju] - ygrid[jl]);

	double z =   (1. - t) * (1. - u) * zgrid[il][jl]
	           + t * (1. - u) * zgrid[iu][jl]
	           + t * u * zgrid[iu][ju]
	           + (1. - t) * u * zgrid[il][ju];
	return z;
}

// For tests:

#include <fstream>

int main()
{

	ofstream alphaFile("alpha.dat", ios::out);
	ofstream rhalfFile("rhalf.dat", ios::out);

	for (double ve=0.; ve<=50; ve+=0.5) {
		for (double logMdot=-12.; logMdot<=-2; logMdot+=0.3) {
			double alpha = getAlpha(pow(10., logMdot), ve);
			double rhalf = getRhalf(pow(10., logMdot), ve);
			alphaFile << logMdot  << "\t" << ve << "\t" << alpha << endl;
			rhalfFile << logMdot  << "\t" << ve << "\t" << log10(rhalf) << endl;
		}
		alphaFile << endl;
		rhalfFile << endl;
	}
	return 0;
}




