////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Attila Juhasz
// Sterrewacht Leiden
// Universiteit Leiden
// juhasz@strw.leidenuniv.nl
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include "modelShu77.h"
#include <math.h>
#include <gsl/gsl_interp.h>

#include <stdexcept>
#include "modelShu77TabulatedSolution.h"
string
ModelShu77::modelID = "Shu77";

const ModelReg
ModelShu77::getModelReg()
{
  ModelReg modelReg;
  modelReg.name = "Shu 1977";
  modelReg.description =  "Expansion-wave collapse solution of hydrostatic singular spheres Shu 1977";
  modelReg.bibref = "Shu 1977, ApJ, 214, 488";

  return modelReg;
}

const set<string>
ModelShu77::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("T");
  paramIDs.insert("time");
  return paramIDs;
}

const ParamReg*
ModelShu77::getParamReg(string paramID)
{
  if (paramID == "T")
    return new ParamRegDouble("T",
                              "Temperature of the cloud",
                              30.,
                              "K");
  else if (paramID == "time")
    return new ParamRegDouble("time",
                              "Time at which the density and velocity field should be calculated",
                              1e+5,
                              "year");
  else
    throw invalid_argument("Invalid paramID " + paramID + " for modelID " + this->modelID);
}



ModelShu77::ModelShu77()
{}

ModelShu77::~ModelShu77()
{}


void 
ModelShu77::onFinalizeConfiguration()
{
  m_T = m_paramDouble["T"];
  m_time = m_paramDouble["time"]*31536000e+0;
}

/******************************************************************************/
const double 
ModelShu77::density(const double x, const double y, const double z)
{
#define pi 3.14159265358979323846264338328e0
#define gg_si 6.674e-11 //
#define mp_si 1.6726e-27  // Proton mass	

  double mp = 1.6726000e-27; // Proton mass
  double kk =1.3807e-23; // Boltzman constant
  double r, a, rc, mod_x, dlx, dly, alpha;
  unsigned long id ;
  double rho;

  r = sqrt(x*x + y*y + z*z);
  a  = sqrt(kk*m_T/(2.3*mp));
  rc = a*m_time;
 
  // Hydrostatic outer envelope
  if (r>=rc){
    mod_x = r / (a*m_time);
	alpha = 2e0 / (mod_x * mod_x);
    rho = alpha * (a*a) / (2e0 * pi* gg_si) / (r*r)/(2.3*mp_si);
  }
 // Collapse solution  
  else if(r>=rc*0.05){

  	mod_x = r / (a*m_time);
// 	hunt(stab_x, 20, mod_x, &id);
    id = gsl_interp_bsearch(stab_x, mod_x, 0, 19);
	dlx = log10(stab_x[id+1]/stab_x[id]);
	dly =  log10(stab_alpha[id+1]/stab_alpha[id]);
	alpha = pow(10., dly/dlx * log10(mod_x/stab_x[id])) * stab_alpha[id]; 

 	rho=  alpha/ ((4e0 * pi * gg_si * m_time*m_time)*(2.3*mp_si));
  	}
  // Polynomial extrapolation inwards of 0.05 * Rcrit, which is the innermost
  // point in Shu's table that has a finite alpha and can therefore be used for
  // interpolation/extrapolation
  else{
  	mod_x = r / (a*m_time);
	dlx = log10(stab_x[1]/stab_x[0]);
	dly =  log10(stab_alpha[1]/stab_alpha[0]);
	alpha = pow(10., dly/dlx * log10(mod_x/stab_x[0])) * stab_alpha[0]; 
	rho   = alpha/ ((4e0 * pi * gg_si * m_time*m_time)*(2.3*mp_si));
  }
  return rho;
}

const double
ModelShu77::temperature(const double x, const double y, const double z)
{
  return m_T; 
}

const double
ModelShu77::t_dust(const double x, const double y, const double z)
{
  return temperature(x, y, z);
}

const void
ModelShu77::velocity(const double x, const double y, const double z, double* v)
{
#define pi 3.14159265358979323846264338328e0
  double mp = 1.6726000e-27; // Proton mass
  double kk = 1.3807e-23; // Boltzman constant
  
  double r, a, rc, mod_x, dlx, dly, sinp, sint, cosp, cost,rcyl, vminus , vr;
  unsigned long id; 
  r = sqrt(x*x + y*y + z*z);
// sound speed

  a  = sqrt(kk*m_T/(2.3*mp));
  rc = a*m_time;

  // Hydrostatic outer envelope
  if (r>=rc){
  	vr = 0.0;
  } 
 // Collapse solution  
  else if(r>=rc*0.05){
  	mod_x = r / (a*m_time);
 	
//	hunt(stab_x, 21, mod_x, &id);
    id = gsl_interp_bsearch(stab_x, mod_x, 0, 20);

	dlx = log10(stab_x[id+1]/stab_x[id]);
	dly =  log10(stab_v[id+1]/stab_v[id]);
	vminus = pow(10e0, dly/dlx * log10(mod_x/stab_x[id])) * stab_v[id]; 
 	vr =  -a * vminus;
  }
  // Polynomial extrapolation inwards of 0.05 * Rcrit, which is the innermost
  // point in Shu's table that has a finite alpha and can therefore be used for
  // interpolation/extrapolation
  else{
  	mod_x = r / (a*m_time);
	dlx = log10(stab_x[1]/stab_x[0]);
	dly =  log10(stab_v[1]/stab_v[0]);
	vminus = pow(10., dly/dlx * log10(mod_x/stab_x[0])) * stab_v[0]; 
	vr  = -a*vminus;
  }
  
  rcyl = sqrt(x*x + y*y);

  sinp = y / (rcyl+1e-90);
  cosp = x / (rcyl+1e-90);
  sint = rcyl / (r+1e-90);
  cost = z    / (r+1e-90);

  v[0] = sint*cosp*vr;
  v[1] = sint*sinp*vr;
  v[2] = cost*vr;

}


