////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <stdexcept>
#include <set>
#include <math.h>
#include <sstream>


#include "modelTabdata.h"
#include "artistTabdataReader.h"

#include "utilities.h"


ModelTabdata::ModelTabdata()
: m_pDensityReader(NULL),
  m_pTemperatureReader(NULL),
  m_pAbundanceReader(NULL),
  m_pTdustReader(NULL),
  m_pDopplerReader(NULL),
  m_pVelocityReader(NULL),
  m_pBmagReader(NULL)
{}

ModelTabdata::~ModelTabdata()
{
	delete m_pDensityReader;
	delete m_pTemperatureReader;
	delete m_pAbundanceReader;
	delete m_pTdustReader;
	delete m_pDopplerReader;
	delete m_pVelocityReader;
	delete m_pBmagReader;
}

string
ModelTabdata::modelID = "Tabdata";

const ModelReg
ModelTabdata::getModelReg()
{
  ModelReg modelReg;

  modelReg.name = "Artist Tabdata Model";
  modelReg.description = "Model from ARTISTS's tabulated data files";
  modelReg.bibref = "";

  return modelReg;
}

const set<string>
ModelTabdata::getParamIDs()
{
  set<string> paramIDs;

  paramIDs.insert("densityFilename");
  paramIDs.insert("temperatureFilename");
  paramIDs.insert("abundanceFilename");
  paramIDs.insert("tdustFilename");
  paramIDs.insert("dopplerFilename");
  paramIDs.insert("velocityFilename");
  paramIDs.insert("bmagFilename");

  return paramIDs;
}

const ParamReg*
ModelTabdata::getParamReg(string paramID)
{
  if (paramID == "densityFilename")
    return new ParamRegString("densityFilename",
                              "Full path to the Tabdata file with density data",
                              "");
  if (paramID == "temperatureFilename")
    return new ParamRegString("temperatureFilename",
                              "Full path to the Tabdata file with temperature data",
                              "");
  if (paramID == "abundanceFilename")
    return new ParamRegString("abundanceFilename",
                              "Full path to the Tabdata file with abundance data",
                              "");
  if (paramID == "tdustFilename")
    return new ParamRegString("tdustFilename",
                              "Full path to the Tabdata file with tdust data",
                              "");
  if (paramID == "dopplerFilename")
    return new ParamRegString("dopplerFilename",
                              "Full path to the Tabdata file with doppler data",
                              "");
  if (paramID == "velocityFilename")
    return new ParamRegString("velocityFilename",
                              "Full path to the Tabdata file with velocity data",
                              "");
  if (paramID == "bmagFilename")
    return new ParamRegString("bmagFilename",
                              "Full path to the Tabdata file with bmag data",
                              "");
  else
    throw invalid_argument("Invalid paramID " + paramID + " for modelID " + this->modelID);
}


void
ModelTabdata::onFinalizeConfiguration()
{
	string& inputFilename = m_paramString["densityFilename"];
	if ( strip(inputFilename).size() ) {
		m_pDensityReader = new ScalarTabdataReader(inputFilename);
		m_pDensityReader->openFile();
		m_pDensityReader->parseFile();
		m_pDensityReader->closeFile();
		if ( m_pDensityReader->getDataname() != "density" )
			throw domain_error("Data file " + inputFilename + " does not contain density data");
	}

	inputFilename = m_paramString["temperatureFilename"];
	if ( strip(inputFilename).size() ) {
		m_pTemperatureReader = new ScalarTabdataReader(inputFilename);
		m_pTemperatureReader->openFile();
		m_pTemperatureReader->parseFile();
		m_pTemperatureReader->closeFile();
		if ( m_pTemperatureReader->getDataname() != "temperature" )
			throw domain_error("Data file " + inputFilename + " does not contain temperature data");
	}

	inputFilename = m_paramString["abundanceFilename"];
	if ( strip(inputFilename).size() ) {
		m_pAbundanceReader = new ScalarTabdataReader(inputFilename);
		m_pAbundanceReader->openFile();
		m_pAbundanceReader->parseFile();
		m_pAbundanceReader->closeFile();
		if ( m_pAbundanceReader->getDataname() != "abundance" )
			throw domain_error("Data file " + inputFilename + " does not contain abundance data");
	}

	inputFilename = m_paramString["tdustFilename"];
	if ( strip(inputFilename).size() ) {
		m_pTdustReader = new ScalarTabdataReader(inputFilename);
		m_pTdustReader->openFile();
		m_pTdustReader->parseFile();
		m_pTdustReader->closeFile();
		if ( m_pTdustReader->getDataname() != "tdust" )
			throw domain_error("Data file " + inputFilename + " does not contain tdust data");
	}

	inputFilename = m_paramString["dopplerFilename"];
	if ( strip(inputFilename).size() ) {
		m_pDopplerReader = new ScalarTabdataReader(inputFilename);
		m_pDopplerReader->openFile();
		m_pDopplerReader->parseFile();
		m_pDopplerReader->closeFile();
		if ( m_pDopplerReader->getDataname() != "doppler" )
			throw domain_error("Data file " + inputFilename + " does not contain doppler data");
	}

	inputFilename = m_paramString["velocityFilename"];
	if ( strip(inputFilename).size() ) {
		m_pVelocityReader = new VectorTabdataReader(inputFilename);
		m_pVelocityReader->openFile();
		m_pVelocityReader->parseFile();
		m_pVelocityReader->closeFile();
		if ( m_pVelocityReader->getDataname() != "velocity" )
			throw domain_error("Data file " + inputFilename + " does not contain velocity data");
	}

	inputFilename = m_paramString["bmagFilename"];
	if ( strip(inputFilename).size() ) {
		m_pBmagReader = new VectorTabdataReader(inputFilename);
		m_pBmagReader->openFile();
		m_pBmagReader->parseFile();
		m_pBmagReader->closeFile();
		if ( m_pBmagReader->getDataname() != "bmag" )
			throw domain_error("Data file " + inputFilename + " does not contain bmag data");
	}

}

const double
ModelTabdata::density(const double x, const double y, const double z)
{
	if ( !m_pDensityReader )
		throw domain_error("No data read for density");

	TabdataIndex index = m_pDensityReader->getIndex(x, y, z);
	if ( !index.isValid )
		return 0.;

	return m_pDensityReader->getData(index);
}

const double
ModelTabdata::temperature(const double x, const double y, const double z)
{
	if ( !m_pTemperatureReader )
		throw domain_error("No data read for temperature");

	TabdataIndex index = m_pTemperatureReader->getIndex(x, y, z);
	if ( !index.isValid )
		return 0.;

	return m_pTemperatureReader->getData(index);
}

const double
ModelTabdata::abundance(const double x, const double y, const double z)
{
	if ( !m_pAbundanceReader )
		throw domain_error("No data read for abundance");

	TabdataIndex index = m_pAbundanceReader->getIndex(x, y, z);
	if ( !index.isValid )
		return 1.;

	return m_pAbundanceReader->getData(index);
}

const double
ModelTabdata::t_dust(const double x, const double y, const double z)
{
	if ( !m_pTdustReader )
		throw domain_error("No data read for tdust");

	TabdataIndex index = m_pTdustReader->getIndex(x, y, z);
	if ( !index.isValid )
		return 0.;

	return m_pTdustReader->getData(index);
}

const double
ModelTabdata::doppler(const double x, const double y, const double z)
{
	if ( !m_pDopplerReader )
		throw domain_error("No data read for doppler");

	TabdataIndex index = m_pDopplerReader->getIndex(x, y, z);
	if ( !index.isValid )
		return 0.;

	return m_pDopplerReader->getData(index);
}

const void
ModelTabdata::velocity(const double x, const double y, const double z, double* v)
{
	if ( !m_pVelocityReader )
		throw domain_error("No data read for velocity");

	TabdataIndex index = m_pVelocityReader->getIndex(x, y, z);
	if ( !index.isValid ) {
		v[0] = 0.;
		v[1] = 0.;
		v[2] = 0.;
	} else {
		const vector<double>& data = m_pVelocityReader->getData(index);

		v[0] = data[0];
		v[1] = data[1];
		v[2] = data[2];
	}
}

const void
ModelTabdata::bmag(const double x, const double y, const double z, double* b)
{
	if ( !m_pBmagReader )
		throw domain_error("No data read for bmag");

	TabdataIndex index = m_pBmagReader->getIndex(x, y, z);
	if ( !index.isValid ) {
		b[0] = 0.;
		b[1] = 0.;
		b[2] = 0.;
	} else {
		const vector<double>& data = m_pBmagReader->getData(index);

		b[0] = data[0];
		b[1] = data[1];
		b[2] = data[2];

	}
}



