////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Attila Juhasz
// Sterrewacht Leiden
// Universiteit Leiden
// juhasz@strw.leidenuniv.nl
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <math.h>

#include "modelUlrich76.h"

#include <stdexcept>
#include <complex>
#include <iostream>
// Define complex double data type */
typedef complex<double> dcomp; 


string
ModelUlrich76::modelID = "Ulrich76";

const ModelReg
ModelUlrich76::getModelReg()
{
  ModelReg modelReg;
  modelReg.name = "Ulrich76";
  modelReg.description = "Cloud collapse model with rotation (Ulrich 1976)";
  modelReg.bibref = "Ulrich 1976, ApJ, 210, 377";

  return modelReg;
}

const set<string>
ModelUlrich76::getParamIDs()
{
  set<string> paramIDs;
  paramIDs.insert("mdot");
  paramIDs.insert("mstar");
  paramIDs.insert("rc");
  return paramIDs;
}

const ParamReg*
ModelUlrich76::getParamReg(string paramID)
{
  if (paramID == "mdot")
    return new ParamRegDouble("mdot",
                              "Accretion rate",
                              1e-6,
                              "Msun/year");
  else if (paramID == "mstar")
    return new ParamRegDouble("mstar",
                              "Mass of the central protostar",
                              0.1,
                              "Msun");
  else if (paramID == "rc")
    return new ParamRegDouble("rc",
                              "Centrifugal radius",
                              60.0,
                              "AU");
  else
    throw invalid_argument("Invalid paramID " + paramID + " for modelID " + this->modelID);
}


ModelUlrich76::ModelUlrich76()
{}

ModelUlrich76::~ModelUlrich76()
{}


void 
ModelUlrich76::onFinalizeConfiguration()
{
  m_mdot      = m_paramDouble["mdot"]*6.3102486e+22;
  m_mstar     = m_paramDouble["mstar"]*1.99e+30;
  m_rc        = m_paramDouble["rc"]*1.496e+11;
}

//*****************************************************************************/
double get_theta0(double r, double theta_inp){

#ifndef min
			#define min( a, b ) ( ((a) < (b)) ? (a) : (b) )
#endif
#ifndef max
			#define max( a, b ) ( ((a) > (b)) ? (a) : (b) )
#endif

#define pi 3.14159265358979323846264338328
	
#define ngrid 100

	int i;
	double theta0_final;
	double theta0[ngrid], sint02[ngrid], diff, rr_old, rr_new, th0_old, th0_new;
	double epsilon, e, phi0, zeta, der, theta, dum, dum2;
    double mu, nu;

    mu = 0.0;
    nu = 0.0;

//	theta = theta_inp;

	if (theta_inp>pi/2.){
		theta = pi - theta_inp;
	} else{
		theta = theta_inp;
	}


// Before we would go on with the full grid interpolation
//  let's check if theta is zero. If it is the grid interpolation
// will fail anyway, and we know that if theta=0 then theta0=0 */

	if (theta==0.){
		theta0_final = 0.0;
	} 
	else {

// OK, so theta is not zero and we need to calculate theta0 via interpolation 
// Set up the theta0 grid
//
		for(i=0;i<ngrid;i++){
			theta0[i] = theta * (i*1.0)/(ngrid*1.0-1.);
			sint02[i] = sin(theta0[i])*sin(theta0[i]);
		}

		diff     = -1.0;
		rr_old   = 0.;
		rr_new   = 0.;
		th0_new  = 0.;
		th0_old  = 0.;
		i        = 10;

		while(diff<0.0&i<ngrid){
			rr_old  = rr_new;
			th0_old = th0_new;
			th0_new = theta0[i];

			epsilon = nu*nu + mu*mu *sint02[i] - 2.0*mu;
			e       = sqrt(1.0 + epsilon*sint02[i]); 
			dum2    = (1.0 - mu*sint02[i])/e; 
			phi0    = acos(min(dum2,1.0)); 
			dum     = min(cos(theta) / cos(th0_new),1.); 
			zeta    = acos(dum) + phi0;

			rr_new  = sint02[i] / (1.0 - e*cos(zeta)) / r;
			diff    = rr_new - 1.0;
  
			
			i  = i+1;
		}

// OK, now we either bracketed r or reached pi/2 now calculate theta0_final

		if (diff==0.){
			theta0_final = th0_new;
		}
		else{
			der   = (th0_new - th0_old) / (rr_new - rr_old);
			theta0_final = th0_old + der * (1.0 - rr_old); 
		}
	}
	
	if (theta_inp>pi/2.){
		theta0_final = pi - theta0_final;
	}

	return cos(theta0_final);

}
//*****************************************************************************/

//*****************************************************************************/
// A function to calculate the mu0 = cos(theta0) parameter in the model
// of Ulrich 1976
const double getmu0(double r, double rc, double mu){

	dcomp a, b, c, d, cc, qq;
	dcomp dum1, dum2, dum3, isq3;
	dcomp x1, x2, x3;
	double mu0;
	double theta, stheta;

	a = dcomp(rc,0.0);//rc + 0.*I;
	b = dcomp(0.0, 0.0);//0.0 + 0.*I;
	c = dcomp((r-rc),0.0); 
	d = dcomp((r*mu), 0.0);
	theta = acos(mu);
    stheta = sin(theta);	
  
	
	dum1 = 2.0 * pow(b,3.0) - 9.0 * a * b * c + 27.0 * pow(a,2.0) * d;
	dum2 = pow(b,2.0) - 3.0*a*c;
	dum3 = 3.*a;
	isq3 = dcomp(0.0, sqrt(3.0));

    qq = sqrt(pow(dum1,2.0) - 4.0 * pow(dum2,3.0));
	cc = pow((0.5 * (qq + dum1)), 1.0/3.0);

	x1 = -b / dum3 - cc / dum3 - (pow(b, 2.0) - dum3*c) / (dum3*cc);
	x2 = -b / dum3 + cc * (1.+isq3) / (dum3*2.0) + (1.0 - isq3) * (pow(b,2.0) - dum3*c) / (2.*dum3*cc);
	x3 = -b / dum3 + cc * (1.-isq3) / (dum3*2.0) + (1.0 + isq3) * (pow(b,2.0) - dum3*c) / (2.*dum3*cc);


/*	if (abs(imag(x1))==0.){
		if (sin(acos(x1))*stheta>0.){
			mu0 = creal(x1); 
		}
	}

	if (abs(imag(x2))==0.){
		if (sin(acos(x2))*stheta>0.){
			mu0 =creal(x2); 
		}
	}	
	if (abs(imag(x3))==0.){
		if (sin(acos(x3))*stheta>0.){
			mu0= creal(x3); 
		}
	}

	return mu0;
*/
	
	mu0 = 0;
 	if (imag(x1)<1e-5){
		if (abs(real(x1))>mu0){
			mu0 = fabs(real(x1));
		}
	}

 	if (imag(x2)<1e-5){
		if (abs(real(x2))>mu0){
			mu0 = fabs(real(x2));
		}
	}	
 	if (imag(x3)<1e-5){
		if (abs(real(x3))>mu0){
			mu0 = fabs(real(x3));
		}
	}	
	
	return mu0;

}
//*****************************************************************************/

const double 
ModelUlrich76::density(const double x, const double y, const double z)
{
#define pi 3.14159265358979323846264338328
#define gg 6.674e-11    // gravitational constant [cm^3 g^-1 s^-2]
#define mmw 2.0        // Mean molecular weight
#define mp 1.6726e-27  // Proton mass	
#define AU 1.49598e11

  double r, mu, mu0, rho;
  double dum1, dum2, dum3;

  r   = sqrt(x*x + y*y + z*z) ;
  // For z=0 we have a divergence in the equations so we'll cheat here a bit
  // and add 1.0 to the z value. Since the usual scale is several AU  this doesn't do any harm
  // to the solution
  mu  = fabs(z+1.) / (r+1e-90);
  mu0 = getmu0(r/m_rc, 1.0, mu);


  dum1 =  m_mdot / (8.0 * pi * sqrt(gg*m_mstar*pow(r,3.))) ;
  dum2 = sqrt(1. + mu/mu0) ;
  dum3 = (0.5*mu/mu0 + mu0*mu0 * m_rc/r) ;

  rho = m_mdot / (8.0 * pi * sqrt(gg*m_mstar*pow(r,3.))) / sqrt(1. + mu/mu0) / (0.5*mu/mu0 + mu0*mu0 * m_rc/r) / (mmw*mp);
	  
  
  return rho;
}

const void
ModelUlrich76::velocity(const double x, const double y, const double z, double* v)
{
#define pi 3.14159265358979323846264338328
#define gg 6.674e-11    // gravitational constant [cm^3 g^-1 s^-2]
#define AU		1.49598e11
  
  double r, rcyl, sinp, cosp, sint, cost, mu, mu0;
  double vr, vt, vp, dum_cost;

  r   = sqrt(x*x + y*y + z*z);
  rcyl = sqrt(x*x + y*y);

  sinp = y / (rcyl+1e-90);
  cosp = x / (rcyl+1e-90);
  sint = rcyl / (r+1e-90);
  cost = z    / (r+1e-90);
  dum_cost = (fabs(z)+1.0) / (r+1e-90); 
  
  mu  = cost;
  mu0 = getmu0(r/m_rc, 1.0, dum_cost);
  if (cost<0.){
	mu0 = -mu0;
  }
 

  vr  = -sqrt(gg*m_mstar/r) * sqrt(1.0 + mu/mu0);
  vt  =  sqrt(gg*m_mstar/r) * (mu0 - mu) * sqrt( (mu0+mu) / (mu0 * sqrt(1.0 - mu*mu)));
  vp  =  sqrt(gg*m_mstar/r) * sqrt(1.0 - mu0*mu0) / sqrt(1.0 - mu*mu) * sqrt(1.0 - mu/mu0);
 

  v[0] = (sint*cosp*vr + cost*cosp*vt - sinp*vp) ;
  v[1] = (sint*sinp*vr + cost*sinp*vt + cosp*vp) ;
  v[2] = (cost*vr - sint*vt) ;
  
}




