////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include "modellib.h"
#include "modelregistry.h"
#include "providerregistry.h"
#include "functionregistry.h"

#include "model.h"
#include "iProvider.h"

#include <stdexcept>

using namespace std;

// Singleton pattern:

ModelLib* ModelLib::m_theModelLib = 0;

ModelLib::ModelLib() :
  m_isInitialized(false),
  m_tdustIdentTemp(false),
  m_tempIdentTdust(false)
{
  m_pCurrentModel = NULL;
  m_currentModelID = "";

  m_pDensityProvider = NULL;
  m_pTemperatureProvider = NULL;
  m_pAbundanceProvider = NULL;
  m_pTdustProvider = NULL;
  m_pDopplerProvider = NULL;
  m_pVelocityProvider = NULL;
  m_pBmagProvider = NULL;
}

ModelLib::~ModelLib()
{
  unsetCurrentModel();
}

ModelLib*
ModelLib::getInstance()
{
  if (m_theModelLib == 0)
    {
      m_theModelLib = new ModelLib();
    }
  return m_theModelLib;
}

// Other:

void
ModelLib::initLibrary()
{
  if (m_isInitialized)
    throw domain_error("Library already initialized");

  ModelRegistry::getInstance()->registerModels();
  ProviderRegistry::getInstance()->registerProviders();
  FunctionRegistry::getInstance()->registerFunctions();

  m_providerIsModel[IDensityProvider::providerID] = false;
  m_providerIsModel[ITemperatureProvider::providerID] = false;
  m_providerIsModel[IAbundanceProvider::providerID] = false;
  m_providerIsModel[ITdustProvider::providerID] = false;
  m_providerIsModel[IDopplerProvider::providerID] = false;
  m_providerIsModel[IVelocityProvider::providerID] = false;
  m_providerIsModel[IBmagProvider::providerID] = false;

  m_providerFunctionIDs[IDensityProvider::providerID] = "";
  m_providerFunctionIDs[ITemperatureProvider::providerID] = "";
  m_providerFunctionIDs[IAbundanceProvider::providerID] = "";
  m_providerFunctionIDs[ITdustProvider::providerID] = "";
  m_providerFunctionIDs[IDopplerProvider::providerID] = "";
  m_providerFunctionIDs[IVelocityProvider::providerID] = "";
  m_providerFunctionIDs[IBmagProvider::providerID] = "";

  m_providerFunctions[IDensityProvider::providerID] = NULL;
  m_providerFunctions[ITemperatureProvider::providerID] = NULL;
  m_providerFunctions[IAbundanceProvider::providerID] = NULL;
  m_providerFunctions[ITdustProvider::providerID] = NULL;
  m_providerFunctions[IDopplerProvider::providerID] = NULL;
  m_providerFunctions[IVelocityProvider::providerID] = NULL;
  m_providerFunctions[IBmagProvider::providerID] = NULL;

  m_tdustIdentTemp = false;
  m_tempIdentTdust = false;

  m_isInitialized = true;
}


// Models:

////////////////////////////////////////////////////////////////////
//! Checks if a model is registered with the model library.
/*!
 * @param[in]   modelID
 *              ID of the model.
 * @returns     true if model with modelID is registered,
 *              else false.
 */
const bool
ModelLib::isRegisteredModel(const string modelID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->isRegisteredModel(modelID);
}

const set<string>
ModelLib::getModelIDs()
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getModelIDs();
}

const string
ModelLib::getModelName(const string modelID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getModelName(modelID);
}

const string
ModelLib::getModelDesc(const string modelID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getModelDesc(modelID);
}

const string
ModelLib::getModelBibref(const string modelID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getModelBibref(modelID);
}

const bool
ModelLib::isRegisteredParam(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->isRegisteredParam(modelID, paramID);
}

const set<string>
ModelLib::getParamIDs(const string modelID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamIDs(modelID);
}

const string
ModelLib::getParamType(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamType(modelID, paramID);
}

const string
ModelLib::getParamName(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamName(modelID, paramID);
}

const string
ModelLib::getParamDesc(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamDesc(modelID, paramID);
}

const string
ModelLib::getParamUnit(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamUnit(modelID, paramID);
}

const int
ModelLib::getParamDefValInt(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamDefValInt(modelID, paramID);
}

const double
ModelLib::getParamDefValDouble(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamDefValDouble(modelID, paramID);
}

const string
ModelLib::getParamDefValString(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamDefValString(modelID, paramID);
}

const vector<double>
ModelLib::getParamEnumValues(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamEnumValues(modelID, paramID);
}

const unsigned
ModelLib::getParamEnumDefIndex(const string modelID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return ModelRegistry::getInstance()->getParamEnumDefIndex(modelID, paramID);
}


// Provider:
const bool
ModelLib::isRegisteredProvider(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();
  return ProviderRegistry::getInstance()->isRegisteredProvider(providerID);
}

const set<string>
ModelLib::getProviderIDs()
{
  if (!m_isInitialized)
    initLibrary();
  return ProviderRegistry::getInstance()->getProviderIDs();
}

const string
ModelLib::getProviderName(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();
  return ProviderRegistry::getInstance()->getProviderName(providerID);
}

const string
ModelLib::getProviderDesc(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();
  return ProviderRegistry::getInstance()->getProviderDesc(providerID);
}

const string
ModelLib::getProviderUnit(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();
  return ProviderRegistry::getInstance()->getProviderUnit(providerID);
}

const string
ModelLib::getProviderType(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();
  return ProviderRegistry::getInstance()->getProviderType(providerID);

}

const set<string>
ModelLib::getModelProviderIDs(const string modelID)
{
  if (!m_isInitialized)
    initLibrary();

  set<string> providerIDs = set<string> ();

  Model *pModel = ModelRegistry::getInstance()->createModel(modelID);

  if (isDensityProvider(pModel))
    providerIDs.insert(IDensityProvider::providerID);
  if (isTemperatureProvider(pModel))
    providerIDs.insert(ITemperatureProvider::providerID);
  if (isAbundanceProvider(pModel))
    providerIDs.insert(IAbundanceProvider::providerID);
  if (isTdustProvider(pModel))
    providerIDs.insert(ITdustProvider::providerID);
  if (isDopplerProvider(pModel))
    providerIDs.insert(IDopplerProvider::providerID);
  if (isVelocityProvider(pModel))
    providerIDs.insert(IVelocityProvider::providerID);
  if (isBmagProvider(pModel))
    providerIDs.insert(IBmagProvider::providerID);

  delete pModel;

  return providerIDs;
}

const set<string>
ModelLib::getProviderModelIDs(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();

  set<string> modelIDs = set<string> ();

  set<string> allModelIDs = getModelIDs();
  set<string>::iterator itModelID;

  for (itModelID = allModelIDs.begin(); itModelID != allModelIDs.end(); itModelID++)
    {
      set<string> providerIDs = getModelProviderIDs(*itModelID);
      if (providerIDs.count(providerID))
        {
          modelIDs.insert(*itModelID);
        }
    }

  return modelIDs;
}

const bool
ModelLib::isProviderModel(const string providerID, const string modelID)
{
  if (!m_isInitialized)
    initLibrary();

  if (getProviderModelIDs(providerID).count(modelID))
    return true;
  else
    return false;
}

// Functions:
const bool
ModelLib::isRegisteredFunction(const string functionID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->isRegisteredFunction(functionID);
}

const set<string>
ModelLib::getFunctionIDs()
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getFunctionIDs();
}

const string
ModelLib::getFunctionName(const string functionID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getFunctionName(functionID);
}

const string
ModelLib::getFunctionDesc(const string functionID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getFunctionDesc(functionID);
}

const string
ModelLib::getFunctionType(const string functionID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getFunctionType(functionID);
}

const bool
ModelLib::isRegisteredFunctionParam(const string functionID,
    const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->isRegisteredParam(functionID, paramID);
}

const set<string>
ModelLib::getFunctionParamIDs(const string functionID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamIDs(functionID);
}

const string
ModelLib::getFunctionParamType(const string functionID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamType(functionID, paramID);
}

const string
ModelLib::getFunctionParamName(const string functionID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamName(functionID, paramID);
}

const string
ModelLib::getFunctionParamDesc(const string functionID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamDesc(functionID, paramID);
}

const int
ModelLib::getFunctionParamDefValInt(const string functionID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamDefValInt(functionID, paramID);
}

const double
ModelLib::getFunctionParamDefValDouble(const string functionID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamDefValDouble(functionID, paramID);
}

const string
ModelLib::getFunctionParamDefValString(const string functionID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamDefValString(functionID, paramID);
}

const vector<double>
ModelLib::getFunctionParamEnumValues(const string functionID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamEnumValues(functionID, paramID);
}

const unsigned
ModelLib::getFunctionParamEnumDefIndex(const string functionID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  return FunctionRegistry::getInstance()->getParamEnumDefIndex(functionID, paramID);
}

const set<string>
ModelLib::getFunctionProviderIDs(const string functionID)
{
  if (!m_isInitialized)
    initLibrary();

  set<string> providerIDs = set<string> ();

  Function *pFunction = FunctionRegistry::getInstance()->createFunction(
      functionID);

  if (isDensityProvider(pFunction))
    providerIDs.insert(IDensityProvider::providerID);
  if (isTemperatureProvider(pFunction))
    providerIDs.insert(ITemperatureProvider::providerID);
  if (isAbundanceProvider(pFunction))
    providerIDs.insert(IAbundanceProvider::providerID);
  if (isTdustProvider(pFunction))
    providerIDs.insert(ITdustProvider::providerID);
  if (isDopplerProvider(pFunction))
    providerIDs.insert(IDopplerProvider::providerID);
  if (isVelocityProvider(pFunction))
    providerIDs.insert(IVelocityProvider::providerID);
  if (isBmagProvider(pFunction))
    providerIDs.insert(IBmagProvider::providerID);

  delete pFunction;

  return providerIDs;
}

const set<string>
ModelLib::getProviderFunctionIDs(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();

  set<string> functionIDs = set<string> ();

  set<string> allFunctionIDs = getFunctionIDs();
  set<string>::iterator itFunctionID;

  for (itFunctionID = allFunctionIDs.begin(); itFunctionID
      != allFunctionIDs.end(); itFunctionID++)
    {
      set<string> providerIDs = getFunctionProviderIDs(*itFunctionID);
      if (providerIDs.count(providerID))
        {
          functionIDs.insert(*itFunctionID);
        }
    }

  return functionIDs;
}

const bool
ModelLib::isProviderFunction(const string providerID, const string functionID)
{
  if (!m_isInitialized)
    initLibrary();

  if (getProviderFunctionIDs(providerID).count(functionID))
    return true;
  else
    return false;
}


// Configuration:
// Model:
void
ModelLib::setCurrentModel(const string modelID)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);

  unsetCurrentModel();

  m_pCurrentModel = ModelRegistry::getInstance()->createModel(modelID);
  m_currentModelID = modelID;

  m_pCurrentModel->initializeParams();

  if (isDensityProvider(m_pCurrentModel)) {
    m_pDensityProvider = dynamic_cast<IDensityProvider*> (m_pCurrentModel);
    m_providerIsModel[IDensityProvider::providerID] = true;
  }
  if (isTemperatureProvider(m_pCurrentModel)) {
    m_pTemperatureProvider = dynamic_cast<ITemperatureProvider*> (m_pCurrentModel);
    m_providerIsModel[ITemperatureProvider::providerID] = true;
  }
  if (isAbundanceProvider(m_pCurrentModel)) {
    m_pAbundanceProvider = dynamic_cast<IAbundanceProvider*> (m_pCurrentModel);
    m_providerIsModel[IAbundanceProvider::providerID] = true;
  }
  if (isTdustProvider(m_pCurrentModel)) {
    m_pTdustProvider = dynamic_cast<ITdustProvider*> (m_pCurrentModel);
    m_providerIsModel[ITdustProvider::providerID] = true;
  }
  if (isDopplerProvider(m_pCurrentModel)) {
    m_pDopplerProvider = dynamic_cast<IDopplerProvider*> (m_pCurrentModel);
    m_providerIsModel[IDopplerProvider::providerID] = true;
  }
  if (isVelocityProvider(m_pCurrentModel)) {
    m_pVelocityProvider = dynamic_cast<IVelocityProvider*> (m_pCurrentModel);
    m_providerIsModel[IVelocityProvider::providerID] = true;
  }
  if (isBmagProvider(m_pCurrentModel)) {
    m_pBmagProvider = dynamic_cast<IBmagProvider*> (m_pCurrentModel);
    m_providerIsModel[IBmagProvider::providerID] = true;
  }

  if (m_providerIsModel[ITemperatureProvider::providerID] ||
	  m_providerIsModel[ITdustProvider::providerID] ) {
	  if (!m_providerIsModel[ITdustProvider::providerID])
		  setTdustIdentTemp();
	  if (!m_providerIsModel[ITemperatureProvider::providerID])
		  setTempIdentTdust();
  }
}

void
ModelLib::unsetCurrentModel()
{
	if (!m_isInitialized)
		initLibrary();

	//  if (m_pCurrentModel) {

	if (isTdustIdentTemp())
		unsetTdustIdentTemp();
	if (isTempIdentTdust())
		unsetTempIdentTdust();

	if (isCurrentProviderFunction(IDensityProvider::providerID))
		unsetFunction(IDensityProvider::providerID);
	m_pDensityProvider = NULL;
	m_providerIsModel[IDensityProvider::providerID] = false;

	if (isCurrentProviderFunction(ITemperatureProvider::providerID))
		unsetFunction(ITemperatureProvider::providerID);
	m_pTemperatureProvider = NULL;
	m_providerIsModel[ITemperatureProvider::providerID] = false;

	if (isCurrentProviderFunction(IAbundanceProvider::providerID))
		unsetFunction(IAbundanceProvider::providerID);
	m_pAbundanceProvider = NULL;
	m_providerIsModel[IAbundanceProvider::providerID] = false;

	if (isCurrentProviderFunction(ITdustProvider::providerID))
		unsetFunction(ITdustProvider::providerID);
	m_pTdustProvider = NULL;
	m_providerIsModel[ITdustProvider::providerID] = false;

	if (isCurrentProviderFunction(IDopplerProvider::providerID))
		unsetFunction(IDopplerProvider::providerID);
	m_pDopplerProvider = NULL;
	m_providerIsModel[IDopplerProvider::providerID] = false;

	if (isCurrentProviderFunction(IVelocityProvider::providerID))
		unsetFunction(IVelocityProvider::providerID);
	m_pVelocityProvider = NULL;
	m_providerIsModel[IVelocityProvider::providerID] = false;

	if (isCurrentProviderFunction(IBmagProvider::providerID))
		unsetFunction(IBmagProvider::providerID);
	m_pBmagProvider = NULL;
	m_providerIsModel[IBmagProvider::providerID] = false;

	if (m_pCurrentModel) {
		delete m_pCurrentModel;
		m_pCurrentModel = NULL;
		m_currentModelID = "";
	}
}

const bool
ModelLib::isSetCurrentModel()
{
  if (m_pCurrentModel)
    return true;
  return false;
}

const string
ModelLib::getCurrentModelID()
{
  return m_currentModelID;
}

const void
ModelLib::setParamDouble(string paramID, double val)
{
  if (!m_pCurrentModel)
    throw domain_error("No model set");
  if (!isRegisteredParam(m_currentModelID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getParamType(m_currentModelID, paramID) != "double")
    throw invalid_argument("Invalid datatype");

  m_pCurrentModel->setParamDouble(paramID, val);
}

const void
ModelLib::setParamInt(string paramID, int val)
{
  if (!m_pCurrentModel)
    throw domain_error("No model set");
  if (!isRegisteredParam(m_currentModelID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getParamType(m_currentModelID, paramID) != "int")
    throw invalid_argument("Invalid datatype");

  m_pCurrentModel->setParamInt(paramID, val);
}

const void
ModelLib::setParamString(string paramID, string val)
{
  if (!m_pCurrentModel)
    throw domain_error("No model set");
  if (!isRegisteredParam(m_currentModelID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getParamType(m_currentModelID, paramID) != "string")
    throw invalid_argument("Invalid datatype");

  m_pCurrentModel->setParamString(paramID, val);
}

const void
ModelLib::setParamEnumIndex(string paramID, unsigned index)
{
  if (!m_pCurrentModel)
    throw domain_error("No model set");
  if (!isRegisteredParam(m_currentModelID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getParamType(m_currentModelID, paramID) != "enum")
    throw invalid_argument("Invalid datatype");

  m_pCurrentModel->setParamEnumIndex(paramID, index);
}


const double
ModelLib::getParamDouble(const string paramID)
{
  if (!m_pCurrentModel)
    throw domain_error("No model set");
  if (!isRegisteredParam(m_currentModelID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getParamType(m_currentModelID, paramID) != "double")
    throw invalid_argument("Invalid datatype");

  return m_pCurrentModel->getParamDouble(paramID);
}

const int
ModelLib::getParamInt(const string paramID)
{
  if (!m_pCurrentModel)
    throw domain_error("No model set");
  if (!isRegisteredParam(m_currentModelID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getParamType(m_currentModelID, paramID) != "int")
    throw invalid_argument("Invalid datatype");

  return m_pCurrentModel->getParamInt(paramID);
}

const string
ModelLib::getParamString(const string paramID)
{
  if (!m_pCurrentModel)
    throw domain_error("No model set");
  if (!isRegisteredParam(m_currentModelID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getParamType(m_currentModelID, paramID) != "string")
    throw invalid_argument("Invalid datatype");

  return m_pCurrentModel->getParamString(paramID);
}

const unsigned
ModelLib::getParamEnumIndex(const string paramID)
{
  if (!m_pCurrentModel)
    throw domain_error("No model set");
  if (!isRegisteredParam(m_currentModelID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getParamType(m_currentModelID, paramID) != "enum")
    throw invalid_argument("Invalid datatype");

  return m_pCurrentModel->getParamEnumIndex(paramID);
}

// Functions:
void
ModelLib::setFunction(const string providerID, const string functionID)
{
	if (!m_isInitialized)
		initLibrary();
	if (!isRegisteredProvider(providerID))
		throw invalid_argument("Invalid providerID " + providerID);
	if (!isRegisteredFunction(functionID))
		throw invalid_argument("Invalid functionID " + functionID);
	if (!isProviderFunction(providerID, functionID))
		throw invalid_argument("Invalid combination of providerID " + providerID
				+ "and functionID " + functionID);

	if (isCurrentProviderFunction(providerID))
		unsetFunction(providerID);

	Function* function = FunctionRegistry::getInstance()->createFunction(functionID);

	function->initializeParams();

	if (providerID == IDensityProvider::providerID)
		m_pDensityProvider = dynamic_cast<IDensityProvider*> (function);
	else if (providerID == ITemperatureProvider::providerID) {
		if (isTempIdentTdust())
			unsetTempIdentTdust();
		m_pTemperatureProvider = dynamic_cast<ITemperatureProvider*> (function);
		if (isTdustIdentTemp()) {
			unsetTdustIdentTemp();
			setTdustIdentTemp();
		}
	}
	else if (providerID == IAbundanceProvider::providerID)
		m_pAbundanceProvider = dynamic_cast<IAbundanceProvider*> (function);
	else if (providerID == ITdustProvider::providerID) {
		if (isTdustIdentTemp())
			unsetTdustIdentTemp();
		m_pTdustProvider = dynamic_cast<ITdustProvider*> (function);
		if (isTempIdentTdust()) {
			unsetTempIdentTdust();
			setTempIdentTdust();
		}
	}
	else if (providerID == IDopplerProvider::providerID)
		m_pDopplerProvider = dynamic_cast<IDopplerProvider*> (function);
	else if (providerID == IVelocityProvider::providerID)
		m_pVelocityProvider = dynamic_cast<IVelocityProvider*> (function);
	else if (providerID == IBmagProvider::providerID)
		m_pBmagProvider = dynamic_cast<IBmagProvider*> (function);
	else
		throw runtime_error("Error setting function for providerID " + providerID);

	m_providerIsModel[providerID] = false;
	m_providerFunctions[providerID] = function;
	m_providerFunctionIDs[providerID] = functionID;
}

void
ModelLib::unsetFunction(const string providerID)
{
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  if (providerID == IDensityProvider::providerID) {
      delete m_providerFunctions[IDensityProvider::providerID];
      m_providerFunctions[IDensityProvider::providerID] = NULL;
      m_providerFunctionIDs[IDensityProvider::providerID] = "";
      if (m_pCurrentModel && isDensityProvider(m_pCurrentModel)) {
          m_pDensityProvider = dynamic_cast<IDensityProvider*> (m_pCurrentModel);
          m_providerIsModel[providerID] = true;
      } else {
          m_pDensityProvider = NULL;
          m_providerIsModel[providerID] = false;
      }
  } else if (providerID == ITemperatureProvider::providerID) {
      delete m_providerFunctions[ITemperatureProvider::providerID];
      m_providerFunctions[ITemperatureProvider::providerID] = NULL;
      m_providerFunctionIDs[ITemperatureProvider::providerID] = "";
      if (m_pCurrentModel && isTemperatureProvider(m_pCurrentModel)) {
          m_pTemperatureProvider = dynamic_cast<ITemperatureProvider*> (m_pCurrentModel);
          m_providerIsModel[providerID] = true;
      } else {
          m_pTemperatureProvider = NULL;
          m_providerIsModel[providerID] = false;
      }
  } else if (providerID == IAbundanceProvider::providerID) {
      delete m_providerFunctions[IAbundanceProvider::providerID];
      m_providerFunctions[IAbundanceProvider::providerID] = NULL;
      m_providerFunctionIDs[IAbundanceProvider::providerID] = "";
      if (m_pCurrentModel && isAbundanceProvider(m_pCurrentModel)) {
          m_pAbundanceProvider = dynamic_cast<IAbundanceProvider*> (m_pCurrentModel);
          m_providerIsModel[providerID] = true;
      } else {
          m_pAbundanceProvider = NULL;
          m_providerIsModel[providerID] = false;
      }
  } else if (providerID == ITdustProvider::providerID) {
      delete m_providerFunctions[ITdustProvider::providerID];
      m_providerFunctions[ITdustProvider::providerID] = NULL;
      m_providerFunctionIDs[ITdustProvider::providerID] = "";
      if (m_pCurrentModel && isTdustProvider(m_pCurrentModel)) {
          m_pTdustProvider = dynamic_cast<ITdustProvider*> (m_pCurrentModel);
          m_providerIsModel[providerID] = true;
      } else {
          m_pTdustProvider = NULL;
          m_providerIsModel[providerID] = false;
      }
  } else if (providerID == IDopplerProvider::providerID) {
      delete m_providerFunctions[IDopplerProvider::providerID];
      m_providerFunctions[IDopplerProvider::providerID] = NULL;
      m_providerFunctionIDs[IDopplerProvider::providerID] = "";
      if (m_pCurrentModel && isDopplerProvider(m_pCurrentModel)) {
          m_pDopplerProvider = dynamic_cast<IDopplerProvider*> (m_pCurrentModel);
          m_providerIsModel[providerID] = true;
      } else {
          m_pDopplerProvider = NULL;
          m_providerIsModel[providerID] = false;
      }
  } else if (providerID == IVelocityProvider::providerID) {
      delete m_providerFunctions[IVelocityProvider::providerID];
      m_providerFunctions[IVelocityProvider::providerID] = NULL;
      m_providerFunctionIDs[IVelocityProvider::providerID] = "";
      if (m_pCurrentModel && isVelocityProvider(m_pCurrentModel)) {
          m_pVelocityProvider = dynamic_cast<IVelocityProvider*> (m_pCurrentModel);
          m_providerIsModel[providerID] = true;
      } else {
          m_pVelocityProvider = NULL;
          m_providerIsModel[providerID] = false;
      }
  } else if (providerID == IBmagProvider::providerID) {
      delete m_providerFunctions[IBmagProvider::providerID];
      m_providerFunctions[IBmagProvider::providerID] = NULL;
      m_providerFunctionIDs[IBmagProvider::providerID] = "";
      if (m_pCurrentModel && isBmagProvider(m_pCurrentModel)) {
          m_pBmagProvider = dynamic_cast<IBmagProvider*> (m_pCurrentModel);
          m_providerIsModel[providerID] = true;
      } else {
          m_pBmagProvider = NULL;
          m_providerIsModel[providerID] = false;
      }
  } else
    throw runtime_error("Error unsetting function for providerID " + providerID);
}

const string
ModelLib::getFunctionID(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);

  return m_providerFunctionIDs[providerID];
}

const void
ModelLib::setFunctionParamDouble(const string providerID, const string paramID,
    double val)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  string functionID = m_providerFunctionIDs[providerID];
  if (!isRegisteredFunctionParam(functionID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getFunctionParamType(functionID, paramID) != "double")
    throw invalid_argument("Invalid datatype");

  m_providerFunctions[providerID]->setParamDouble(paramID, val);
}

const void
ModelLib::setFunctionParamInt(const string providerID, const string paramID,
    int val)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  string functionID = m_providerFunctionIDs[providerID];
  if (!isRegisteredFunctionParam(functionID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getFunctionParamType(functionID, paramID) != "int")
    throw invalid_argument("Invalid datatype");

  m_providerFunctions[providerID]->setParamInt(paramID, val);
}

const void
ModelLib::setFunctionParamString(const string providerID, const string paramID,
    string val)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  string functionID = m_providerFunctionIDs[providerID];
  if (!isRegisteredFunctionParam(functionID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getFunctionParamType(functionID, paramID) != "string")
    throw invalid_argument("Invalid datatype");

  m_providerFunctions[providerID]->setParamString(paramID, val);
}

const void
ModelLib::setFunctionParamEnumIndex(const string providerID, const string paramID, unsigned index)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  string functionID = m_providerFunctionIDs[providerID];
  if (!isRegisteredFunctionParam(functionID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getFunctionParamType(functionID, paramID) != "string")
    throw invalid_argument("Invalid datatype");

  m_providerFunctions[providerID]->setParamEnumIndex(paramID, index);
}

const double
ModelLib::getFunctionParamDouble(const string providerID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  string functionID = m_providerFunctionIDs[providerID];
  if (!isRegisteredFunctionParam(functionID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getFunctionParamType(functionID, paramID) != "double")
    throw invalid_argument("Invalid datatype");

  return m_providerFunctions[providerID]->getParamDouble(paramID);
}

const int
ModelLib::getFunctionParamInt(const string providerID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  string functionID = m_providerFunctionIDs[providerID];
  if (!isRegisteredFunctionParam(functionID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getFunctionParamType(functionID, paramID) != "int")
    throw invalid_argument("Invalid datatype");

  return m_providerFunctions[providerID]->getParamInt(paramID);
}

const string
ModelLib::getFunctionParamString(const string providerID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  string functionID = m_providerFunctionIDs[providerID];
  if (!isRegisteredFunctionParam(functionID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getFunctionParamType(functionID, paramID) != "string")
    throw invalid_argument("Invalid datatype");

  return m_providerFunctions[providerID]->getParamString(paramID);
}

const unsigned
ModelLib::getFunctionParamEnumIndex(const string providerID, const string paramID)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);
  if (!isCurrentProviderFunction(providerID))
    throw domain_error("No function set for providerID " + providerID);

  string functionID = m_providerFunctionIDs[providerID];
  if (!isRegisteredFunctionParam(functionID, paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (getFunctionParamType(functionID, paramID) != "string")
    throw invalid_argument("Invalid datatype");

  return m_providerFunctions[providerID]->getParamEnumIndex(paramID);
}

// Identification of Tdust with Temperature:
void
ModelLib::setTdustIdentTemp()
{
	if (isTempIdentTdust())
		throw domain_error("Cannot identify Tdust with Temperature, since Temperature already identified with Tdust");

	if (!isTdustIdentTemp()) {
		if (isCurrentProviderFunction(ITdustProvider::providerID))
			unsetFunction(ITdustProvider::providerID);
		m_pTdustProvider = new TempTdustAdaptor(m_pTemperatureProvider);
		m_providerIsModel[ITdustProvider::providerID] = false;
		m_tdustIdentTemp = true;
	}
}

void
ModelLib::unsetTdustIdentTemp()
{
	if (!isTdustIdentTemp())
		throw domain_error("Tdust not identified with Temperature");

	delete m_pTdustProvider;
	m_pTdustProvider = NULL;
	m_tdustIdentTemp = false;

	if (isSetCurrentModel()) {
		if (isTdustProvider(m_pCurrentModel)) {
		    m_pTdustProvider = dynamic_cast<ITdustProvider*> (m_pCurrentModel);
		    m_providerIsModel[ITdustProvider::providerID] = true;
		}
	}
}

const bool
ModelLib::isTdustIdentTemp()
{
	return m_tdustIdentTemp;
}


// Identification of Temperature with Tdust:
void
ModelLib::setTempIdentTdust()
{
	if (isTdustIdentTemp())
		throw domain_error("Cannot identify Temperature with Tdust, since Tdust already identified with Temperature");

	if (!isTempIdentTdust()) {
		if (isCurrentProviderFunction(ITemperatureProvider::providerID))
			unsetFunction(ITemperatureProvider::providerID);
		m_pTemperatureProvider = new TdustTempAdaptor(m_pTdustProvider);
		m_providerIsModel[ITemperatureProvider::providerID] = false;
		m_tempIdentTdust = true;
	}
}

void
ModelLib::unsetTempIdentTdust()
{
	if (!isTempIdentTdust())
		throw domain_error("Temperature not identified with Tdust");

	delete m_pTemperatureProvider;
	m_pTemperatureProvider = NULL;
	m_tempIdentTdust = false;

	if (isSetCurrentModel()) {
		if (isTemperatureProvider(m_pCurrentModel)) {
		    m_pTemperatureProvider = dynamic_cast<ITemperatureProvider*> (m_pCurrentModel);
		    m_providerIsModel[ITemperatureProvider::providerID] = true;
		}
	}
}

const bool
ModelLib::isTempIdentTdust()
{
	return m_tempIdentTdust;
}



// Providers:
const bool
ModelLib::isCurrentProviderModel(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);

  return m_providerIsModel[providerID];
}


const bool
ModelLib::isCurrentProviderFunction(const string providerID)
{
  if (!m_isInitialized)
    initLibrary();
  if (!isRegisteredProvider(providerID))
    throw invalid_argument("Invalid providerID " + providerID);

  if (m_providerFunctionIDs[providerID] == "")
    return false;

  return true;
}


const bool
ModelLib::isConfigurationComplete()
{
	if ( m_pDensityProvider &&
		 m_pTemperatureProvider &&
		 m_pAbundanceProvider &&
		 m_pTdustProvider &&
		 m_pDopplerProvider &&
		 m_pVelocityProvider &&
		 m_pBmagProvider )
		return true;
	return false;
}

const void
ModelLib::finalizeConfiguration()
{
	if (m_pCurrentModel)
		m_pCurrentModel->finalizeConfiguration();

	map<string, Function*>::iterator itFunction;
	for (itFunction=m_providerFunctions.begin(); itFunction!=m_providerFunctions.end(); itFunction++) {
		string providerID = itFunction->first;
		if (isCurrentProviderFunction(providerID)) {
			Function* pFunction = itFunction->second;
			pFunction->finalizeConfiguration();
		}
	}
}

const bool
ModelLib::isFinalizedConfiguration()
{
	if (m_pCurrentModel)
		if (!m_pCurrentModel->isFinalizedConfiguration())
			return false;

	map<string, Function*>::iterator itFunction;
	for (itFunction=m_providerFunctions.begin(); itFunction!=m_providerFunctions.end(); itFunction++) {
		string providerID = itFunction->first;
		if (isCurrentProviderFunction(providerID)) {
			Function* pFunction = itFunction->second;
			if(!pFunction->isFinalizedConfiguration())
				return false;
		}
	}

	return true;
}

// Usage:

const double
ModelLib::density(const double x, const double y, const double z)
{
  return m_pDensityProvider->density(x, y, z);
}

const double
ModelLib::temperature(const double x, const double y, const double z)
{
  return m_pTemperatureProvider->temperature(x, y, z);
}

const double
ModelLib::abundance(const double x, const double y, const double z)
{
  return m_pAbundanceProvider->abundance(x, y, z);
}

const double
ModelLib::t_dust(const double x, const double y, const double z)
{
  return m_pTdustProvider->t_dust(x, y, z);
}

const double
ModelLib::doppler(const double x, const double y, const double z)
{
  return m_pDopplerProvider->doppler(x, y, z);
}

const void
ModelLib::velocity(const double x, const double y, const double z, double* v)
{
  m_pVelocityProvider->velocity(x, y, z, v);
}

const void
ModelLib::bmag(const double x, const double y, const double z, double* b)
{
  m_pBmagProvider->bmag(x, y, z, b);
}

