////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef MODELLIB_H
#define MODELLIB_H

#include <string>
#include <set>
#include <vector>

#include "model.h"
#include "iProvider.h"

using namespace std;

class Function;

class 
ModelLib
{
  ////////////////////////////////////////////////////////////////////
  //! Implementation of the Singleton pattern.
  /**
   * @name Singleton
   */
  //@{
public:
  static ModelLib* getInstance();
  virtual ~ModelLib();
protected:
  ModelLib();
private:
  static ModelLib* m_theModelLib;
  //@}

 // Other:
 public:
  void initLibrary();

  ////////////////////////////////////////////////////////////////////
  //! Functions that allow the exploration of the model library.
  /**
   * @name Exploration
   */
  //@{

  // Models:
  const bool isRegisteredModel(const string modelID);
  const set<string> getModelIDs();
  const string getModelName(const string modelID);
  const string getModelDesc(const string modelID);
  const string getModelBibref(const string modelID);

  const bool isRegisteredParam(const string modelID, const string paramID);
  const set<string> getParamIDs(const string modelID);
  const string getParamType(const string modelID, const string paramID);
  const string getParamName(const string modelID, const string paramID);
  const string getParamDesc(const string modelID, const string paramID);
  const string getParamUnit(const string modelID, const string paramID);
  const int getParamDefValInt(const string modelID, const string paramID);
  const double getParamDefValDouble(const string modelID, const string paramID);
  const string getParamDefValString(const string modelID, const string paramID);
  const vector<double> getParamEnumValues(const string modelID, const string paramID);
  const unsigned getParamEnumDefIndex(const string modelID, const string paramID);

  // Providers:
  const bool isRegisteredProvider(const string providerID);
  const set<string> getProviderIDs();
  const string getProviderName(const string providerID);
  const string getProviderDesc(const string providerID);
  const string getProviderUnit(const string providerID);
  const string getProviderType(const string providerID);

  const set<string> getModelProviderIDs(const string modelID);
  const set<string> getProviderModelIDs(const string providerID);
  const bool isProviderModel(const string providerID, const string modelID);

  // Functions:
  const bool isRegisteredFunction(const string functionID);
  const set<string> getFunctionIDs();
  const string getFunctionName(const string functionID);
  const string getFunctionDesc(const string functionID);
  const string getFunctionType(const string functionID);

  const set<string> getFunctionProviderIDs(const string functionID);
  const set<string> getProviderFunctionIDs(const string providerID);
  const bool isProviderFunction(const string providerID, const string functionID);

  const bool isRegisteredFunctionParam(const string functionID, const string paramID);
  const set<string> getFunctionParamIDs(const string functionID);
  const string getFunctionParamType(const string functionID, const string paramID);
  const string getFunctionParamName(const string functionID, const string paramID);
  const string getFunctionParamDesc(const string functionID, const string paramID);
  const int getFunctionParamDefValInt(const string functionID, const string paramID);
  const double getFunctionParamDefValDouble(const string functionID, const string paramID);
  const string getFunctionParamDefValString(const string functionID, const string paramID);
  const vector<double> getFunctionParamEnumValues(const string functionID, const string paramID);
  const unsigned getFunctionParamEnumDefIndex(const string functionID, const string paramID);



  //@}

  // Configuration:
  // Model:
  void setCurrentModel(const string modelID);
  void unsetCurrentModel();
  const bool isSetCurrentModel();
  const string getCurrentModelID();

  const void setParamDouble(string paramID, double val);
  const void setParamInt(string paramID, int val);
  const void setParamString(string paramID, string val);
  const void setParamEnumIndex(string paramID, unsigned index);

  const double getParamDouble(const string paramID);
  const int getParamInt(const string paramID);
  const string getParamString(const string paramID);
  const unsigned getParamEnumIndex(const string paramID);

  // Functions:
  void setFunction(const string providerID, const string functionID);
  void unsetFunction(const string providerID);
  const string getFunctionID(const string providerID);

  const void setFunctionParamDouble(const string providerID, const string paramID, double val);
  const void setFunctionParamInt(const string providerID, const string paramID, int val);
  const void setFunctionParamString(const string providerID, const string paramID, string val);
  const void setFunctionParamEnumIndex(const string resultID, const string paramID, unsigned index);

  const double getFunctionParamDouble(const string providerID, const string paramID);
  const int getFunctionParamInt(const string providerID, const string paramID);
  const string getFunctionParamString(const string providerID, const string paramID);
  const unsigned getFunctionParamEnumIndex(const string resultID, const string paramID);

  // Identification of Tdust with Temperature:
  void setTdustIdentTemp();
  void unsetTdustIdentTemp();
  const bool isTdustIdentTemp();

  // Identification of Temperature with Tdust:
  void setTempIdentTdust();
  void unsetTempIdentTdust();
  const bool isTempIdentTdust();

  // Providers:
  const bool isCurrentProviderModel(const string providerID);
  const bool isCurrentProviderFunction(const string providerID);

  const bool isConfigurationComplete();

  const void finalizeConfiguration();
  const bool isFinalizedConfiguration();

  // Usage:
  const double density(const double x, const double y, const double z);
  const double temperature(const double x, const double y, const double z);
  const double abundance(const double x, const double y, const double z);
  const double t_dust(const double x, const double y, const double z);
  const double doppler(const double x, const double y, const double z);
  const void velocity(const double x, const double y, const double z, double* v);
  const void bmag(const double x, const double y, const double z, double* b);

 private:
  bool m_isInitialized;

  Model* m_pCurrentModel;
  string m_currentModelID;

  IDensityProvider* m_pDensityProvider;
  ITemperatureProvider* m_pTemperatureProvider;
  IAbundanceProvider* m_pAbundanceProvider;
  ITdustProvider* m_pTdustProvider;
  IDopplerProvider* m_pDopplerProvider;
  IVelocityProvider* m_pVelocityProvider;
  IBmagProvider* m_pBmagProvider;

  map <string, bool> m_providerIsModel;
  map <string, string> m_providerFunctionIDs;
  map <string, Function*> m_providerFunctions;

  bool m_tdustIdentTemp;
  bool m_tempIdentTdust;
};

#endif //MODELLIB_H
