////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include "modelregistry.h"
#include "parameterregistry.h"
#include "providerregistry.h"
#include "model.h"

#include <stdexcept>

using namespace std;

////////////////////////////////////////////////////////////////////
// Edit here when adding a new model:
//
// The example gives the additions for class TemplateModel with header file templateModel.h
//
#include "modelUser.h"
#include "modelAllen.h"
#include "modelDDN01.h"
#include "modelUlrich76.h"
#include "modelMendoza09.h"
#include "modelBonnorEbert56.h"
#include "modelShu77.h"
#include "modelLiShu96.h"
#include "modelCG97.h"
#include "modelMamon88.h"
#include "modelTabdata.h"
//#include "templateModel.h"

void
ModelRegistry::registerModels()
{

  registerModel(ModelUser::modelID);
  registerModel(ModelAllen::modelID);
  registerModel(ModelDDN01::modelID);
  registerModel(ModelUlrich76::modelID);
  registerModel(ModelMendoza09::modelID);
  registerModel(ModelBonnorEbert56::modelID);
  registerModel(ModelShu77::modelID);
  registerModel(ModelLiShu96::modelID);
  registerModel(ModelCG97::modelID);
  registerModel(ModelMamon88::modelID);
  registerModel(ModelTabdata::modelID);
//  registerModel(TemplateModel::modelID);
}

Model*
ModelRegistry::createModel(const string modelID)
const
{
  if (modelID == ModelUser::modelID)
    return new ModelUser();
  if (modelID == ModelAllen::modelID)
    return new ModelAllen();
  if (modelID == ModelDDN01::modelID)
    return new ModelDDN01();
  if (modelID == ModelUlrich76::modelID)
    return new ModelUlrich76();
  if (modelID == ModelMendoza09::modelID)
    return new ModelMendoza09();
  if (modelID == ModelBonnorEbert56::modelID)
    return new ModelBonnorEbert56();
  if (modelID == ModelShu77::modelID)
    return new ModelShu77();
  if (modelID == ModelLiShu96::modelID)
    return new ModelLiShu96();
  if (modelID == ModelCG97::modelID)
    return new ModelCG97();
  if (modelID == ModelMamon88::modelID)
    return new ModelMamon88();
  if (modelID == ModelTabdata::modelID)
    return new ModelTabdata();
//  if (modelID == TemplateModel::modelID)
//    return new TemplateModel();

  throw invalid_argument("Invalid modelID " + modelID);
}
//
// No further editing necessary below!
////////////////////////////////////////////////////////////////////


////////////////////////////////////////////////////////////////////
// Singleton pattern:
//
ModelRegistry* ModelRegistry::m_theModelRegistry = 0;

ModelRegistry::ModelRegistry(){}

ModelRegistry::~ModelRegistry(){}

ModelRegistry* ModelRegistry::getInstance()
{
  if (m_theModelRegistry == 0) {
    m_theModelRegistry = new ModelRegistry();
  }
  return m_theModelRegistry;
}
//
////////////////////////////////////////////////////////////////////

////////////////////////////////////////////////////////////////////
// Other:
//
void
ModelRegistry::addModelReg(const string& modelID,
                           const ModelReg modelReg)
{
  m_modelRegs[modelID] = modelReg;
  m_modelIDs.insert(modelID);
}

const ModelReg&
ModelRegistry::getModelReg(const string& modelID)
const
{
  map<string, ModelReg>::const_iterator pos;
  pos = m_modelRegs.find(modelID);
  if (pos == m_modelRegs.end())
    throw invalid_argument("Invalid modelID " + modelID);
  return pos->second;
}

void
ModelRegistry::registerModel(const string& modelID)
{
  if (isRegisteredModel(modelID))
    throw invalid_argument("Multiple registration of modelID " + modelID);

  Model* model = createModel(modelID);

  ModelReg modelReg = model->getModelReg();
  addModelReg(modelID, modelReg);

  m_parameterRegistries[modelID] = ParameterRegistry();
  ParameterRegistry& parameterRegistry = m_parameterRegistries[modelID];

  set<string> paramIDs = model->getParamIDs();
  set<string>::iterator paramID;
  for (paramID = paramIDs.begin(); paramID != paramIDs.end(); paramID++) {
    if (parameterRegistry.isRegisteredParam(*paramID))
      throw invalid_argument("Multiple registration of modelID, paramID "
                             + modelID + "," + *paramID);
    const ParamReg* paramReg = model->getParamReg(*paramID);
    parameterRegistry.addParamReg(*paramID, paramReg);
    delete paramReg;
  }

  delete model;
}

bool
ModelRegistry::isRegisteredModel(const string& modelID)
const
{
  if (m_modelRegs.count(modelID))
    return true;
  return false;
}

set<string>
ModelRegistry::getModelIDs()
const
{
  return m_modelIDs;
}

string
ModelRegistry::getModelName(const string& modelID)
const
{
  return getModelReg(modelID).name;
}

string
ModelRegistry::getModelDesc(const string& modelID)
const
{
  return getModelReg(modelID).description;
}

string
ModelRegistry::getModelBibref(const string& modelID)
const
{
  return getModelReg(modelID).bibref;
}

const ParameterRegistry&
ModelRegistry::getParameterRegistry(const string& modelID)
const
{
  map<string, ParameterRegistry>::const_iterator pos;
  pos = m_parameterRegistries.find(modelID);
  if (pos == m_parameterRegistries.end())
    throw invalid_argument("Invalid modelID " + modelID);
  return pos->second;

}

bool
ModelRegistry::isRegisteredParam(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).isRegisteredParam(paramID);
}

set<string>
ModelRegistry::getParamIDs(const string& modelID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamIDs();
}

string
ModelRegistry::getParamType(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamType(paramID);
}

string
ModelRegistry::getParamName(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamName(paramID);
}

string
ModelRegistry::getParamDesc(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamDesc(paramID);
}

string
ModelRegistry::getParamUnit(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamUnit(paramID);
}

int
ModelRegistry::getParamDefValInt(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamDefValInt(paramID);
}

double
ModelRegistry::getParamDefValDouble(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamDefValDouble(paramID);
}

string
ModelRegistry::getParamDefValString(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamDefValString(paramID);
}

vector<double>
ModelRegistry::getParamEnumValues(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamEnumValues(paramID);
}

unsigned
ModelRegistry::getParamEnumDefIndex(const string& modelID, const string& paramID)
const
{
  if (!isRegisteredModel(modelID))
    throw invalid_argument("Invalid modelID " + modelID);
  return getParameterRegistry(modelID).getParamEnumDefIndex(paramID);
}


