////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <string>
#include <typeinfo>
#include <stdexcept>

#include <iostream>

#include "parameterregistry.h"

using namespace std;

ParamReg::ParamReg(const string& name, const string& description)
: m_name(name),
  m_description(description)
{}

ParamRegInt::ParamRegInt(const string& name, const string& description, const int& defValue)
: ParamReg(name, description),
  m_defValue(defValue)
{}

ParamRegDouble::ParamRegDouble(const string& name, const string& description, const double& defValue, const string& unit)
: ParamReg(name, description),
  m_defValue(defValue),
  m_unit(unit)
{}

ParamRegString::ParamRegString(const string& name, const string& description, const string& defValue)
: ParamReg(name, description),
  m_defValue(defValue)
{}

ParamRegEnum::ParamRegEnum(const string& name,
                           const string& description,
                           const unsigned& nValues,
                           const double values[],
                           const int& defIndex,
                           const string& unit)
: ParamReg(name, description),
  m_defIndex(defIndex),
  m_unit(unit)
{
  m_values.assign(values, values + nValues);
}








void
ParameterRegistry::addParamReg(const string& paramID,
                               const ParamReg* paramReg)
{
  if (isRegisteredParam(paramID))
    throw invalid_argument("Multiple registration of paramID " + paramID);

  bool foundType = false;

  const ParamRegInt* iParamReg = dynamic_cast<const ParamRegInt*>(paramReg);
  if (iParamReg) {
      foundType = true;
      m_intParamRegs[paramID] = ParamRegInt(*iParamReg);
  }

  const ParamRegDouble* dParamReg = dynamic_cast<const ParamRegDouble*>(paramReg);
  if (dParamReg) {
      foundType = true;
      m_doubleParamRegs[paramID] = ParamRegDouble(*dParamReg);
  }

  const ParamRegString* sParamReg = dynamic_cast<const ParamRegString*>(paramReg);
  if (sParamReg) {
      foundType = true;
      m_stringParamRegs[paramID] = ParamRegString(*sParamReg);
  }

  const ParamRegEnum* eParamReg = dynamic_cast<const ParamRegEnum*>(paramReg);
  if (eParamReg) {
      foundType = true;
      m_enumParamRegs[paramID] = ParamRegEnum(*eParamReg);
  }

  if (!foundType)
    throw domain_error("Invalid datatype found in registration of paramID " + paramID);

  m_paramIDs.insert(paramID);
}


bool
ParameterRegistry::isRegisteredParam(const string& paramID)
const
{
  if (m_intParamRegs.count(paramID)    ||
      m_doubleParamRegs.count(paramID) ||
      m_stringParamRegs.count(paramID) ||
      m_enumParamRegs.count(paramID)   )
    return true;
  return false;
}

set<string>
ParameterRegistry::getParamIDs()
const
{
  return m_paramIDs;
}


string
ParameterRegistry::getParamType(const string& paramID)
const
{
  if (!isRegisteredParam(paramID))
    throw invalid_argument("Invalid paramID " + paramID);
  if (m_intParamRegs.count(paramID))
    return "int";
  if (m_doubleParamRegs.count(paramID))
    return "double";
  if (m_stringParamRegs.count(paramID))
    return "string";
  if (m_enumParamRegs.count(paramID))
    return "enum";

  throw domain_error("Invalid datatype for paramID " + paramID);
}

string
ParameterRegistry::getParamName(const string& paramID)
const
{
  string type = getParamType(paramID);
  if (type=="int") {
    map<string, ParamRegInt>::const_iterator pos;
    pos = m_intParamRegs.find(paramID);
    if (pos == m_intParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getName();
  }
  if (type=="double") {
    map<string, ParamRegDouble>::const_iterator pos;
    pos = m_doubleParamRegs.find(paramID);
    if (pos == m_doubleParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getName();
  }
  if (type=="string") {
    map<string, ParamRegString>::const_iterator pos;
    pos = m_stringParamRegs.find(paramID);
    if (pos == m_stringParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getName();
  }
  if (type=="enum") {
    map<string, ParamRegEnum>::const_iterator pos;
    pos = m_enumParamRegs.find(paramID);
    if (pos == m_enumParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getName();
  }
  throw domain_error("Invalid datatype for paramID " + paramID);
}

string
ParameterRegistry::getParamDesc(const string& paramID)
const
{
  string type = getParamType(paramID);
  if (type=="int") {
    map<string, ParamRegInt>::const_iterator pos;
    pos = m_intParamRegs.find(paramID);
    if (pos == m_intParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getDescription();
  }
  if (type=="double") {
    map<string, ParamRegDouble>::const_iterator pos;
    pos = m_doubleParamRegs.find(paramID);
    if (pos == m_doubleParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getDescription();
  }
  if (type=="string") {
    map<string, ParamRegString>::const_iterator pos;
    pos = m_stringParamRegs.find(paramID);
    if (pos == m_stringParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getDescription();
  }
  if (type=="enum") {
    map<string, ParamRegEnum>::const_iterator pos;
    pos = m_enumParamRegs.find(paramID);
    if (pos == m_enumParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getDescription();
  }
  throw domain_error("Invalid datatype for paramID " + paramID);
}

string
ParameterRegistry::getParamUnit(const string& paramID)
const
{
  string type = getParamType(paramID);
  if (type=="double") {
    map<string, ParamRegDouble>::const_iterator pos;
    pos = m_doubleParamRegs.find(paramID);
    if (pos == m_doubleParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getUnit();
  }
  if (type=="enum") {
      map<string, ParamRegEnum>::const_iterator pos;
      pos = m_enumParamRegs.find(paramID);
      if (pos == m_enumParamRegs.end())
        throw invalid_argument("Invalid paramID " + paramID);
      return pos->second.getUnit();
  }
  throw domain_error("Parameter paramID " + paramID + " of type " + type + " does not have unit");
}

int
ParameterRegistry::getParamDefValInt(const string& paramID)
const
{
  string type = getParamType(paramID);
  if (type=="int") {
    map<string, ParamRegInt>::const_iterator pos;
    pos = m_intParamRegs.find(paramID);
    if (pos == m_intParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getDefValue();
  }
  throw domain_error("Invalid paramID " + paramID);
}

double
ParameterRegistry::getParamDefValDouble(const string& paramID)
const
{
  string type = getParamType(paramID);
  if (type=="double") {
    map<string, ParamRegDouble>::const_iterator pos;
    pos = m_doubleParamRegs.find(paramID);
    if (pos == m_doubleParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getDefValue();
  }
  throw domain_error("Invalid paramID " + paramID);
}

string
ParameterRegistry::getParamDefValString(const string& paramID)
const
{
  string type = getParamType(paramID);
  if (type=="string") {
    map<string, ParamRegString>::const_iterator pos;
    pos = m_stringParamRegs.find(paramID);
    if (pos == m_stringParamRegs.end())
      throw invalid_argument("Invalid paramID " + paramID);
    return pos->second.getDefValue();
  }
  throw domain_error("Invalid paramID " + paramID);
}

vector<double>
ParameterRegistry::getParamEnumValues(const string& paramID)
const
{
  string type = getParamType(paramID);
  if (type=="enum") {
      map<string, ParamRegEnum>::const_iterator pos;
      pos = m_enumParamRegs.find(paramID);
      if (pos == m_enumParamRegs.end())
        throw invalid_argument("Invalid paramID " + paramID);
      return pos->second.getValues();
  }
  throw domain_error("Invalid paramID " + paramID);
}

unsigned
ParameterRegistry::getParamEnumDefIndex(const string& paramID)
const
{
  string type = getParamType(paramID);
  if (type=="enum") {
      map<string, ParamRegEnum>::const_iterator pos;
      pos = m_enumParamRegs.find(paramID);
      if (pos == m_enumParamRegs.end())
        throw invalid_argument("Invalid paramID " + paramID);
      return pos->second.getDefIndex();
  }
  throw domain_error("Invalid paramID " + paramID);
}



