////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#ifndef PARAMETERREGISTRY_H
#define PARAMETERREGISTRY_H

#include <string>
#include <set>
#include <map>
#include <vector>

using namespace std;

class ParamReg
{
public:
  ParamReg()  {};
  ParamReg(const string& name, const string& description);
  virtual ~ParamReg() {};

  string getName() const { return m_name; };
  string getDescription() const  { return m_description; };
private:
  string m_name;
  string m_description;
};


class ParamRegInt
: public ParamReg
{
public:
  ParamRegInt()  {};
  ParamRegInt(const string& name, const string& description, const int& defValue);

  int getDefValue() const { return m_defValue; };
private:
  int m_defValue;
};

class ParamRegDouble
: public ParamReg
{
public:
  ParamRegDouble()  {};
  ParamRegDouble(const string& name, const string& description, const double& defValue, const string& unit);

  double getDefValue() const { return m_defValue; };
  string getUnit() const { return m_unit; };
private:
  double m_defValue;
  string m_unit;
};

class ParamRegString
: public ParamReg
{
public:
  ParamRegString()  {};
  ParamRegString(const string& name, const string& description, const string& defValue);

  string getDefValue() const { return m_defValue; };
private:
  string m_defValue;
};

class ParamRegEnum
: public ParamReg
{
public:
  ParamRegEnum()  {};
  ParamRegEnum(const string& name,
               const string& description,
               const unsigned& nValues,
               const double values[],
               const int& defIndex,
               const string& unit);

  vector<double> getValues() const { return m_values; };
  unsigned getDefIndex() const { return m_defIndex; };
  string getUnit() const { return m_unit; };
private:
  vector<double> m_values;
  unsigned m_defIndex;
  string m_unit;
};




class ParameterRegistry
{
public:
  ParameterRegistry() {};
//  ParameterRegistry(const ParameterRegistry& parameterRegistry);
  virtual ~ParameterRegistry() {};

//  ParameterRegistry& operator=(const ParameterRegistry& rhs) {return *this;};

  void addParamReg(const string& paramID, const ParamReg* paramReg);

  bool isRegisteredParam(const string& paramID) const;
  set<string> getParamIDs() const;
  string getParamType(const string& paramID) const;
  string getParamName(const string& paramID) const;
  string getParamDesc(const string& paramID) const;
  string getParamUnit(const string& paramID) const;
  int getParamDefValInt(const string& paramID) const;
  double getParamDefValDouble(const string& paramID) const;
  string getParamDefValString(const string& paramID) const;
  vector<double> getParamEnumValues(const string& paramID) const;
  unsigned getParamEnumDefIndex(const string& paramID) const;

private:
  // All paramIDs, regardless of type:
  set<string> m_paramIDs;

  // The ParamRegs:
  // Key paramID; Val: ParamReg object
  map<string, ParamRegInt> m_intParamRegs;
  map<string, ParamRegDouble> m_doubleParamRegs;
  map<string, ParamRegString> m_stringParamRegs;
  map<string, ParamRegEnum> m_enumParamRegs;

};





#endif // PARAMETERREGISTRY_H
