////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <fstream>
#include <sstream>
#include <stdexcept>

#include "tablereader.h"

using namespace std;

const string TableReader::m_comment = "#";
const string TableReader::m_wordsep = " \t\f\v\n\r";

TableReader::TableReader(const string filename)
	: m_filename(filename),
	  m_fileParsed(false)
{}

TableReader::~TableReader()
{}

void
TableReader::stripComment(string& line)
const
{
  size_t pos;

  if ( line.length() ) {
    pos = line.find(m_comment);
    if ( pos != string::npos ) {
      line.erase(pos);
    }
  }
}

int
TableReader::countWords(const string& line)
const
{
  size_t pos0, pos1;

  int wc = 0;
  pos0 = line.find_first_not_of(m_wordsep);
  while ( pos0 != string::npos ) {
    wc += 1;
    pos1 = line.find_first_of(m_wordsep, pos0);
    if ( pos1 != string::npos )
      pos0 = line.find_first_not_of(m_wordsep, pos1);
    else
      pos0 = string::npos;
  }
  return wc;
}

vector<double>
TableReader::parseLine(string line)
const
{
	vector<double> vals;

	stripComment(line);

    int nwords = countWords(line);
    if ( nwords ) {
     	istringstream isline(line);
		for ( int iword=0; iword<nwords; iword++) {
			double val;
			isline >> val;
			vals.push_back(val);
		}
    }
    return vals;
}


void
TableReader::parseFile()
{
	ifstream file;
	unsigned int ncols = 0;
	string line;
	int irow = 0;
	vector<double> linevals;

	file.open(m_filename.c_str(), ios_base::in);
	if ( file.fail() )
		throw domain_error("Cannot open file " + m_filename);

	while ( getline(file, line) ) {
		// Count rows (1-based):
		irow++;

		linevals = parseLine(line);
	    if ( linevals.size() ) {
	    	// First non-empty line defines # columns:
	    	if ( ncols==0 ) {
	    		ncols = linevals.size();

	    		// Create a vector for each column:
	    		// Column numbers 1-based
	    		for ( unsigned int icol=1; icol<=ncols; icol++) {
	    			vector<double> colvals;
	    			m_vals[icol] = colvals;
	    		}
	    	}

	    	// Check consistent content of this line:
	    	if ( linevals.size()!=ncols ){
	    		ostringstream mes;
	    		mes << "Inconsistent data in row " << irow << " in file " << m_filename;
	    		throw domain_error(mes.str());
	    	}

	    	// Index for columns 1-based:
    		for ( unsigned int icol=1; icol<=ncols; icol++) {
    			m_vals[icol].push_back(linevals[icol-1]);
    		}
	    }
	}

	file.close();

	m_fileParsed = true;
}

vector<double>&
TableReader::read(const int icol)
{
	if ( !m_fileParsed )
		parseFile();
	return m_vals[icol];
}

