////////////////////////////////////////////////////////////////////////////////
//
// Copyright 2012, Reinhold Schaaf
// Argelander-Institut fuer Astronomie
// Universitaet Bonn
// rschaaf@astro.uni-bonn.de
//
// and the ARTIST team.
//
// For further info about copyright and license see included license.txt file.
//
////////////////////////////////////////////////////////////////////////////////

#include <algorithm> // std::transform
#include <cctype>    // std::tolower
#include <stdexcept>

#include "utilities.h"

using namespace std;

string whitespace(" \t\f\v\n\r");

vector<string>
split(const string& s)
{
  return split(s, whitespace);
}

vector<string>
split(const string& s, const string& sep)
{
  vector<string> words;

  if ( !s.empty() ) {
    size_t pos0, pos1;
    pos0 = pos1 = 0;
    while ( pos0 != string::npos && pos1 != string::npos ) {
      pos0 = s.find_first_not_of(sep, pos1);
      if ( pos0 != string::npos ) {
        pos1 = s.find_first_of(sep, pos0);
        if ( pos0 != string::npos ) {
          words.push_back( s.substr(pos0, pos1 - pos0) );
        }
      }
    }
  }
  return words;
}


string
lstrip(string s)
{
  return lstrip(s, whitespace);
}

string
lstrip(string s, const string& chars)
{
  if ( !s.empty() ) {
    size_t pos;
    pos = s.find_first_not_of(chars);
    if ( pos != string::npos )
      s = s.erase(0, pos);
  }
  return s;
}

string
rstrip(string s)
{
  return rstrip(s, whitespace);
}

string
rstrip(string s, const string& chars)
{
  if ( !s.empty() ) {
    size_t pos;
    pos = s.find_last_not_of(chars);
    if ( pos != string::npos )
      s.erase(pos+1);
    else
      s.clear();
  }
  return s;
}

string
strip(string s)
{
  return strip(s, whitespace);
}

string
strip(string s, const string& chars)
{
  return rstrip(lstrip(s, chars), chars);
}

string
upper(string s)
{
  transform(s.begin(), s.end(), s.begin(), ::toupper);
  return s;
}

string
lower(string s)
{
  transform(s.begin(), s.end(), s.begin(), ::tolower);
  return s;
}


void
heapsort(vector<double>& arr, vector<int>& arr2)
{
  // Sorts arr into ascending numerical order, while
  // making the corresponding rearrangements of arr2
  // Adopted from Press et al. "Numerical Recipes in C".
  // See also
  // http://en.wikibooks.org/wiki/Algorithm_Implementation/Sorting/Heapsort

  // Check validity of input data:
  if ( arr.size() != arr2.size() )
    throw invalid_argument("Sizes of arrays do not match");

  // Trivial case:
  if ( arr.size() == 0 )
    return;

  // temporary values
  double t;
  int t2;
  // Heap Indices:
  unsigned n = arr.size();
  unsigned parent = n / 2;
  unsigned index, child;

  while (1) {
    if ( parent > 0 ) {
      // First stage - Sorting the heap
      t  = arr[--parent];
      t2 = arr2[parent];
    } else {
      // Second stage - Extracting elements in-place
      n--;
      if ( n == 0 ) {
        // Heap empty - Done
        return;
      }
      // Save lost heap entry
      t  = arr[n];
      t2 = arr2[n];
      // Save root entry beyond heap
      arr[n]  = arr[0];
      arr2[n] = arr2[0];
    }
    // Insert operation - pushing t and t2 down the heap
    // to replace the parent
    index = parent;
    child = index * 2 + 1;
    while ( child < n ) {
      // Choose the last child
      if ( child + 1 < n && arr[child+1] > arr[child] ) {
        // Right child exists and is bigger
        child++;
      }
      // Is the largest child larger than the entry?
      if ( arr[child] > t ) {
        // Overwrite entry with child
        arr[index]  = arr[child];
        arr2[index] = arr2[child];
        // Move index to the child
        index = child;
        // Get the left child and go around again
        child = index * 2 + 1;
      } else {
        // t's place is found
        break;
      }
    }
    // Store the temporary values at their new locations
    arr[index]  = t;
    arr2[index] = t2;
  }
}
